#include <rccl.h>
#include "manex.260.h"

#define BOXSIZE         100.0

#ifdef Vxw_TARGET
#define main box_260_
#endif

main()
{
        TRSF_PTR e, b, c[4];
        POS_PTR corners[4], start;
        MANIP *mnp;
        char *robotName;
        int k, ia, iv;

        static float accelValues[] = { F_DEFAULT, 2000.0, 20.0 };
        static char *accelDesc[] = { "default", "high", "low" };
        static float velValues[] = { F_DEFAULT, 300.0, 30.0 };
        static char *velDesc[] = { "default", "fast", "slow" };

        rcclSetOptions (RCCL_ERROR_EXIT);
        robotName = getDefaultRobot();

        e = allocTransXyz ("E", UNDEF, 0.0, 0.0, TOOLZ);
        b = allocTransRot ("B", UNDEF, B_X, B_Y, B_Z, xunit, 180.0);
        start = makePosition ("start", T6, e, EQ, b, TL, e);

        c[0] = allocTransXyz (NULL, UNDEF,  BOXSIZE/2, 0.0, -BOXSIZE/2);
        c[1] = allocTransXyz (NULL, UNDEF, -BOXSIZE/2, 0.0, -BOXSIZE/2);
        c[2] = allocTransXyz (NULL, UNDEF, -BOXSIZE/2, 0.0,  BOXSIZE/2);
        c[3] = allocTransXyz (NULL, UNDEF,  BOXSIZE/2, 0.0,  BOXSIZE/2);
        for (k=0; k<4; k++)
         { corners[k] = makePosition (NULL, T6, e, EQ, b, c[k], TL, e);
         }

        mnp = rcclCreate (robotName, 0);
        rcclStart();

        move (mnp, start);
        stop (mnp, 1000.0);
        waitForCompleted (mnp);

        setMod (mnp, 'c');

        for (ia=0; ia<ARRAY_SIZE(accelValues); ia++)
         { for (iv=0; iv<ARRAY_SIZE(velValues); iv++)
            { float vel, accel;

              setCartAccel (mnp, accelValues[ia], F_UNDEF);
              setCartVel (mnp, velValues[iv], F_UNDEF);
              getCartAccel (mnp, &accel, (float*)NULL);
              getCartVel (mnp, &vel, (float*)NULL);

              printf ("%s vel (%g), %s acceleration (%g):\n",
                      velDesc[iv], vel, accelDesc[ia], accel);

              moveSet (mnp, corners, ARRAY_SIZE(corners));
              waitForCompleted (mnp);
            }
         }

        setMod (mnp, 'j');
        move (mnp, start);
        stop (mnp, 1000.0);
        waitForCompleted (mnp);

        rcclRelease (YES);
}

moveSet (mnp, p, size)
MANIP *mnp;
POS_PTR *p;
int size;
{
        int k;

        for (k=0; k<size; k++)
         { move (mnp, p[k]);
         }
}
