#ifdef X86sol_TARGET
#include "pmc_pid.h"
#include "mc8prim.h"
#include "rciioctl.h"
#include <sys/types.h>
#include <fcntl.h>
#include <strings.h>
#include <stdio.h>

#include "../h/ctree.h"

char key_buf [256];

#define PMC_ADDR	0x380

union pack3 {
	short s[4];
	int i[2];
  };

static void writeInt (mc8_struct *pmc8, int addr, int val)
{
	union pack3 xfer;
	
	xfer.i[0] = val;
	mc8WriteMem (pmc8, addr, xfer.s, /*numWords=*/1);
}

static int readInt (mc8_struct *pmc8, int addr)
{
	union pack3 xfer;

	mc8ReadMem (pmc8, addr, xfer.s, /*numWords=*/1);
	return xfer.i[0];
}

main()
{
	mc8_struct *pmc8;
	char *keytree;
	int k, rciFd;
	int chan;

	if ((rciFd = open ("/dev/genrci0", O_RDWR)) == -1)
	 { perror ("open genrci0");
	   exit (1);
	 }

	if (ioctl (rciFd, RCIIOC_ENABLE_IOPORTS, 1) == -1 ||
	    ioctl (rciFd, RCIIOC_ENABLE_IOPORTS, 0) == -1)
	 { perror ("ioctl RCIIOC_ENABLE_IOPORTS");
	   exit (1);
	 }

	if ((pmc8 = mc8Open (PMC_ADDR, 0, 0)) == NULL)
	 { fprintf (stderr, "no response from PMC\n");
	   exit (1);
	 }
	/* check magic and version number */
	k = readInt (pmc8, PID_MAGIC_ADDR);
	if (k != PID_MAGIC)
	 { fprintf (stderr, "bad magic number\n");
	   exit (1);
	 }
	k = readInt (pmc8, PID_VERSION_ADDR);
	if (k != PID_VERSION)
	 { fprintf (stderr, "wrong version: %d vs. %d\n",
			     k, PID_VERSION);
	   exit (1);
	 }


	keytree = tree_match_parse("					\
(	quit								\
	show								\
)",0);
#if 0
	"[free lock]							\
	 (all-joints							\
	  one-joint							\
	   (%d",&joint,",1,6,		\"joint number (1 to 6)\")	\
	 )								\
	show (encoder-values)						"
#endif
	if (keytree == 0)
	 { printf ("ERROR/Parse error at %.30s\n", tree_match_err_at());
	   exit (-1);
	 }
	do
 	 { tree_match (keytree, "PROMPT> ", key_buf);

	   if (COMMAND("show"))
	    { for (k=0; k<9; k++)
	       { for (chan=0; chan<PID_NCHAN; chan++)
		  { int x;
		    x = readInt (pmc8, PID_CHAN_BASE(chan)+k);
		    printf ("%8x ", x);
		  }
		 printf ("\n");
	       }
	    }
	 }
	while (!(COMMAND("quit")));
}
#endif

