/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#include <stdio.h>
#include <math.h>

#include "../h/cdefs.h"
#include "../h/robotClass.h"
#include "../h/gantry_kynvar.h"
#include "../h/errorCodes.h"
#include "../h/errorStack.h"
#include "../h/robotMath.h"
#include "../h/robotTypes.h"

#define EPSILONF 	1.0e-7

printRot (char *s, TRSF *R)
{
	printf ("%s:\n", s);
	printf (
"%10.4f %10.4f %10.4f\n%10.4f %10.4f %10.4f\n%10.4f %10.4f %10.4f\n",
R->n.x, R->o.x, R->a.x, 
R->n.y, R->o.y, R->a.y, 
R->n.z, R->o.z, R->a.z);
}

gantry_kyn (kyn, fp)
KYN *kyn;
FILE *fp;
{
	GANTRY_KYN *skyn = (GANTRY_KYN*)kyn->ext;
	ConfCodeDesc *desc;
	char buf[64];
	int j;

	char *fxnName = "gantry_kyn()";

	readInit(fp,"=");

	if (!readExactString("BeginDHParams"))
	 { errorMsg (EBadFile,
"%s -- missing keyword 'BeginDHParam'\n", fxnName);
	   return -1;
	 }

	kyn->gen->dh[0].alpha = -PI/2;
	kyn->gen->dh[1].alpha =  PI/2;
	kyn->gen->dh[1].theta = -PI/2;
	kyn->gen->dh[2].alpha =  PI;
	kyn->gen->dh[2].theta =  PI;
	kyn->gen->dh[3].alpha = -PI/2;
	kyn->gen->dh[4].alpha =  PI/2;


	do
	 { double value;

	   if (!readIdentifier (buf, 64))
	    { errorMsg (EBadFile,
"%s -- can't read DHparam identifier\n", fxnName);
	      return -1;
	    }
	   if (strcmp (buf, "EndDHParams") != 0)
	    { if (!readExactString("=") ||
		  !readDouble (&value))
	       { errorMsg (EBadFile, 
"%s -- can't read value associated with '%s'\n", fxnName, buf);
		 return -1;
	       }
	      j = buf[strlen(buf)-1]-'1';
	      if (j < 0 || j > 4)
	       { errorMsg (EBadFile, 
"%s -- bogus joint number at end of DH param identifier '%s'\n",
			   fxnName, buf);
		 return -1;
	       }
	      if (strncmp (buf, "alpha", 5) == 0)
	       { kyn->gen->dh[j].alpha = DEGTORAD*value;
	       }
	      else if (strncmp (buf, "theta", 5) == 0)
	       { kyn->gen->dh[j].theta = DEGTORAD*value;
	       }
	      else if (strncmp (buf, "d", 1) == 0)
	       { kyn->gen->dh[j].d = value;
	       }
	      else if (strncmp (buf, "a", 1) == 0)
	       { kyn->gen->dh[j].a = value;
	       }
	      else if (strncmp (buf, "beta", 4) == 0)
	       { kyn->gen->dh[j].beta = DEGTORAD*value;
	       }
	      else
	       { errorMsg (EBadFile,
"%s -- bogus DH param identifier '%s'\n", fxnName, buf);
		 return -1; 
	       }
	    }
	 }
	while (strcmp (buf, "EndDHParams") != 0);

/*  	if (!readExactString("Alpha2") || !readExactString("=") || */
/*  	    !readFloat(&alpha2) || */
/*  	    !readExactString("Theta2") || !readExactString("=") || */
/*  	    !readFloat(&theta2)) */
/*  	 { errorMsg (EBadFile,  */
/*  "%s -- can't read DH parameters Theta2 and Alpha2\n", fxnName); */
/*  	   return (-1); */
/*  	 } */
	printf ("Gantry DH params:\n");
	for (j=0; j<5; j++)
	 { printf ("%8.3f %8.3f %8.3f %8.3f %8.3f\n",
		   kyn->gen->dh[j].d, RADTODEG*kyn->gen->dh[j].theta,
		   kyn->gen->dh[j].a, RADTODEG*kyn->gen->dh[j].alpha,
		   RADTODEG*kyn->gen->dh[j].beta);
	 }

	if (gen_readNomVelAcc (fxnName, kyn->gen) == -1)
	 { return (-1);
	 }

	skyn->s2 = sin(kyn->gen->dh[1].theta);
	skyn->c2 = cos(kyn->gen->dh[1].theta);
	skyn->sa = sin(kyn->gen->dh[1].alpha);
	skyn->ca = cos(kyn->gen->dh[1].alpha);

	identTrsf (&skyn->R03);
	multTrsfRotY (&skyn->R03, RADTODEG*PI/2);
	skyn->axis1.x = 1; 
	skyn->axis1.y = 0; 
	skyn->axis1.z = 0; 

	multTrsfRotX (&skyn->R03, RADTODEG*kyn->gen->dh[0].alpha);

	skyn->axis2.x = skyn->R03.a.x;
	skyn->axis2.y = skyn->R03.a.y;
	skyn->axis2.z = skyn->R03.a.z; 

	multTrsfRotZ (&skyn->R03, RADTODEG*kyn->gen->dh[1].theta);
	multTrsfRotX (&skyn->R03, RADTODEG*kyn->gen->dh[1].alpha);
	skyn->axis3.x = skyn->R03.a.x;
	skyn->axis3.y = skyn->R03.a.y;
	skyn->axis3.z = skyn->R03.a.z;

	multTrsfRotZ (&skyn->R03, RADTODEG*PI);
	multTrsfRotX (&skyn->R03, RADTODEG*kyn->gen->dh[2].alpha);
	multTrsfRotY (&skyn->R03, RADTODEG*kyn->gen->dh[2].beta);

	/* ---- Configuration codes ---- */
	desc = kyn->gen->confCodes;
	kyn->gen->numConf = 1;
	desc[0].code  = GANTRY_FLIP_CONF;
	desc[0].set   = 'f';
	desc[0].clear = 'n';
	
	/* ---- Instantiate function fields ---- */

	kyn->gen->fwdKinematics = gantryFwdKinematics;
	kyn->gen->invKinematics = gantryInvKinematics;
	kyn->gen->rangeAngles = gantryRangeAngles;
	kyn->gen->solveConf = gantrySolveConf;
	kyn->gen->fwdJacob = gantryFwdJacob;
	kyn->gen->invJacob = gantryInvJacob;
	kyn->gen->fwdJacobT = gantryFwdJacobT;
	kyn->gen->invJacobT = gantryInvJacobT;
	kyn->gen->gravload = gantryGravload;
	kyn->gen->fwdJacobVar = gantryFwdJacobVar;
	kyn->gen->invJacobVar = gantryInvJacobVar;
	kyn->gen->fwdJacobTVar = gantryFwdJacobTVar;
	kyn->gen->invJacobTVar = gantryInvJacobTVar;
	kyn->gen->gravloadVar = gantryGravloadVar;
	kyn->gen->updateVar = gantryUpdateVar;
	kyn->gen->updateVarXsincos = gantryUpdateVarXsincos;

	kyn->gen->rbtclass = GANTRY;

	return (1);
}
