#include <stdio.h>
#include "../h/stringDefs.h"
#include <sys/types.h>

#ifdef Qnx_TARGET
#define SYSV
#endif

#ifndef SYSV
#include <sys/wait.h>
#endif

#define MAXARGS		2048
#define DEFAULT_CPP	"/usr/X11/bin/cpp"

int cppArgc;
char *cppArgv[MAXARGS+2];

char **argvec;
int argcnt;

addArg (arg)
char *arg;
{
	if (cppArgc >= MAXARGS)
	 { fprintf (stderr, "%s: too many arguments\n", *argvec);
	   exit (1);
	 }
	cppArgv[cppArgc++] = arg;
}

isCfile (arg)
const char *arg;
{
	char *cp;

	if ((cp = charRindex(arg, '.')) &&
	     strcmp (cp, ".c") == 0)
	 { return (1);
	 }
	else
	 { return (0);
	 }
}

execute (arg)
char *arg;
{
#ifdef SYSV
	int	status;
#else	/* !SYSV */
	union wait	status;
#endif	/* !SYSV */
	int pid;

	cppArgv[cppArgc] = arg;	
	cppArgv[cppArgc+1] = NULL;

	if ((pid = fork()))
	 { if (pid < 0)
	    { fprintf (stderr, "%s: cannot fork\n", *argvec);
	      perror ("");
	      exit (1);
	    }
	   else
	    { fprintf (stderr, "%s ... ", arg);
	      fflush (stderr);
	      while (wait(&status) > 0)
	       { 
#ifdef SYSV
		 if ((status >> 8) & 0xff)
		  { fprintf (stderr, "%s: Signal %d.\n", 
			     cppArgv[0], (status >> 8) & 0xff);
		    return;
		  }
		 if (status & 0xff)
		  { fprintf (stderr, "%s: Exit code %d.\n",
			     cppArgv[0], status & 0xff);
		    return;
		  }
#else	/* !SYSV */
		 if (status.w_termsig)
		  { fprintf (stderr, "%s: Signal %d.\n", 
			     cppArgv[0], status.w_termsig);
		    return;
		  }
		 if (status.w_retcode)
		  { fprintf (stderr, "%s: Exit code %d.\n", 
			     cppArgv[0], status.w_retcode);
		    return;
		  }
#endif	/* !SYSV */
	       }
	      fprintf (stderr, "done\n");
	    }
	 }
	else
	 { if (execv (DEFAULT_CPP, cppArgv) < 0)
	    { fprintf (stderr, "%s: can't exec %s\n", *argvec, DEFAULT_CPP);
	      perror ("");
	      exit (1);
	    }
	 }
}

main(argc, argv)
int argc;
char **argv;
{
	argvec = argv;
	argcnt = argc;

	addArg (DEFAULT_CPP);
	addArg ("-M");

	argc--; argv++;
	while (argc > 0)
	 { if (!isCfile (*argv))
	    { addArg (*argv);
	    }
	   argc--; argv++;
	 }

	argc = argcnt;
	argv = argvec;

	argc--; argv++;
	while (argc > 0)
	 { if (isCfile (*argv))
	    { execute (*argv);
	    }
	   argc--; argv++;
	 }
	exit (0);
}
