/*=-=-=-=-=-=-=-=-=-=-  RCCL/RCI Notice of Copyright  -=-=-=-=-=-=-=-=-=-=-=*
 *									    *
 *	Copyright (C) 1987 by Mike Parker. Permission to use, distribute    *
 *	and copy is granted for non-commercial and research purposes only,  *
 *	providing the authors are acknowledged and this copyright notice is *
 *	displayed. The authors provide this software in hope that it will   *
 *	useful, but do not assume responsibility for its use.		    *
 *									    *
 *	This software was developed using the facilities of the McGill 	    *
 *	University Research Center for Intelligent Machines, Montreal,	    *
 *	Canada, for a project undertaken and supported by the RCA Advanced  *
 *	Technology Laboratories, Moorestown, New Jersey.		    *
 *									    *
 *=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/

#include <stdio.h>
#include <ctype.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>
#include <utmp.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/types.h>
#include <fcntl.h>
#ifndef FD_ZERO
#include "broken-select.h"
#endif
#ifdef HAS_UNISTD_H
#include <unistd.h>
#endif

#include "../h/stringDefs.h"
#include "../h/termline.h"

#ifdef HAS_SGTTY
#define TURN_OFF_SUPPORTED		1
#else
#define TURN_OFF_SUPPORTED		0
#endif

char *malloc();
char *realloc();
char *copyofblk();
extern void free();
#define NEW(type) ((type *) malloc(sizeof(type)))
#define OLD(x) free((char *) x)

char **argvec;

char *tree_match_parse();

char *device;
char *init_file;
int send_init_file;
int errs;
int speed;
int breakc;
int pstrip;
int tty_fd;
int line_fd;
int turn_off;
int old_active;
int debug_mode;
int nlmap;
int echo;
int dohup;
int dohup_after;
int fax_sync;
#define SYNC_BLIND 0
#define SYNC_ECHO 1
int fax_echo;

TermProps old_tty;
TermProps old_line;
TermProps new_tty;
TermProps new_line;
int strip_parity;
int photofd;
char *tree;
char kt_file[1024];
int baud_rate;
int kt_baud;
char cmd_buf[1024];
int fax_fd;

do_hangup()
{
	if (debug_mode)
	 { fprintf(stderr,"Doing a hangup...");
	 }
	do_setspeed(0);
  	nap(0.5);
   	if (debug_mode)
	 { fprintf(stderr,"\r\n");
	 }
	do_setspeed(speed);
}

do_setspeed(s)
int s;
{
	termlineSpeed (&new_line, s);
	termlineSet (line_fd, &new_line);
	termlineOutputFlow (line_fd, /*on=*/1);
}

int iswhite(c)
char c;
{
 return((c==' ')||(c=='\t'));
}

int matches(s,pat)
register char *s;
register char *pat;
{
 for (;*pat;s++,pat++)
  { if (*s == *pat)
     { continue;
     }
    if (iswhite(*pat) && iswhite(*s))
     { continue;
     }
    return(0);
  }
 return (1);
}

int skipfields(n,str)
int n;
char *str;
{
 int quoted;
 char *cp;

 cp = str;
 quoted = 0;
 while (isspace(*cp))
  { cp ++;
  }
 for (;n>0;n--)
  { while (quoted || !isspace(*cp))
     { quoted ^= (*cp == '"');
       cp ++;
     }
    while (isspace(*cp))
     { cp ++;
     }
  }
 return(cp-str);
}

#if TURN_OFF_SUPPORTED
int set_line_active(tty,active)
char *tty;
int active;
{
 static char utmp_file[] = "/etc/utmp";
 static char ttys_file[] = "/etc/ttys";
 char ttys_line[1024];
 struct utmp ut;
 char *dev;
 int fd;
 FILE *f;
 int pos;
 int old;
 int didchange;
 int i;

 if (strncmp(tty,"/dev/",5) != 0)
  { return(0);
  }
 if (geteuid() != 0)
  { return(0);
  }
 dev = tty + 5;
 fd = open(utmp_file,O_RDONLY,0);
 if (fd < 0)
  { fprintf(stderr,"%s: cannot open %s: ",argvec[0],utmp_file);
    perror(0);
    exit(1);
  }
 while (read(fd,&ut,sizeof(struct utmp)) == sizeof(struct utmp)) /* foo */
  { if (strncmp(ut.ut_line,dev,sizeof(ut.ut_line)) == 0)
     { if (ut.ut_name[0])
	{ fprintf(stderr,"%s: %s is busy\n",argvec[0],tty);
	  exit(1);
	}
       break;
     }
  }
 close(fd);
 f = fopen(ttys_file,"r+");
 if (f == NULL)
  { return(0);
  }
 old = 0;
 didchange = 0;
 while (1)
  { pos = ftell(f);
    if (fgets(ttys_line,sizeof(ttys_line),f) != ttys_line)
     { break;
     }
    if (strncmp(ttys_line,dev,strlen(dev)) == 0)
     { i = skipfields(3,ttys_line);
       if (i > 0)
	{ int ok;
	  ok = 0;
	  if (matches(&ttys_line[i],"off "))
	   { old = 0;
	     ok = 1;
	   }
	  if (matches(&ttys_line[i],"on  "))
	   { old = 1;
	     ok = 1;
	   }
	  if (ok && (old != active))
	   { if (active)
	      { ttys_line[i+1] = 'n';
		ttys_line[i+2] = ' ';
	      }
	     else
	      { ttys_line[i+1] = 'f';
		ttys_line[i+2] = 'f';
	      }
	     fseek(f,pos,0);
	     fputs(ttys_line,f);
	     didchange = 1;
	   }
	}
       break;
     }
  }
 fclose(f);
 if (didchange)
  { kill(1,SIGHUP);
    sleep(1);
  }
 return(old);
}
#endif

maybe_turnoff()
{
#if TURN_OFF_SUPPORTED
 if (old_active >= 0)
  { set_line_active(device,old_active);
  }
#endif
}

setup_tty()
{
	termlineSet (tty_fd, &new_tty);
	termlineOutputFlow (tty_fd, /*on=*/1);
}

reset_tty()
{
	termlineSet (tty_fd, &old_tty);
	termlineOutputFlow (tty_fd, /*on=*/1);
}

/* max number of nested IFs */
#define MAXIF 10
/* values for the leading word on the line */
#define SEND 1
#define EXPECT 2
#define IF 3
#define ELSEIF 4
#define ELSE 5
#define ENDIF 6
#define FAIL 7
#define SUCCEED 8
#define TIMEOUT 9
/* conditions for IF */
#define OK 1001
#define ERR 1002
#define GOT 1003

typedef struct {
	  char *key;
	  int value; } KEY;

int ifstack[MAXIF];
#define IF_NOW 1
#define IF_NOTYET 2
#define IF_PAST 3
int ifstackd;
#define INFALSEIF() ((ifstackd>=0)&&(ifstack[ifstackd]!=IF_NOW))
char *lastex = 0;
int gotit;
int sendfail;

sendstring(str)
char *str;
{
 register char *cp;
 char c;
 int nonl;

 nonl = 0;
 for (cp=str;*cp;cp++)
  { switch (*cp)
     { case '\\':
	  switch (*++cp)
	   { case '\0':
		cp --;
		c = '\\';
		break;
	     case 'b':
		c = '\b';
		break;
	     case 'c':
		nonl = 1;
		break;
	     case 'e':
		c = '\33';
		break;
	     case 'n':
		c = '\n';
		break;
	     case 'r':
		c = '\r';
		break;
	     case 't':
		c = '\t';
		break;
	     case '0': case '1': case '2': case '3':
	     case '4': case '5': case '6': case '7':
		c = *cp++ - '0';
		if (isascii(*cp) && charIndex("01234567",*cp))
		 { c = (c << 3) + (*cp++ - '0');
		   if (isascii(*cp) && charIndex("01234567",*cp))
		    { c = (c << 3) + (*cp - '0');
		    }
		   else
		    { cp --;
		    }
		 }
		else
		 { cp --;
		 }
		break;
	   }
	  break;
       default:
	  c = *cp;
	  break;
     }
    write(line_fd,&c,1);
  }
}

int getstring(expect,timeout,got)
char *expect;
double timeout;
char **got;
{
 char *sofar;
 int have;
 int used;
 struct timeval end;
 struct timeval now;
 struct timeval tv;
 fd_set fds;
 int selrv;
 char c;
 int xlen;

 gettimeofday(&end,(struct timezone *)0);
 end.tv_sec += (int) timeout;
 timeout -= (int) timeout;
 end.tv_usec += (int) (timeout*1e6);
 while (end.tv_usec >= 1000000)
  { end.tv_usec -= 1000000;
    end.tv_sec ++;
  }
 xlen = strlen(expect);
 sofar = malloc(8);
 have = 8;
 used = 0;
 while (1)
  { gettimeofday(&now,(struct timezone *)0);
    tv.tv_sec = end.tv_sec - now.tv_sec;
    tv.tv_usec = end.tv_usec - now.tv_usec;
    if (tv.tv_usec < 0)
     { tv.tv_usec += 1000000;
       tv.tv_sec --;
     }
    FD_ZERO(&fds);
    FD_SET(line_fd,&fds);
    selrv = select(FD_SETSIZE,&fds,(fd_set *)0,(fd_set *)0,&tv);
    if (selrv < 0)
     { perror("select");
       continue;
     }
    if (selrv == 0)
     { *got = sofar;
       return(0);
     }
    if (FD_ISSET(line_fd,&fds) && (read(line_fd,&c,1) == 1))
     { while (have <= used)
	{ sofar = realloc(sofar,have+=8);
	}
       sofar[used++] = c;
       if ((used >= xlen) && (strncmp(sofar+used-xlen,expect,xlen) == 0))
	{ sofar = realloc(sofar,used+1);
	  sofar[used] = '\0';
	  *got = sofar;
	  return(1);
	}
     }
  }
}

int findstr(sub,str)
char *sub;
char *str;
{
 char *maybe;
 int sublen;

 sublen = strlen(sub);
 if (sublen == 0)
  { return(1);
  }
 while (1)
  { maybe = charIndex(str,*sub);
    if (! maybe)
     { return(0);
     }
    if (strncmp(maybe,sub,sublen) == 0)
     { return(1);
     }
    str = maybe + 1;
  }
}

keymatch(cpp,keys,valp)
char **cpp;
KEY *keys;
int *valp;
{
 int i;
 register char *cp;
 register char *kp;
 register char *sp;

 for (i=0;kp=keys[i].key;i++)
  { cp = *cpp;
    while (*kp)
     { if ((isascii(*cp) && isupper(*cp)) ? (tolower(*cp) != *kp) : (*cp != *kp))
	{ break;
	}
       cp ++;
       kp ++;
     }
    if (! *kp)
     { if (!*cp || (isascii(*cp) && isspace(*cp)))
	{ *cpp = cp + !!*cp;
	  *valp = keys[i].value;
	  return;
	}
     }
  }
}

doif(str)
char *str;
{
 static KEY ifkeys[] = { { "err", ERR },
			 { "got", GOT },
			 { "ok", OK },
			 { 0, 0 } };
 int cond;

 cond = -1;
 keymatch(&str,ifkeys,&cond);
 switch (cond)
  { case -1:
       fprintf(stderr,"%s: bad if condition in init file\n",argvec[0]);
       ifstack[ifstackd] = IF_PAST;
       return;
       break;
    case ERR:
       cond = !gotit;
       break;
    case GOT:
       cond = findstr(str,lastex);
       break;
    case OK:
       cond = gotit;
       break;
  }
 ifstack[ifstackd] = cond ? IF_NOW : IF_NOTYET;
}

int init_conversation()
{
 FILE *initf;
 char ifbuf[512];
 int mode;
 int defmode;
 static KEY keywords[] = { { "else", ELSE },
		    { "elseif", ELSEIF },
		    { "endif", ENDIF },
		    { "expect", EXPECT },
		    { "fail", FAIL },
		    { "if", IF },
		    { "send", SEND },
		    { "succeed", SUCCEED },
		    { "timeout", TIMEOUT },
		    { 0, 0 } };
 int l;
 int kl;
 int i;
 char *cp;
 double timeout;
 int rv;

 initf = fopen(init_file,"r");
 if (initf == 0)
  { fprintf(stderr,"%s: can't open init file %s\n",init_file);
    return(1);
  }
 sendfail = 0;
 defmode = SEND;
 ifstackd = -1;
 timeout = 60;
 while (fgets(ifbuf,sizeof(ifbuf),initf) == ifbuf)
  { l = strlen(ifbuf);
    if ((l > 0) && (ifbuf[l-1] == '\n'))
     { ifbuf[--l] = '\0';
     }
    if (ifbuf[0] == '#')
     { continue;
     }
    mode = defmode;
    cp = ifbuf;
    keymatch(&cp,keywords,&mode);
    switch (mode)
     { case SEND:
	  if (sendfail)
	   { rv = FAIL;
	     goto out;
	   }
	  if (! INFALSEIF())
	   { sendstring(cp);
	     defmode = EXPECT;
	   }
	  break;
       case EXPECT:
	  if (! INFALSEIF())
	   { if (lastex)
	      { free(lastex);
	      }
	     gotit = getstring(cp,timeout,&lastex);
	     sendfail = ! gotit;
	     defmode = SEND;
	   }
	  break;
       case IF:
	  sendfail = 0;
	  if (ifstackd >= MAXIF-1)
	   { fprintf(stderr,"%s: sorry, init file if nesting too deep\n",argvec[0]);
	     rv = FAIL;
	     goto out;
	   }
	  if (INFALSEIF())
	   { ifstack[++ifstackd] = IF_PAST;
	   }
	  else
	   { ifstackd ++;
	     doif(cp);
	   }
	  break;
       case ELSEIF:
	  if (ifstackd < 0)
	   { fprintf(stderr,"%s: sorry, if-less elseif in init file\n",argvec[0]);
	     rv = FAIL;
	     goto out;
	   }
	  if (ifstack[ifstackd] != IF_NOTYET)
	   { ifstack[ifstackd] = IF_PAST;
	   }
	  else
	   { doif(cp);
	   }
	  break;
       case ELSE:
	  if (ifstackd < 0)
	   { fprintf(stderr,"%s: sorry, if-less else in init file\n",argvec[0]);
	     rv = FAIL;
	     goto out;
	   }
	  if (ifstack[ifstackd] == IF_NOTYET)
	   { ifstack[ifstackd] = IF_NOW;
	   }
	  else
	   { ifstack[ifstackd] = IF_PAST;
	   }
	  break;
       case ENDIF:
	  if (ifstackd < 0)
	   { fprintf(stderr,"%s: sorry, if-less endif in init file\n",argvec[0]);
	     rv = FAIL;
	     goto out;
	   }
	  ifstackd --;
	  break;
       case FAIL:
	  if (! INFALSEIF())
	   { rv = FAIL;
	     goto out;
	   }
	  break;
       case SUCCEED:
	  if (! INFALSEIF())
	   { rv = SUCCEED;
	     goto out;
	   }
	  break;
       case TIMEOUT:
	  sscanf(cp,"%lf",&timeout);
	  break;
     }
  }
 rv = sendfail ? FAIL : SUCCEED;
out:
 fclose(initf);
 return(rv==SUCCEED);
}

#if !TURN_OFF_SUPPORTED
printNoTurnOff()
{
	fprintf(stderr, "\
The -T option is not supported on this system. Normally, this option turns\n\
off logins on the port that you wish to connect to, for the duration of the\n\
termlink session. This may not be necessary because logins are often not\n\
enabled on spare serial ports. However, if they are enabled on the port you\n\
wish to use, they can be turned off permanently by your system admistrator.\n\
");
}
#endif

main(ac,av)
int ac;
char **av;
{
 int skip;
 int argno;
 int val;
 int i;

 argvec = av;

 tree = tree_match_parse("\
	(quit								\
	 hang-up-phone							\
	 talk								\
	 facsimile (%s",kt_file,")					\
	 kermit (send (%s",kt_file,")					\
		 receive						\
		)							\
	 set (echo (off on)						\
	      newline-mapping (off on)					\
	      speed (%d",&kt_baud,")					\
	      facsimile (sync (blind echo)				\
			 echo (on off)					\
			)						\
	     )								\
	)",0);
 if (tree == 0)
  { char *tree_match_err_at();
    fprintf(stderr,"Parse error: %.30s",tree_match_err_at());
    exit(1);
  }
 speed = 9600;
 device = 0;
 breakc = '~';
 errs = 0;
 skip = 0;
 dohup = 0;
 dohup_after = 0;
#if TURN_OFF_SUPPORTED
 turn_off = 0;
#else
 turn_off = -1;
#endif
 echo = 0;
 nlmap = 1;
 old_active = -1;
 fax_sync = SYNC_BLIND;
 fax_echo = 1;
 init_file = 0;
 send_init_file = 0;
 photofd = -1;
 debug_mode = 0;
 argno = 0;
 if (ac <= 1)
  { fprintf(stderr,
"Usage: %s -d device [ -h ] [ -H ] [ -s speed ] [ -b break-char ]\n\
	   [ -p photo-file ] %s[ -e ] [ -n ] [ -i init-file ]\n",
		turn_off == 0 ? "[ -T ] " : "", argvec[0]);
    fprintf(stderr,"or `%s -U' for full description\n",argvec[0]);
    exit(1);
  }
 for (ac--,av++;ac;ac--,av++)
  { if (skip > 0)
     { skip --;
       continue;
     }
    if (**av == '-')
     { for (++*av;**av;++*av)
	{ switch (**av)
	   { default:
		fprintf(stderr,"%s: bogus flag -%c\n",argvec[0],**av);
		errs = 1;
		break;
	     case 'b':
		if (++skip >= ac)
		 { fprintf(stderr,"%s: -%c needs another argument\n",
					argvec[0],**av);
		   errs = 1;
		 }
		breakc = av[skip][0];
		if ((breakc == '^') && av[skip][1])
		 { if (av[skip][1] == '?')
		    { breakc = 0177;
		    }
		   else
		    { breakc = av[skip][1] & 0x1f;
		    }
		 }
		break;
	     case 'd':
		if (++skip >= ac)
		 { fprintf(stderr,"%s: -%c needs another argument\n",
					argvec[0],**av);
		   errs = 1;
		 }
		else if (device)
		 { fprintf(stderr,"%s: at most one -%d flag\n",argvec[0],**av);
		   errs = 1;
		 }
		else
		 { device = av[skip];
		 }
		break;
	     case 'D':
		debug_mode ++;
		break;
	     case 'e':
		echo = 1;
		break;
	     case 'h':
		dohup = 1;
		break;
	     case 'H':
		dohup_after = 1;
		break;
	     case 'i':
		if (++skip >= ac)
		 { fprintf(stderr,"%s: -%c needs another argument\n",
					argvec[0],**av);
		   errs = 1;
		 }
		init_file = av[skip];
		send_init_file = 1;
		break;
	     case 'n':
		nlmap = 0;
		break;
	     case 'p':
	     case 'P':
		strip_parity = (**av == 'p');
		if (++skip >= ac)
		 { fprintf(stderr,"%s: -%c needs another argument\n",
					argvec[0],**av);
		   errs = 1;
		 }
		if (photofd >= 0)
		 { fprintf(stderr,"%s: extra photo file spec ignored\n",
								argvec[0]);
		   errs = 1;
		 }
		else
		 { int bad;
		   if (access(av[skip],W_OK) == 0)
		    { bad = 0;
		    }
		   else if (errno != ENOENT)
		    { bad = 1;
		    }
		   else
		    { char *cp = charRindex(av[skip],'/');
		      if (cp)
		       { *cp = '\0';
			 bad = access(av[skip],W_OK);
			 *cp = '/';
		       }
		    }
		   if (bad || ((photofd=open(av[skip],O_WRONLY|O_CREAT|O_TRUNC,0644)) < 0))
		    { fprintf(stderr,"%s: cannot open photo file %s: ",
						argvec[0],av[skip]);
		      perror((char *)0);
		      photofd = -1;
		      errs = 1;
		    }
		   else
		    { fchown(photofd,getuid(),getgid());
		      if (debug_mode)
		       { printf("ruid %d rgid %d euid %d egid %d\n",
				getuid(),getgid(),geteuid(),getegid());
		       }
		    }
		 }
		break;
	     case 's':
		if (++skip >= ac)
		 { fprintf(stderr,"%s: -%c needs another argument\n",
					argvec[0],**av);
		   errs = 1;
		 }
		if (sscanf(av[skip],"%d",&val) != 1)
		 { fprintf(stderr,"%s: -%c %s: cannot convert number\n",
				   argvec[0],**av,av[skip]);
		   errs = 1;
		 }
		else
		 { if (termlineSpeed ((TermProps*)0, val) < 0)
		    { fprintf(stderr,"%s: -%c %s: speed not available\n",
				      argvec[0],**av,av[skip]);
		      errs = 1;
		    }
		   else
		    { speed = val;
		    }
		 }
		break;
	     case 'T':
#if TURN_OFF_SUPPORTED
		turn_off = 1;
#else
		printNoTurnOff();
		exit (1);
#endif
		break;
	     case 'U':
		fprintf(stderr,"\
Arguments:\n\
\n\
-d device	Specifies what device to link to.  This must be the full\n\
		device name (with the /dev/ and all).\n\
\n\
-h              Tells termlink to hang up the line before starting\n\
\n\
-H              Tells termlink to hang up the line after finishing\n\
\n\
-s speed	Specifies what baud rate to set the device to.\n\
\n\
-n		Tells termlink that when sending files, it should send\n\
		them exactly as they are.  Normally newlines are sent as\n\
		carriage returns.\n\
\n\
-e		Tells termlink to echo everything you type back to your\n\
		terminal.\n\
\n\
-b break-char	Specifies the break character.  This character will break\n\
		back to your shell.  To specify a control character, you\n\
		can specify a ^ followed by some other character and it\n\
		will be turned into the appropriate control character (^?\n\
		gets turned into DELETE).  Note that to specify ^? or ~ or\n\
		some other special characters you may have to quote them\n\
		to protect them from the shell.  By the way, ^ alone\n\
		specifies that the break character is a ^.\n\
\n\
-p photo-file	Names a file to which a photo of the entire session is to\n\
-P photo-file	be written.  If -p is used, the highest bit (usually a\n\
		parity bit) will be stripped from each character before it\n\
		is written to the photo file; with -P this bit will be\n\
		preserved.\n\
\n\
-D              Turns on various cryptic debugging messages.\n\
\n\
");
#if TURN_OFF_SUPPORTED
		fprintf (stderr, "\
-T              Tells termlink to turn the specified terminal line off in\n\
		/etc/ttys before starting (and turn it back on again when\n\
		exiting).  This option will be ignored if termlink is not\n\
		installed set-uid or if the specified line is not listed\n\
		in /etc/ttys.  If someone is logged in on the specified\n\
		line, termlink will merely print a message and exit.\n\
");
#else
		printNoTurnOff();
#endif
		exit(1);
		break;
	   }
	  if (skip >= ac)
	   { fprintf(stderr,"%s: -%c needs another argument\n",
			     argvec[0],**av);
	     errs = 1;
	   }
	}
     }
    else
     { fprintf(stderr,"%s: extra argument `%s'\n",argvec[0],*av);
     }
  }
 if (errs)
  { exit(1);
  }
 tty_fd = open("/dev/tty",O_RDWR,0);
 if (tty_fd < 0)
  { fprintf(stderr,"%s: cannot open /dev/tty, serious failure: ",argvec[0]);
    perror(0);
    exit(1);
  }
 if (*device == '\0')
  { fprintf(stderr,"%s: no device specified\n",argvec[0]);
    exit(1);
  }
#if TURN_OFF_SUPPORTED
 if (turn_off)
  { old_active = set_line_active(device,0);
  }
#endif
 line_fd = open(device,O_RDWR,0);
 if (line_fd < 0)
  { fprintf(stderr,"%s: cannot open %s: ",argvec[0],device);
    maybe_turnoff();
    perror(0);
    exit(1);
  }
#if TURN_OFF_SUPPORTED
 if (turn_off)
  { int pid;
    int fd;
    while ((pid=fork()) < 0)
     { perror("fork");
       sleep(3);
     }
    if (pid == 0)
     { pid = getppid();
       for (fd=getdtablesize()-1;fd>=0;fd--)
	{ close(fd);
	}
       fd = open("/dev/tty",O_RDWR,0);
       if (fd < 0)
	{ _exit(0);
	}
       ioctl(fd,TIOCNOTTY,0);
       close(fd);
       fd = open(device,O_RDWR,0);
       if (fd < 0)
	{ _exit(0);
	}
       close(fd);
       while (kill(pid,0) == 0)
	{ sleep(15);
	}
       _exit(0);
     }
  }
#endif
 termlineGet (tty_fd, &old_tty);
 new_tty = old_tty;
 termlineRawMode (&new_tty);
 termlineEchoOff (&new_tty);

 termlineGet (line_fd, &old_line);
 new_line = old_line;
 termlineRawMode (&new_line);
 termlineEchoOff (&new_line);

 termlineOutputFlow (tty_fd, /*on=*/1);
 termlineOutputFlow (line_fd, /*on=*/1);
 while (1)
  { setup_tty();
    do_setspeed(speed);
    if (send_init_file)
     { send_init_file = 0;
       if (init_file)
	{ init_conversation();
	}
     }
    while (1)
     { fd_set fdmask;
       int selrv;
       FD_ZERO(&fdmask);
       FD_SET(tty_fd,&fdmask);
       FD_SET(line_fd,&fdmask);
       selrv = select(FD_SETSIZE,&fdmask,(fd_set *)0,(fd_set *)0,(struct timeval *)0);
       if (selrv < 0)
	{ perror("select");
	  continue;
	}
       if (FD_ISSET(tty_fd,&fdmask))
	{ char c;
	  int n;
	  n = read(tty_fd,&c,1);
	  if (n > 0)
	   { if (c == breakc)
	      { break;
	      }
	     write(line_fd,&c,1);
	     if (echo)
	      { write(tty_fd,&c,1);
		if (photofd >= 0)
		 { write(photofd,&c,1);
		 }
	      }
	   }
	}
       if (FD_ISSET(line_fd,&fdmask))
	{ char buf[512];
	  int n;
	  n = read(line_fd,buf,sizeof(buf));
	  if (n > 0)
	   { write(tty_fd,buf,n);
	     if (photofd >= 0)
	      { if (strip_parity)
		 { register int i;
		   register char *cp;
		   for (i=n,cp=buf;i>0;i--,cp++)
		    { *cp &= 0x7f;
		    }
		 }
		write(photofd,buf,n);
	      }
	   }
	}
     }
    reset_tty();
    while (1)
     { tree_match(tree,"Termlink> ",cmd_buf);
       if (strcmp(cmd_buf,"talk") == 0)
	{ break;
	}
       else if (strcmp(cmd_buf,"quit") == 0)
	{
	  termlineSet (line_fd, &old_line);
	  maybe_turnoff();
	  if (dohup_after)
	   { do_hangup();
	   }
	  exit(0);
	}
       else if (strcmp(cmd_buf,"hang-up-phone") == 0)
	{ do_hangup();
	}
       else if (strcmp(cmd_buf,"facsimile") == 0)
	{ if ( (access(kt_file,R_OK) < 0) ||
	       ((fax_fd=open(kt_file,O_RDONLY,0)) < 0) )
	   { printf("Cannot open %s: ",kt_file);
	     fperror(stdout,(char *)0);
	   }
	  else
	   { send_facsimile();
	   }
	}
       else if (strcmp(cmd_buf,"set echo on") == 0)
	{ echo = 1;
	}
       else if (strcmp(cmd_buf,"set echo off") == 0)
	{ echo = 0;
	}
       else if (strcmp(cmd_buf,"set speed") == 0)
	{ if (termlineSpeed ((TermProps*)0, kt_baud) < 0)
	   { printf("Speed %d not available\n",kt_baud);
	   }
	  else
	   { speed = kt_baud;
	   }
	}
       else if (strcmp(cmd_buf,"set facsimile sync blind") == 0)
	{ fax_sync = SYNC_BLIND;
	}
       else if (strcmp(cmd_buf,"set facsimile sync echo") == 0)
	{ fax_sync = SYNC_ECHO;
	}
       else if (strcmp(cmd_buf,"set facsimile echo on") == 0)
	{ fax_echo = 1;
	}
       else if (strcmp(cmd_buf,"set facsimile echo off") == 0)
	{ fax_echo = 0;
	}
       else if (strcmp(cmd_buf,"kermit send") == 0)
	{ int pid;
	  if (access(kt_file,R_OK) < 0)
	   { printf("Cannot read %s: ",kt_file);
	     fperror(stdout,(char *)0);
	   }
	  else
	   { pid = fork();
	     if (pid < 0)
	      { printf("Cannot fork: ");
		fperror(stdout,(char *)0);
	      }
	     else if (pid == 0)
	      { char brc[15];
		close(tty_fd);
		close(line_fd);
		sprintf(brc,"%d",baud_rate);
		setuid(getuid());
		setgid(getgid());
		execlp("kermit","kermit","slb",device,brc,kt_file,0);
		printf("Cannot exec kermit!\n");
		exit(0);
	      }
	     else
	      { while (wait(0) != pid) ;
	      }
	   }
	}
       else if (strcmp(cmd_buf,"kermit receive") == 0)
	{ int pid = fork();
	  if (pid < 0)
	   { printf("Cannot fork: ");
	     fperror(stdout,(char *)0);
	   }
	  else if (pid == 0)
	   { char brc[15];
	     close(tty_fd);
	     close(line_fd);
	     sprintf(brc,"%d",baud_rate);
	     setuid(getuid());
	     setgid(getgid());
	     execlp("kermit","kermit","rlb",device,brc,0);
	     printf("Cannot exec kermit!\n");
	     exit(0);
	   }
	  else
	   { while (wait(0) != pid) ;
	   }
	}
     }
  }
}

send_facsimile()
{
 char buf[512];
 int n;
 int i;

 while (1)
  { n = read(fax_fd,buf,sizeof(buf));
    if (n < 0)
     { printf("Error reading from %s: ",kt_file);
       fperror(stdout,(char *)0);
     }
    for (i=0;i<n;i++)
     { send_f_char(buf[i]);
     }
  }
}

send_f_char(c)
char c;
{
 char ec;

 if (fax_echo)
  { write(tty_fd,&c,1);
  }
 if (nlmap && (c == '\n'))
  { c = '\r';
  }
 write(line_fd,&c,1);
 switch (fax_sync)
  { case SYNC_BLIND:
	{ char ebuf[10];
	  read_timeout(line_fd,ebuf,sizeof(ebuf),0);
	}
       return;
       break;
    case SYNC_ECHO:
       while (1)
	{ switch (read_timeout(line_fd,&ec,1,250)) /* .25 second timeout */
	   { case 0:
		return;
		break;
	     case 1:
		break;
	     case -1:
		fperror(stdout,"echo read");
		return;
		break;
	   }
	  if ((ec == c) || ((c == '\r') && ((ec == '\n') || (ec == '\r'))))
	   { return;
	   }
	}
       break;
  }
}

extern int sys_nerr;
#ifndef Sgi_TARGET
extern const char *const sys_errlist[];
#endif

fperror(f,s)
FILE *f;
char *s;
{
 static char buf[64];

 if (s && *s)
  { fprintf(f,"%s: ",s);
  }
 if (errno < sys_nerr)
  { strcpy (buf, sys_errlist[errno]);
  }
 else
  { static char buf[64];
    sprintf(buf,"Unknown error (%d)",errno);
  }
 fprintf(f,"%s\n",buf);
 fflush(f);
}
