/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1991 by John E. Lloyd.				    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*#include <stdio.h>*/

#include "../h/stringDefs.h"

#include "../h/cdefs.h"
#include "../h/rciTask.h"
#include "../h/robotControl.h"
#include "../h/errorCodes.h"
#include "../h/errorStack.h"
#include "../h/moper_com.h"
#include "../h/robot_kynvar.h"
#include "../h/pmcIO.h"
#include "../h/priDevice.h"
#include "../h/IOdevice.h"
#include "../h/robotMath.h"
#include "../h/rciMonitor.h"
#include "../h/rciMacros.h"
#include "../h/tval.h"
#include "../h/packByte.h"

extern int rciCpu;
extern char *genErrorString();

void onLimitError (IO_DEVICE *dev, JLS *jls, int enc[], int lastenc[])
{
	extern int pmcio_encset0[];   
	extern int pmcio_encset1[];
	extern int pmcio_chan[];
	extern int pmcio_numtrys;
	int j;
	int enccheck[12];
	int encset0[12];
	int encset1[12];
	int mode[12];

	for (j=0; j<12; j++)
	 { enccheck[j] = 0;
	   encset0[j] = pmcio_encset0[j];
	   encset1[j] = pmcio_encset1[j];
	 }
	printf ("%d tries\n", pmcio_numtrys);
	nap (0.01);
	pmcio_get_enc (dev, enccheck);
	pmcio_get_mode (dev, mode);
	printf ("lastenc  enc      encset0  encset1   ch enccheck mode\n");
/*		 XXXXXXXX XXXXXXXX XXXXXXXX XXXXXXXX  XX XXXXXXXX XX */
	for (j=0; j<jls->numJ; j++)
	 { printf ("%8d %8d %8d %8d  %2d %8d %2d\n", lastenc[j], enc[j],
		    encset0[j], encset1[j], pmcio_chan[j], enccheck[j], mode[j]);
	 }
}


void makeName (char *buf, char *fxn, RCI_RBT *rbt)
{
	strcpy (buf, fxn);
	strcat (buf, ", ");
	strcat (buf, rbt->name);
	strcat (buf, " -- ");
}

pmc_input(td, rbt)
RCI_DESC *td;
RCI_RBT *rbt;
{
	char pfx[64];

	makeName (pfx, "rbtInput()", rbt);
	/* If we are running under the simulator, use the generic functions,
	   because the simulator controller emulates an old Unimation
	   controller. We are running under the simulator if the 
	   scheduling discipline is ON_SIGNAL */
	if (td->tcb->scheduleDisp == ON_SIGNAL ||
	    td->tcb->scheduleDisp == ON_CLOCK_SIM)
	 { return (gen_input (td, rbt));
	 }
	else
	 { return (_pmc_input (td, rbt, pfx));
	 }
}

static Tval timeCheck;
extern Tval rciWallClock();

_pmc_input(td, rbt, pfx)
RCI_DESC *td;
RCI_RBT *rbt;
char *pfx;
{
 	register HOW *how = rbt->how;
 	register JLS *jls = rbt->jls;
	RCI_TCB *tcb = td->tcb;
	IO_DEVICE *dev = (IO_DEVICE*)rbt->dev;

	/* ---- read stuff in from the robot and load the 'how' struct ---- */

	 { unsigned char dataBuf[MAX_DATABUF];
	   float ang[MAXJNTS];
	   int   enc[MAXJNTS];
	   int lastenc[MAXJNTS];
	   register unsigned char *bp = dataBuf;
	   register int n, j, k;
	   register float rate;
	   int readFlags;
	   int stateFlags;

	   pmcio_get_enc (dev, enc);

	   /* ---- compute angle and velocity ---- */

	   (*jls->encoderToAngle)(ang, enc, jls);	   

	   how->time = rciTime();

	   if (tcb->cycleCount == 1)
	    { for (j=0; j<jls->numJ; j++)
	       { how->angVel[j] = 0.0;
		 how->encVel[j] = 0.0;
		 how->angSet[j] = ang[j];
		 how->encSet[j] = enc[j];
	       }
	    }
	   else
	    { rate = 1000.0/(float)(tcb->interval);

	      for (j=0; j<jls->numJ; j++)
	       { how->angVel[j] = (ang[j] - how->ang[j]) * rate;
		 how->encVel[j] = (enc[j] - how->enc[j]) * rate;
	       }
	    }

	   if (k = how->loadSettings)
	    { if (how->controlLevel == MOTOR_LEVEL)
	       { for (j=0; j<jls->numJ; j++)
		  { if (k & (1<<j))
		     { how->encSet[j] = enc[j];
		     }
		  }
	       }
	      else
	       { for (j=0; j<jls->numJ; j++)
		  { if (k & (1<<j))
		     { how->angSet[j] = ang[j];
		     }
		  }
	       }
	      how->loadSettings = 0;		    
	    }

	   for (j=0; j<jls->numJ; j++)
	    { lastenc[j] = how->enc[j];
	      how->ang[j] = ang[j];
	      how->enc[j] = enc[j];
	    }

	   if (how->checking & CHK_MAXPOS)
	    { int jnts = 0;

	      for (j=0; j<jls->numJ; j++)
	       { if (ang[j] < jls->jmin[j] || ang[j] > jls->jmax[j])
		  { jnts |= (1<<j);
		  }
	       }
	      if (jnts)
	       { onLimitError (dev, jls, enc, lastenc);
		 rciAbort (EMaxPos,
		"%s + joint limit exceeded for joints 0x%x\n", pfx, jnts);
		 return (-1);
	       }
	    }

	   if (how->checking & CHK_MAXVEL)
	    { int jnts = 0;

	      for (j=0; j<jls->numJ; j++)
	       { if (FABS (how->angVel[j]) > jls->maxVel[j])
		  { jnts |= (1<<j);
		  }
	       }

	      if (jnts)
	       { int j;
		 float cvt;

		 onLimitError (dev, jls, enc, lastenc);
		 rciAbort (EMaxVel,
		"%s + velocity limit exceeded for joints 0x%x\n", pfx, jnts);

		 for (j=0; j<jls->numJ; j++)
		  { if (jnts & (1<<j))
		     { break;
		     }
		  }
		 cvt = (jls->jtype[j] == 'P' ? 1.0 : RADTODEG);
		 rciPrintf ("\nangVel[%d] = %g ang[%d] = %g rate = %g\n",
			j, (double)cvt*how->angVel[j],
			j, (double)cvt*how->ang[j], (double)rate);
		 return (-1);
	       }
	      timeCheck = rciWallClock();
	    }
	   how->state 	    = 0;

	   stateFlags = pmcio_get_state (dev);
	   how->power	    = ((stateFlags & POWER_ON) != 0);
	   how->calibrated  = ((stateFlags & IS_CALIBRATED) != 0);

	   if (stateFlags & MARK_FOUND)
	    { how->markFound = pmcio_get_mark (dev);
	    }

	   if (how->options & TORQUE_AVAIL)
	    { int rawTor[MAXJNTS];

	      pmcio_get_cur (dev, rawTor);
	      for (j=0; j<jls->numJ; j++)
	       { how->mtorque[j] = jls->currentToMtorque[j] * rawTor[j];
	       }

	      if (how->checking & CHK_MAXTOR)
	       { int jnts = 0;
		 for (j=0; j<jls->numJ; j++)
		  { if (FABS (how->mtorque[j]) > jls->maxMtorque[j])
		     { jnts |= (1<<j);
		     }
		  }
		 if (jnts)
		  { rciAbort (EMaxTor,
"%s maximum motor torque exceeded for joints 0x%x\n", pfx, jnts);
		    return (-1);
		  }
	       }
	    }
	   how->potOK = 0;

	   /* Gripper info N/A for pmc */

	   how->hand = 0;
	   how->handForce = 0;

	   /* no force sensor so zero those values: */
	   zeroVf ((float*)how->forceInput, 6);

	   how->jstickOK = 0;
	   how->teleopOK = 0;

	   how->adc = 0;
	   how->slu = 0;
	 }

	return (1);
}

pmc_output(td, rbt)
RCI_DESC *td;
RCI_RBT *rbt;
{
	char pfx[64];

	makeName (pfx, "rbtOutput()", rbt);
	/* See comment in pmc_input() */
	if (td->tcb->scheduleDisp == ON_SIGNAL ||
	    td->tcb->scheduleDisp == ON_CLOCK_SIM)
	 { return (gen_output (td, rbt));
	 }
	else
	 { return (_pmc_output (td, rbt, pfx));
	 }
}

_pmc_output(td, rbt, pfx)
RCI_DESC *td;
RCI_RBT *rbt;
char *pfx;
{
	RCI_TCB *tcb = td->tcb;
	IO_DEVICE *dev = (IO_DEVICE*)rbt->dev;
	HOW *how = rbt->how;
	JLS *jls = rbt->jls;
	register CHG *chg = rbt->chg;
	unsigned long jntsMask = ((1<<jls->numJ)-1);
	int j;

	/* ---- do the control commands first ---- */

	if (tcb->newRate)
	 { pmcio_set_rate (dev, MIN(tcb->interval,255));
	 }
	   
	if (chg->ctrlcom)
  	 { if (chg->paramset)
	    { rciAbort (ENotSupported,
	"%s setting servo parameters not supported for this robot\n", pfx);
	      return (-1);
	    }

	   if (chg->integration.com)
	    { int mask = chg->integration.val.l & jntsMask;
	      how->integration = mask;
	      pmcio_set_integration (dev, mask);
	      chg->integration.com = 0;
	    }

	   if (chg->calibMode.com)
	    { int mask = chg->calibMode.val.l & jntsMask;
	      how->calibMode = mask;
	      pmcio_set_calibMode (dev, mask);
	      chg->calibMode.com = 0;
	    }

	   if (chg->calibflag.com)
	    { rciAbort (ENotSupported,
	"%s SET_CALIBFLAG() not supported for this robot\n", pfx);
	      return (-1);
	    }

	   if (chg->power.com)
	    { pmcio_set_power (dev, chg->power.val.i);
	      chg->power.com = 0;
	    }
	    
	   if (chg->maintCom.com)
	    { rciAbort (ENotSupported,
	"%s SET_MAINT_COM() not supported for this robot\n", pfx);
	      return (-1);
	    }
	    
	   if (chg->checking.com)
	    { how->checking = chg->checking.val.l & CHK_RBTALL;
	      chg->checking.com = 0;
	    }

	   if (chg->powerOnRelease.com)
	    { how->powerOnRelease = !!chg->powerOnRelease.val.i;
	      chg->powerOnRelease.com = 0;
	    }
	    
	   chg->ctrlcom = 0;
	 }


	/* ----- now do the motion control stuff ----- */

	 { unsigned long newTorqueMode;
	   unsigned long modeChange;
	   float deltaT;
	   register int j;
	   char *illCmd;
	
	   deltaT = tcb->interval / 1000.;

	   /* check for a change in the control level */

	   if (chg->controlLevel.com)
	    { how->controlLevel = chg->controlLevel.val.i;
	      chg->controlLevel.com = 0;
	    }

	   /* look for the stop command first */

	   if (chg->stop.com)
	    { how->torqueMode = 0;
	      how->loadSettings = jntsMask;

	      for (j=0; j<jls->numJ; j++)
	       { chg->motion[j].com = 0;
		 how->mtorqueSet[j] = 0.0;
		 how->jtorqueSet[j] = 0.0;
		 pmcio_set_stop (dev, j);
	       }
	      chg->stop.com = 0;
	    }

	   if (how->controlLevel == JOINT_LEVEL)
	    {
	      int newEnc[MAXJNTS];
	      float newMtorque[MAXJNTS];
	      char com[MAXJNTS];

	      newTorqueMode = how->torqueMode;

	      for (j=0; j<jls->numJ; j++)
	       { switch (com[j] = chg->motion[j].com)
		  { case 0:
		     { break;
		     }
		    case CHG_ANGLE:
		     { float ang = chg->motion[j].val.f;
   
		       if (how->checking & CHK_REQPOS)
			{ if (ang < jls->jmin[j] || ang > jls->jmax[j])
			   { rciAbort (EReqPos,
	"%s requested angle %g exceeds limits, joint %d\n",pfx,
			     ang*RADTODEG, j+1);
			     return (-1);
			   }
			}
		       if (how->checking & CHK_REQVEL)
			{ if (FABS(ang - how->angSet[j]) > jls->maxVel[j]*deltaT)
			   { rciAbort (EReqVel,
	"%s requested velocity %g exceeds velocity limits, joint %d\n", pfx,
			     RADTODEG*FABS(ang - how->angSet[j])/deltaT, j+1);
			     return (-1);
			   }
			}
		       how->angSet[j] = ang;
		       how->jtorqueSet[j] = 0.0;
		       newTorqueMode &= ~(1<<j);
		       break;
		     }
		    case CHG_JTORQUE:
		     {
		       how->jtorqueSet[j] = chg->motion[j].val.f;
		       newTorqueMode |= (1<<j);
		       break;
		     }
		    case CHG_STOP:
		     {
		       how->jtorqueSet[j] = 0.0;
		       newTorqueMode &= ~(1<<j);
		       break;
		     }
		    case CHG_ANGLE_VALUE:
		     {
		       rciAbort (EBadCmd,
	"%s SET_ANGLE_VALUE() not supported for this robot\n", pfx);
		       return (-1);
		       break;
		     }
		    case CHG_ENCODER:
		     { illCmd = "SET ENCODER";
		       break;
		     }
		    case CHG_ENCODER_VALUE:
		     { illCmd = "SET ENCODER VALUE";
		       break;
		     }
		    case CHG_MTORQUE:
		     { illCmd = "SET MTORQUE";
		       rciAbort (EBadCmd,
     "%s illegal joint level motion command %s\n", pfx, illCmd);
		       return (-1);
		       break;
		     }
		    default:
		     { rciAbort (EBadCmd,
     "%s illegal joint level motion command (%d)\n", pfx, com[j]);
		       return (-1);
		       break;
		     }
		  }
		 chg->motion[j].com = 0;

		 switch (chg->bias[j].com)
		  { case 0:
		     { break;
		     }
		    case CHG_JTORQUE_BIAS:
		     { rciAbort (EBadCmd,
	"%s SET_JTORQUE_BIAS() not supported for this robot\n", pfx);
		       return (-1);
		       break;
		     }
		    case CHG_MTORQUE_BIAS:
		     { rciAbort (EBadCmd,
	"%s SET_MTORQUE_BIAS() not supported for this robot\n", pfx);
		       return (-1);
		       break;
		     }
		    default:
		     { rciAbort (EBadCmd,
	"%s illegal joint level bias command (%d)\n", pfx, chg->bias[j].com);
		       return (-1);
		       break;
		     }
		  }		    
		 chg->bias[j].com = 0;
	       }

	      modeChange = newTorqueMode ^ how->torqueMode;
	      how->torqueMode = newTorqueMode;
			  
	      (*jls->angleToEncoder)(newEnc, how->angSet, jls);
	      if (how->torqueMode)
	       { (*jls->jtorqueToMtorque)(newMtorque, how->jtorqueSet, jls);
	       }
	      
	      /* now go and dispatch the commands */
   
	      for (j=0; j<jls->numJ; j++)
	       { if (how->torqueMode & (1<<j))
		  { float mtor = newMtorque[j];
		    if (mtor != how->mtorqueSet[j] || modeChange & (1<<j))
		     { if (how->checking & CHK_REQTOR)
			{ if (FABS(mtor) > jls->maxMtorque[j])
			   { rciAbort (EReqTor,
      "%s requested motor torque %f exceeds bounds, joint %d\n",
				       pfx,(double)mtor,j+1);
			     return (-1);
			   }
			}
		       pmcio_set_cur (dev, mtor*jls->mtorqueToCurrent[j]
					    + jls->currentBias[j], j);
		     }
		    how->encSet[j] = newEnc[j];
		    how->mtorqueSet[j] = mtor;
		  }
		 else
		  { if (newEnc[j] != how->enc[j] || modeChange & (1<<j))
		     { pmcio_set_enc (dev, newEnc[j], j);
		     }
		    how->encSet[j] = newEnc[j];
		    how->mtorqueSet[j] = how->torqueMode ? newMtorque[j]:0.0;
		  }
	       }
	      how->loadSettings |= newTorqueMode;
	    }
	   else 	/* control level == MOTOR LEVEL */
	    { int newEnc[MAXJNTS];
	      float newAng[MAXJNTS];
	      float newMtorque[MAXJNTS];
	      int encValueSet = NO;
	      char com[MAXJNTS];
   
	      newTorqueMode = how->torqueMode;

	      for (j=0; j<jls->numJ; j++)
	       { switch (com[j] = chg->motion[j].com)
		  { case 0:
		     { newEnc[j] = how->encSet[j];
		       newMtorque[j] = how->mtorqueSet[j];
		       break;
		     }
		    case CHG_ENCODER:
		     { newEnc[j] = chg->motion[j].val.i;
		       newMtorque[j] = 0.0;
		       newTorqueMode &= ~(1<<j);
		       break;
		     }
		    case CHG_MTORQUE:
		     { newEnc[j] = how->encSet[j];
		       newMtorque[j] = chg->motion[j].val.f;
		       newTorqueMode |= (1<<j);
		       break;
		     }
		    case CHG_STOP:
		     { newEnc[j] = how->encSet[j];
		       newMtorque[j] = 0.0;
		       newTorqueMode &= ~(1<<j);
		       break;
		     }
		    case CHG_ENCODER_VALUE:
		     { newEnc[j] = chg->motion[j].val.i;
		       newMtorque[j] = 0.0;
		       newTorqueMode &= ~(1<<j);

		       how->enc[j] = chg->motion[j].val.i;
		       how->encSet[j] = chg->motion[j].val.i;
		       encValueSet = YES;
		       break;
		     }
		    case CHG_ANGLE:
		     { illCmd = "SET ANGLE";
		       break;
		     }
		    case CHG_ANGLE_VALUE:
		     { illCmd = "SET ANGLE VALUE";
		       break;
		     }
		    case CHG_JTORQUE:
		     { illCmd = "SET JTORQUE";
		       rciAbort (EBadCmd,
     "%s illegal motor level motion command %s\n", pfx, illCmd);
		       return (-1);
		       break;
		     }
		    default:
		     { rciAbort (EBadCmd,
	"%s illegal motor level motion command (%d)\n", pfx, com[j]);
		       return (-1);
		       break;
		     }
		  }

		 switch (chg->bias[j].com)
		  { case 0:
		     { break;
		     }
		    case CHG_MTORQUE_BIAS:
		     { rciAbort (ENotSupported,
	"%s SET_MTORQUE_BIAS() not supported for this robot\n", pfx);
		       return (-1);
		     }
		    case CHG_JTORQUE_BIAS:
		     { rciAbort (ENotSupported,
	"%s SET_JTORQUE_BIAS() not supported for this robot\n", pfx);
		       return (-1);
		     }
		    default:
		     { rciAbort (EBadCmd,
	"%s illegal motor level bias command (%d)\n", pfx, chg->bias[j].com);
		       return (-1);
		       break;
		     }
		  }		    
		 chg->bias[j].com = 0;
	       }

	      if (encValueSet)
	       { (*jls->encoderToAngle)(how->ang, how->enc, jls);
		 (*jls->encoderToAngle)(how->angSet, how->encSet, jls);
	       }
	      modeChange = newTorqueMode ^ how->torqueMode;
	      how->torqueMode = newTorqueMode;
   
	      (*jls->encoderToAngle)(newAng, newEnc, jls);
	      if (how->torqueMode)
	       { (*jls->mtorqueToJtorque)(how->jtorqueSet, newMtorque, jls);
	       }
	      else
	       { for (j=0; j<jls->numJ; j++)
		  { how->jtorqueSet[j] = 0.0;
		  }
	       }

	      for (j=0; j<jls->numJ; j++)
	       { 		 
		 if (how->checking & CHK_REQPOS)
		  { if (newAng[j] < jls->jmin[j] || newAng[j] > jls->jmax[j])
		     { rciAbort (EReqPos,
	"%s requested angle %g exceeds limits, joint %d\n",pfx,
		       RADTODEG*newAng[j],j+1);
		       return (-1);
		     }
		  }
		 if (how->checking & CHK_REQVEL)
		  { if (FABS(newAng[j] - how->angSet[j]) > jls->maxVel[j]*deltaT)
		     { rciAbort (EReqVel,
	"%s requested velocity %g exceeds limits, joint %d\n", pfx,
		       RADTODEG*FABS(newAng[j] - how->angSet[j])/deltaT, j+1);
		       return (-1);
		     }
		  }

		 if (how->torqueMode & (1<<j))
		  { float mtor = newMtorque[j];
		    if (mtor != how->mtorqueSet[j] || modeChange&(1<<j))
		     { if (how->checking & CHK_REQTOR)
			{ if (FABS(mtor) > jls->maxMtorque[j])
			   { rciAbort (EReqTor,
  "%s requested motor torque %f exceeds bounds, joint %d\n",
				       pfx,(double)mtor,j+1);
			     return (-1);
			   }
			}
		       pmcio_set_cur (dev, mtor*jls->mtorqueToCurrent[j] +
					    jls->currentBias[j], j);
		     }
		  }
		 else
		  { if (com[j] == CHG_ENCODER_VALUE)
		     { pmcio_set_encvalue (dev, chg->motion[j].val.i, j);
		       how->calibMode &= ~(1<<j);
		     }
		    else if (com[j] == CHG_STOP)
		     { how->loadSettings |= (1<<j);
		       pmcio_set_stop (dev, j);
		     }
		    else if (newEnc[j] != how->enc[j] || modeChange&(1<<j))
		     { pmcio_set_enc (dev, newEnc[j], j);
		     }
		  }
		 how->mtorqueSet[j] = newMtorque[j];
		 how->encSet[j] = newEnc[j];
		 how->angSet[j] = newAng[j];
		 chg->motion[j].com = 0;
	       }
	      how->loadSettings |= how->torqueMode;
	    }
#if 1
	   if (chg->hand.com)
	    { rciAbort (EBadCmd, 
	"%s SET_HAND() and SET_HAND_FORCE() not supported for this robot\n", 
			pfx, chg->hand.com);
	      return (-1);
	    }
#endif
	 }
       
	if (chg->readcom)
 	 {
	   if (chg->torque.com)
	    { chg->torque.com = 0;	/* torque read enabled by default */
	    }
	   if (chg->pot.com)
	    { rciAbort (ENotSupported,
	"%s SET_POT_READ() not supported for this robot\n", pfx);
	      return (-1);
	    }
	   if (chg->pendant.com)
	    { rciAbort (ENotSupported,
	"%s SET_PENDANT_READ() not supported for this robot\n", pfx);
	      return (-1);
	    }
	   if (chg->force.com)
	    { rciAbort (ENotSupported,
	"%s SET_FORCE_READ() not supported for this robot\n", pfx);
	      return (-1);
	    }
	   if (chg->teleop.com)
	    { rciAbort (ENotSupported,
	"%s SET_TELEOP_READ() not supported for this robot\n", pfx);
	      return (-1);
	    }
	   if (chg->jstick.com)
	    { rciAbort (ENotSupported,
	"%s SET_JSTICK_READ() not supported for this robot\n", pfx);
	      return (-1);
	    }
	   if (chg->adc.com)
	    { rciAbort (ENotSupported,
	"%s SET_ADC_CHAN() not supported for this robot\n", pfx);
	      return (-1);
	    }
	   if (chg->slu.com)
	    { rciAbort (ENotSupported,
	"%s SET_SLU_CHAN() not supported for this robot\n", pfx);
	      return (-1);
	    }
	   if (chg->setTest.com)
	    { rciAbort (ENotSupported,
	"%s SET_TEST() not supported for this robot\n", pfx);
	      return (-1);
	    }
	   chg->readcom = 0;
	 }

	if (chg->writecom)
  	 { int n;

	   for (n=0; n<how->numPioOut; n++)
	    { int cmd;
	      if (cmd = chg->pioOutput[n].com)
	       { rciAbort (ENotSupported,
	"%s PUT,SET,CLEAR_PIO() not supported for this robot\n", pfx);
		 return (-1);
	       }
	    }
	   for (n=0; n<how->numDac; n++)
	    { if (chg->dacOutput[n].com)
	       { rciAbort (ENotSupported,
	"%s PUT_DAC() not supported for this robot\n", pfx);
		 return (-1);
	       }
	    }
	   for (n=0; n<how->numSlu; n++)
	    { if (chg->sluOutput[n].com)
	       { rciAbort (ENotSupported,
	"%s PUT_SLU() not supported for this robot\n", pfx);
		 return (-1);
	       }
	    }
	   if (chg->pendantMsg.com)
	    { rciAbort (ENotSupported,
	"%s PUT_PENDANT_MSG() not supported for this robot\n", pfx);
	      return (-1);
	    }
	   if (chg->putTest.com)
	    { rciAbort (ENotSupported,
	"%s PUT_TEST() not supported for this robot\n", pfx);
	      return (-1);
	    }
	   chg->writecom = 0;
	 }

	if (chg->putVectorSet)
	 { rciAbort (ENotSupported,
	"%s PUT_VECTOR() not supported for this robot\n", pfx);
	   return (-1);
	 }
	return (1);
}

#define WAKEUP_SIG	1
#define RELEASE_SIG	2

pmc_signal (dev)
IO_DEVICE *dev;
{
	return (1);
}

pmc_startup(td, rbt)
RCI_DESC *td;
RCI_RBT *rbt;
{
	char pfx[64];

	makeName (pfx, "rbtStartup()", rbt);
	/* See comment in pmc_input() */
	if (td->tcb->scheduleDisp == ON_SIGNAL ||
	    td->tcb->scheduleDisp == ON_CLOCK_SIM)
	 { return (gen_startup (td, rbt));
	 }
	else
	 { return (_pmc_startup (td, rbt, pfx));
	 }
}

_pmc_startup (td, rbt, pfx)
RCI_DESC *td;
RCI_RBT *rbt;
char *pfx;
{
	int j, n;
	RCI_TCB *tcb = td->tcb;
	IO_DEVICE *dev = (IO_DEVICE*)rbt->dev;
 	register HOW *how = rbt->how;
	JLS *jls = rbt->jls;
	CHG *chg = rbt->chg;

	if (pmcio_attach (dev, jls->numJ) < 0)
	 { errorMsg (getErrorCode(0), "%s cannot connect to robot\n",pfx);
	   rciAbortErrors();
	   return (-1);
	 }

	/* set any necessary commands */

	if (chg->checking.com)
	 { how->checking = chg->checking.val.l & CHK_RBTALL;
	   chg->checking.com = 0;
	 }
	if (chg->powerOnRelease.com)
	 { how->powerOnRelease = !!chg->powerOnRelease.val.i;
	   chg->powerOnRelease.com = 0;
	 }

	how->robotID = 0;
	how->controllerID = 0;
	how->options = pmcio_get_options (dev);

	how->numAdc = 0;
	how->numDac = 0;
	how->numPioIn = 0;
	how->numPioOut = 0;
	how->numSlu = 0;
	how->integration = pmcio_get_integration (dev);
	how->calibMode = pmcio_get_calibMode (dev);
	how->markFound = pmcio_get_mark (dev);

	for (j=0; j<jls->numJ; j++)
	 { how->jtorqueBias[j] = 0;
	   how->mtorqueBias[j] = 0;
	 }
	pmcio_set_rate (dev, MIN(tcb->interval,255));

	rbt->ready = dev->testRead;
	rbt->readyArg = (int)dev;

	rbt->wakeup = pmc_signal;
	rbt->wakeupArg = (int)dev;

	return (1);
}

pmc_release (td, rbt, code)
RCI_DESC *td;
RCI_RBT *rbt;
int code;
{
	char pfx[64];

	makeName (pfx, "rbtRelease()", rbt);
	/* See comment in pmc_input() */
	if (td->tcb->scheduleDisp == ON_SIGNAL ||
	    td->tcb->scheduleDisp == ON_CLOCK_SIM)
	 { return (gen_release (td, rbt, code));
	 }
	else
	 { return (_pmc_release (td, rbt, code, pfx));
	 }
}

_pmc_release (td, rbt, code, pfx)
RCI_DESC *td;
RCI_RBT *rbt;
int code;
char *pfx;
{
	/* Send the release signal to the robot.
	   Send down release commands.
	   Read back information, if any.
	*/

	RCI_TCB *tcb = td->tcb;
	IO_DEVICE *dev = (IO_DEVICE*)rbt->dev;
	HOW *how = rbt->how;
	int j;

	/* clear how->powerOnRelease if code is not normal release request */
	
	if (code == RBT_HARDERROR)
	 { how->powerOnRelease = NO;
	 }
	for (j=0; j<rbt->jls->numJ; j++)
	 { pmcio_set_stop (dev, j);
	 }
	if (how->powerOnRelease)
	 { pmcio_set_power (dev, 0);
	 }
	BZERO ((char*)rbt->chg, sizeof(CHG));
	pmcio_detach (dev);
	return (0);
}
