/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1996 by John E. Lloyd.			    	    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#include <stdio.h>
#include <ctype.h>
#include <rccl.h>
#include <ctree.h>
#include <cdefs.h>        
#include "../h/signalFxns.h"
#include "../h/stringDefs.h"
#include "../h/fatalMsg.h"

#include "forceCal.h"

#define SGN(x)	((x) >= 0 ? 1 : -1)

#define SEC_TO_TICKS    (1000000/64.)

typedef struct _pair {
	struct _pair *next;
	char *name;
	double dist;
	double mass;
} MassDistPair;

MassDistPair *mdList;

MassDistPair *findMassDistPair (name)
char *name;
{
	MassDistPair *pair;

	for (pair=mdList; pair; pair=pair->next)
	 { if (strcmp (pair->name, name) == 0)
	    { return (pair);
	    }
	 }
	return (NULL);
}

int addMassDistPair (name, mass, dist)
char *name;
double mass, dist;
{
	MassDistPair *pair;
	int newpair = 0;

	if (!isalpha(name[0]))
	 { printf ("Error, name must begin with an alphabetic character\n");
	   return (-1);
	 }

	if ((pair = findMassDistPair (name)) == NULL)
	 { pair = (MassDistPair*)malloc(sizeof(MassDistPair));
	   pair->next = mdList;
	   pair->name = copyofstr (name);
	   mdList = pair;
	   newpair = 1;
	 }
	pair->dist = dist;
	pair->mass = mass;
	return (newpair);
}

/* x,y,x and roll,pitch,yaw terms defining the transformation ST
   from T6 to sensor coordinates */

double TSx = 0.0;
double TSy = 0.0;
double TSz = 101.4;
double TSroll = 0.0;
double TSpitch = 0.0;
double TSyaw = 0.0;

/* x,y,x and roll,pitch,yaw terms defining the transformation JT
   from sensor coordinates to jig coordinates */

double SJx = 0.0;
double SJy = 0.0;
double SJz = 14.08;
double SJroll = 46.0;
double SJpitch = 0.0;
double SJyaw = 0.0;

/* Diameter of pin holding weight in place */

#define PIN_DIAMETER	4.66 /* mm */

double d1Offset = 0.0;
double d2Offset = 0.0;
double d1Slope = 1.0;
double d2Slope = 1.0;

#define NSAMP	10	/* number od samples to collect */

int dataCount = 0;

writeDataFile (fp, data, force, m1, d1, m2, d2, ang)
FILE *fp;
float *data;
FORCE *force;
double m1, d1, m2, d2;
float *ang;
{
	fprintf (fp, "%% m1=%6.3f d1=%7.2f m2=%6.3f d2=%7.2f (kg,mm)\n",
			 m1, d1, m2, d2);
	fprintJf (fp, "%% angles = %8.3f\n", ang, 6);

	if (dataCount == 0)
	 { fprintf (fp, "        f = [");
	   fprintVf (fp, "%9.4f", (float*)force, 6);
	   fprintf (fp, "]';\n");
	   fprintf (fp, "        d = [");
	   fprintVf (fp, "%9.6f", data, 6);
	   fprintf (fp, "]';\n");
	   fflush (fp);
	 }
	else
	 { fprintf (fp, "    f = [f, [");
	   fprintVf (fp, "%9.4f", (float*)force, 6);
	   fprintf (fp, "]'];\n");
	   fprintf (fp, "    d = [d, [");
	   fprintVf (fp, "%9.6f", data, 6);
	   fprintf (fp, "]'];\n");
	   fflush (fp);
	 }
	dataCount++;
}

getRealT6 (mnp,T)
MANIP *mnp;
TRSF *T;
{
	readMem (mnp->t6o, T);
}

getRealJ6 (mnp, j6)
MANIP *mnp;
JNTS *j6;
{
	readMem (mnp->j6o, j6);
}

getForceData (data, fd)
float *data;
ForceData *fd;
{
	ForceData buf;

	readMem (fd, &buf);
	BCOPY (buf.data, data, sizeof(float)*6);
	return (buf.valid);
}

char **argvec;

#define DEFAULT_OUTFILE "forceCal.out"

char key_buf[256];

extern forceMonitor();

int interruptHit = 0;
int useRawData = 1;

SIGHAN_RETURN onIntr(n)
int n;
{
	interruptHit++;
}

parseMassDist (fp, massp, distp)
FILE *fp;
double *massp, *distp;
{
	extern double strtod();

	MassDistPair *pair;

	char sbuf[256];

	fscanf (fp, "%s", sbuf);
	if (isalpha(sbuf[0]))
	 { if ((pair = findMassDistPair(sbuf)) == NULL)
	    { printf ("Error, massDistPair %s not found\n", sbuf);
	      return (-1);
	    }
	   else
	    { *massp = pair->mass;
	      *distp = pair->dist;
	    }
	 }
	else
	 { char *checkp;
	   *massp = strtod (sbuf, &checkp);
	   if (sbuf == checkp)
	    { printf ("Error, error parsing numeric mass value\n");
	      return (-1);
	    }
	   if (fscanf (fp, "%lf", distp) != 1)
	    { printf ("Error, error parsing numeric dist value\n");
	      return (-1);
	    }
	 }
	return (1);
}

main(argc, argv)
int argc;
char **argv;
{
	/* Ctree matcher variables */
	char *keytree;
	char sbuf[256];
	float fval, fval1, fval2;
	float ival;
	int joint;
	double mass1, d1;
	double mass2, d2;
	
	/* General program variables */
	char *robotName = NULL;
	MANIP *mnp;
	POS *pos;
	TRSF *SensorToJig;
	TRSF *T6toSensor;
	TRSF *Target;
	static char outFileName[256];
	FILE *ofp = NULL;
	float forceBias[6];
	ForceData *fd;
	int dataCount;

	int status;
	int count;

	argvec = argv;
	argc--; argv++;
	while (argc)
	 { if (ARGUMENTS ("-sim="))
	    { char *sname;
	      if (strlen(sname=charIndex(*argv,'=')+1) == 0)
	       { goto usage;
	       }
	      rcclSetSimulator(sname);
	    }
	   else if (ARGUMENT ("-sim"))
	    { rcclSetOptions (RCCL_SIMULATE);
	    }
	   else if (ARGUMENT ("-on"))
	    { rcclSetOptions (RCCL_LEAVE_POWER_ON);
	    }
	   else if (ARGUMENT ("-f"))
	    { argc--; argv++;
	      strcpy (outFileName,*argv);
	    }
	   else if (ARGUMENT ("-cal"))
	    { useRawData = 0;
	    }
	   else if (!robotName)
	    { robotName = *argv;
	    }
	   else
	    { goto usage;
	    }
	   argc--; argv++;
	   continue;
usage:
	   usageMsg (-1, 
"%s [<robotName>] [-sim[=<simulatorAddr>]] [-f <filename>] [-on] [-cal]\n", 
		     *argvec);
	 }
	if (!robotName)
	 { if ((robotName = (char*)getDefaultRobot()) == NULL)
	    { printErrors();
	      exit (-1);
	    }
	   printf("robotName = %s\n", robotName);
	 }
	if (outFileName[0] != '\0')
	 { if ((ofp = fopen (outFileName, "w")) == NULL)
	    { fatalMsg (-1, "Can't open file %s\n", outFileName);
	    }
	   printf ("Calibration data will be written to %s\n", outFileName);
	 }

	keytree = tree_match_parse("					\
(	quit								\
	set (								\
	  [T6toSensor SensorToJig] (translation roll-pitch-yaw)		\
	  [d1Offset d2Offset] (%f", &fval, ", \"offset value (mm)\")	\
	  [d1Slope d2Slope] (%f", &fval, ", \"offset value (mm)\")	\
	  massDistPair (%s", sbuf, ",	\"pair name\"			\
		(%f", &fval1, ",	\"mass (kg)\"			\
		  (%f", &fval2, ",	\"dist (kg)\")))		\
	)								\
	move 								\
	 (to 								\
	   (position (%s", sbuf, ",	\"position name\"))		\
	  joint 							\
	   (%d",&joint,",1,6,           \"joint number (1 to 6)\"       \
	     ([to by] (%f", &fval, ",	\"joint value (degrees)\")))	\
	  [x y z] 							\
	    (by								\
	      (%f", &fval, ",    \"displacement (mm)\"))     		\
	 )								\
	open								\
	 (output-file							\
	   (%s", sbuf, ",     \"file name\"))				\
	close (output-file)						\
	show (position force transforms continuous-force (raw biased)	\
	      dSlopeAndOffsets massDistPairs)				\
	enter								\
	 (data 								\
	   ([x-axis y-axis]))						\
	parseMassDist							\
	reset (bias)							\
)",0);

	if (keytree == 0)
	 { printf ("ERROR/Parse error at %.30s\n", tree_match_err_at());
	   exit (-1);
	 }

	/* Set default values for the ST and JT transformations: */

	T6toSensor = allocTrans (NULL, UNDEF);
	SensorToJig = allocTrans (NULL, UNDEF);
	Target = allocTrans (NULL, UNDEF);

	pos = makePosition (NULL, T6, T6toSensor, SensorToJig, EQ, 
				  Target, TL, SensorToJig);

	xyzToTrsf (T6toSensor, TSx, TSy, TSz);
	rpyToTrsf (T6toSensor, TSroll, TSpitch, TSyaw);
	xyzToTrsf (SensorToJig, SJx, SJy, SJz);
	rpyToTrsf (SensorToJig, SJroll, SJpitch, SJyaw);

	if ((mnp = rcclCreate (robotName, 0)) == NULL)
	 { printErrors();
	   exit (1);
	 }
	if (rcclStart() < 0)
	 { printErrors();
	   exit (1);
	 }
	if (useRawData)
	 { SET_FORCE_READ (mnp->rbt, 2); /* the 2 gets us uncalibrated vals */
	 }
	else
	 { SET_FORCE_READ (mnp->rbt, 1); /* the 2 gets us uncalibrated vals */
	 }
	fd = (ForceData*)allocMem ("ForceData", sizeof(ForceData), UNDEF);
	runMonitorFxn (mnp, forceMonitor, /*arg=*/0);
	rcclSetModes (mnp, T6O_EVAL);

	/* wait a couple of cycles for the force reading to
	   initialize */
	rcclBlock();
	rcclBlock();
	rcclBlock();
	if (!getForceData (forceBias, fd))
	 { fprintf (stderr, "Warning: force data not valid when bias set\n");
	 }

	do
 	 { FILE *tfp;

	   tree_match (keytree, "FORCECAL> ", key_buf);

	   tfp = tree_match_input_stream();

	   if (COMMAND("set T6toSensor translation"))
	    { float x, y, z;

	      printf ("Enter x,y,z > ");
	      if (fscanf (tfp, "%f%f%f", &x, &y, &z) != 3)
	       { fprintf (stderr, "Error scanning values\n");
		 while (getc(tfp) != '\n')
		    ;
		 continue;
	       }
	      xyzToTrsf (T6toSensor, x, y, z);
	      TSx = x;
	      TSy = y;
	      TSz = z;
	    }
	   else if (COMMAND("set T6toSensor roll-pitch-yaw"))
	    { float roll, pitch, yaw;

	      printf ("Enter roll,pitch,yaw (in degrees) > ");
	      if (fscanf (tfp, "%f%f%f", &roll, &pitch, &yaw) != 3)
	       { fprintf (stderr, "Error scanning values\n");
		 while (getc(tfp) != '\n')
		    ;
		 continue;
	       }
	      rpyToTrsf (T6toSensor, roll, pitch, yaw);
	      TSroll = roll;
	      TSpitch = pitch;
	      TSyaw = yaw;
	    }
	   else if (COMMAND("set SensorToJig translation"))
	    { float x, y, z;

	      printf ("Enter x,y,z > ");
	      if (fscanf (tfp, "%f%f%f", &x, &y, &z) != 3)
	       { fprintf (stderr, "Error scanning values\n");
		 while (getc(tfp) != '\n')
		    ;
		 continue;
	       }
	      xyzToTrsf (SensorToJig, x, y, z);
	      SJx = x;
	      SJy = y;
	      SJz = z;
	    }
	   else if (COMMAND("set SensorToJig roll-pitch-yaw"))
	    { float roll, pitch, yaw;

	      printf ("Enter roll,pitch,yaw (in degrees) > ");
	      if (fscanf (tfp, "%f%f%f", &roll, &pitch, &yaw) != 3)
	       { fprintf (stderr, "Error scanning values\n");
		 while (getc(tfp) != '\n')
		    ;
		 continue;
	       }
	      rpyToTrsf (SensorToJig, roll, pitch, yaw);
	      SJroll = roll;
	      SJpitch = pitch;
	      SJyaw = yaw;
	    }
	   else if (COMMAND("set d1Offset"))
	    { d1Offset = fval;
	    }
	   else if (COMMAND("set d2Offset"))
	    { d2Offset = fval;
	    }
	   else if (COMMAND("set d1Slope"))
	    { d1Slope = fval;
	    }
	   else if (COMMAND("set d2Slope"))
	    { d2Slope = fval;
	    }
	   else if (COMMAND("set massDistPair"))
	    { int status;

	      if ((status = addMassDistPair (sbuf, fval1, fval2)) == 1)
	       { printf ("new pair added\n");
	       }
	      else if (status == 0)
	       { printf ("old pair overwritten\n");
	       }
	    }
	   else if (COMMAND("move to position"))
	    { int status;
	      JNTS jnts;
	
	      status = getRobotPosition (jnts.v, sbuf, robotName);
	      if (status == 0)
	       { fprintf (stderr, "Position not found for robot %s\n",
				   robotName);	
		 continue;
	       }
	      else if (status < 0)
	       { printErrors();
		 clearErrors();
		 continue;
	       }
	      if (movej (mnp, &jnts) < 0)
	       { printErrors();
		 clearErrors();
	       }
	      else
	       { waitForCompleted (mnp);
	       }
	    }

	   else if (COMMANDS ("move joint"))
	    { JNTS jnts;

	      jnts = *mnp->j6;
	      if (COMMAND ("move joint by"))
	       { jnts.v[joint-1] += DEGTORAD*fval;
printf ("move joint %d by %f\n", joint, fval);
	       }
	      else if (COMMAND ("move joint to"))
	       { jnts.v[joint-1] = DEGTORAD*fval;
printf ("move joint %d to %f\n", joint, fval);
	       }
	      if (movej (mnp, &jnts) < 0)
	       { printErrors();
		 clearErrors();
	       }
	      else
	       { waitForCompleted (mnp);
	       }
	    }
	   else if (COMMANDS ("move"))
	    { TRSF Delta;

	      solveTrans (Target, pos, Target, mnp->t6);
	      identTrsf (&Delta);
	      if (COMMAND ("move x by"))
	       { Delta.p.x = fval;
	       }
	      else if (COMMAND ("move y by"))
	       { Delta.p.y = fval;
	       }
	      else if (COMMAND ("move z by"))
	       { Delta.p.z = fval;
	       }
	      multTrsf (Target, Target, &Delta);
	      if (move (mnp, pos) < 0)
	       { printErrors();
		 clearErrors();
	       }
	      else
	       { waitForCompleted (mnp);
	       }
	    }
	   else if (COMMANDS ("open output-file"))
	    { if (ofp)
	       { printf ("Closing %s\n", outFileName);
		 fprintf (ofp, "c = f / d\n");
		 fclose (ofp);
	       }
	      if ((ofp = fopen (sbuf, "w")) == NULL)
	       { perror ("fopen");
	       }
	      else
	       { strcpy (outFileName, sbuf);
	       }
	      dataCount = 0;
	    }
	   else if (COMMANDS ("close output-file"))
	    { if (ofp)
	       { printf ("Closing %s\n", outFileName);
		 fprintf (ofp, "c = f / d\n");
		 fclose (ofp);
	       }
	      dataCount = 0;
	    }
	   else if (COMMANDS ("show force"))
	    { float data[6];

	      if (!getForceData(data, fd))
	       { printf ("Warning: force data not valid\n");
	       }
	      printVf ("Raw:    %9.4f\n", data, 6);
	      subtractVf (data, data, forceBias, 6);
	      printVf ("Biased: %9.4f\n", data, 6);
	    }
	   else if (COMMANDS ("show continuous-force"))
	    { SIGHAN_RETURN (*oldHandler)();
	      float data[6];
	
	      printf ("Hit <INTR> character to stop display\n");
	      interruptHit = 0;
	      oldHandler = signal (SIGINT, onIntr);
	      while (!interruptHit)
	       { 
		 if (!getForceData(data, fd))
		  { printf ("Warning: force data not valid\n");
		  }
		 if (COMMAND ("show continuous-force raw"))
		  { printVf ("Raw:    %9.4f\n", data, 6);
		  }
		 else
		  { subtractVf (data, data, forceBias, 6);
	   	    printVf ("Biased: %9.4f\n", data, 6);
		  }
		 nap (1.0);
	       }
	      signal (SIGINT, oldHandler);
	    }
	   else if (COMMANDS ("show transforms"))
	    { printTrsf ("T6toSensor:  %m\n", T6toSensor);
	      printf    ("      (RPY:  %8.3f %8.3f %8.3f)\n\n",
			  TSroll, TSpitch, TSyaw);			    
	      printTrsf ("SensorToJig: %m\n", SensorToJig);
	      printf    ("      (RPY: %8.3f %8.3f %8.3f)\n\n",
			  SJroll, SJpitch, SJyaw);			    
	    }
	   else if (COMMANDS ("show position"))
	    { TRSF T6real;
	      JNTS j6real;

	      getRealT6 (mnp, &T6real);
	      getRealJ6 (mnp, &j6real);
	      printTrsf ("T6 (target): %m\n\n", mnp->t6);
	      printTrsf ("T6 (actual): %m\n\n", &T6real);
	      printJf   ("Angles:      %8.3f\n\n", j6real.v, 6);

	    }
	   else if (COMMAND ("show dSlopeAndOffsets"))
	    { printf ("d1Slope=%g d1Offset=%g\n", d1Slope, d1Offset);
	      printf ("d2Slope=%g d2Offset=%g\n", d2Slope, d2Offset);
	    }
	   else if (COMMAND("show massDistPairs"))
	    { MassDistPair *pair;

	      if (mdList == NULL)
	       { printf ("no pairs currently defined\n");
	       }
	      else
	       { printf ("     Name  mass(kg) dist(mm)\n");
			/*XXXXXXXXXX XXXX.XXX XXXX.XXX*/
		 for (pair=mdList; pair; pair=pair->next)
		  { printf ("%10s %8.3f %8.3f\n", 
			     pair->name, pair->mass, pair->dist);
		  }
	       }
	    }
	   else if (COMMANDS ("reset bias"))
	    { if (!getForceData (forceBias, fd))
	       { printf ("Warning: force data not valid; bias probably bad\n");
	       }
	    }
	   else if (COMMANDS ("enter data"))
	    { printf ("Mass1, d1, mass2, d2 (units: kg,mm)> ");
	      if (parseMassDist (tfp, &mass1, &d1) == 1 &&
		  parseMassDist (tfp, &mass2, &d2) == 1)
	       { TRSF TT, D1, D2, FT1, FT2;
		 JNTS j6real;
		 FORCE g, f1, f2, ftotal;
		 float dataSum[6];
		 int i;

		 d1 = d1*d1Slope + d1Offset;
		 d2 = d2*d2Slope + d2Offset;

		 getRealT6 (mnp, &TT);
		 getRealJ6 (mnp, &j6real);
		 multTrsf (&TT, &TT, T6toSensor);
		 multTrsf (&TT, &TT, SensorToJig);
		 identTrsf (&D1);
		 identTrsf (&D2);
		 if (COMMAND ("enter data x-axis"))
		  { xyzToTrsf (&D1, d1, 0, 0);
		    xyzToTrsf (&D2, d2, 0, 0);
		  }
		 else
		  { xyzToTrsf (&D1, 0, d1, 0);
		    xyzToTrsf (&D2, 0, d2, 0);
		  }
		 invertTrsf (&FT1, SensorToJig);
		 multLiTrsf (&FT1, &D1, &FT1);
		 invertTrsf (&FT2, SensorToJig);
		 multLiTrsf (&FT2, &D2, &FT2);

		 zeroForce (&g);
		 g.f.z = mass1 * 9.8;
		 rotateVect (&g.f, &g.f, &TT);
		 transForce (&f1, &g, &FT1);

		 zeroForce (&g);
		 g.f.z = mass2 * 9.8;
		 rotateVect (&g.f, &g.f, &TT);
		 transForce (&f2, &g, &FT2);

		 addVf ((float*)&ftotal, (float*)&f1, (float*)&f2, 6);
		
		 zeroVf (dataSum, 6);

		 for (i=0; i<NSAMP; i++)
		  { float data[6];
		    int numTries = 3;

		    while (!getForceData(data, fd) && --numTries)
		       ;
		    if (numTries == 0)
		     { break;
		     }
		    subtractVf (data, data, forceBias, 6);
		    addVf (dataSum, dataSum, data, 6);
		  }
		 if (i<NSAMP)
		  { printf ("Error: couldn't get valid data\n");
		  }
		 else
		  { scaleVf (dataSum, 1/(512.0*NSAMP), dataSum, 6);

		    printVf ("Force: %9.4f\n", (float*)&ftotal, 6);
		    printVf ("Data:  %9.6f\n", dataSum, 6);

		    if (ofp)
		     { writeDataFile (ofp, dataSum, &ftotal,
				      mass1, d1, mass2, d2, j6real.v);
		     }
		  }
	       }
	      while (getc(tfp) != '\n')
		 ;
	    }
	   else if (COMMAND ("parseMassDist"))
	    { int code;
	      double mass, dist;

	      if (parseMassDist (tfp, &mass, &dist) < 0)
	       { printf ("parseMassDist returns error\n");
	       }
	      else
	       { printf ("mass=%g dist=%g\n", mass, dist);
		 if (parseMassDist (tfp, &mass, &dist) < 0)
		  { printf ("parseMassDist returns error\n");
		  }
		 else
		  { printf ("mass=%g dist=%g\n", mass, dist);
		  }
	       }
	      while (getc(tfp) != '\n')
		 ;
	    }
	 }
	while (!COMMAND ("quit"));

	if (ofp)
	 { fprintf (ofp, "c = f / d\n");
	   fclose (ofp);
	 }
	rcclRelease (1);	
}
