#include "../rccl/completionQueue.h"

#define LOCK_MEM_WORKS	0

rcclCqInit (cq, buf, size)
CompletionQueue *cq;
CompletionQueueEntry *buf;
int size;	
{
	cq->size = size;
	cq->head = 1;
	cq->tail = 0;
	cq->buf = buf;
	return (1);
}

int rcclCqFlush (cq)
CompletionQueue *cq;
{
	int queueNotEmpty = 0;
	int newTail;

#if LOCK_MEM_WORKS
	lockMem (cq);
#endif
	newTail = cq->head-1;
	if (newTail < 0)
	 { newTail += cq->size;
	 }
	if (newTail != cq->tail)
	 { queueNotEmpty = 1;
	 }
	cq->tail = newTail;
#if LOCK_MEM_WORKS
	unlockMem (cq);
#endif
	return (queueNotEmpty);
}

rcclCqNumItems (cq)
CompletionQueue *cq;
{
	/* return the number of items on the completion queue */
	int n;

#if LOCK_MEM_WORKS
	lockMem (cq);
#endif
	n = cq->head - cq->tail - 1;
#if LOCK_MEM_WORKS
	unlockMem (cq);
#endif
	if (n < 0)
	 { n += cq->size;
	 }
	return (n);
}

rcclCqAdd (cq, ent)
CompletionQueue *cq;
CompletionQueueEntry *ent;
{
	/* add an item to the queue, overwriting stuff at the end
	   if necessary */

	int status = 0;
#if LOCK_MEM_WORKS
	int nextHead;
				
	lockMem (cq);
	nextHead = (cq->head+1)%cq->size;
	cq->buf[cq->head] = *ent;
	if (cq->head == cq->tail)
	 { cq->tail = nextHead;
	   status = 1;
	 }
	cq->head = nextHead;
	unlockMem (cq);
#else
	if (cq->head != cq->tail)
	 { cq->buf[cq->head] = *ent;
	   cq->head = (cq->head+1)%cq->size;
	   status = 1;
	 }	
#endif
	return (status);
}

rcclCqRemove (cq, ent)
CompletionQueue *cq;
CompletionQueueEntry *ent;
{
	int nextTail;
	int gotSomething = 0;

#if LOCK_MEM_WORKS
	lockMem (cq);
#endif
	nextTail = (cq->tail+1)%cq->size;
	if (nextTail != cq->head)
	 { *ent = cq->buf[nextTail];
	   cq->tail = nextTail;
	   gotSomething = 1;
	 }
#if LOCK_MEM_WORKS
	unlockMem (cq);
#endif
	return (gotSomething);
}
