/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

/*---------------------------------------------------------------------------
 *		
 * MODULE NAME:	getMisc
 *
 * DESCRIPTION:	Routines to read back compliance settings.
 *
 * AUTHOR/DATE:	John Lloyd, Fall 1988
 *
 * MODIFICATION HISTORY:
 *
 *--------------------------------------------------------------------------*/

#include "../h/cdefs.h"
#include "../h/rcclDefs.h"

#include "../rccl/motionRequest.h"

#define NmToNmm         1000.

getComply(mnp, mask, force)
MANIP *mnp;
unsigned long *mask;
FORCE *force;
{
	/* RCCL primitive to read back the set compliance specification.
	 */

	char *fxnName = "getComply()";

	MOTION_CONTROL_BLOCK *mcb = (MOTION_CONTROL_BLOCK*)mnp->mcb;
	register int index;

	do
	 { index = mcb->complyIndex;
	   if (force)
	    { *force = mcb->setComplyValues[index&1];
	      force->m.x /= NmToNmm;
	      force->m.y /= NmToNmm;
	      force->m.z /= NmToNmm;
	    }
	   *mask = mcb->setComplyMask[index&1];
	 }
	while ((index&~1) != (mcb->complyIndex&~1));
	return (1);
}

getImpedance(MANIP *mnp, float *k, float *invB, TRSF *COC)
{
	/* RCCL primitive to read back the set compliance specification.
	 */

	char *fxnName = "getImpedance()";

	MOTION_CONTROL_BLOCK *mcb = (MOTION_CONTROL_BLOCK*)mnp->mcb;
	register int index;

	do
	 { index = mcb->impedanceIndex;
	   if (k)
	    { copyVf (k, &mcb->setImpedanceK[index&1][0], 6);
	    }
	   if (invB)
	    { copyVf (k, &mcb->setImpedanceInvB[index&1][0], 6);
	    }
	   if (COC)
	    { *COC = mcb->setImpedanceCOC[index&1];
	    }
	 }
	while ((index&~1) != (mcb->impedanceIndex&~1));
	return (1);
}

getActiveComply(mnp, mask, force)
MANIP *mnp;
unsigned long *mask;
register FORCE *force;
{
	/* RCCL primitive to read back the compliance spec. currently
	   active on a particular manipulator.
	 */

	char *fxnName = "getActiveComply()";

	MOTION_CONTROL_BLOCK *mcb = (MOTION_CONTROL_BLOCK*)mnp->mcb;
	register int index;

	do
	 { index = mcb->pathsegIndex;
	   if (force)
	    { *force = mcb->activeComplyVals[index&1];
	      force->m.x /= NmToNmm;
	      force->m.y /= NmToNmm;
	      force->m.z /= NmToNmm;
	    }
	   *mask = mcb->activeComplyMask[index&1];
	 }
	while ((index&~1) != (mcb->pathsegIndex&~1));
	return (1);
}

getActiveImpedance (MANIP *mnp, float *k, float *invB, TRSF *COC)
{
	/* RCCL primitive to read back the compliance spec. currently
	   active on a particular manipulator.
	 */

	char *fxnName = "getActiveImpedance()";

	MOTION_CONTROL_BLOCK *mcb = (MOTION_CONTROL_BLOCK*)mnp->mcb;
	int index;

	do
	 { index = mcb->pathsegIndex;
	   if (k)
	    { copyVf (k, &mcb->activeImpedanceK[index&1][0], 6);
	    }
	   if (invB)
	    { copyVf (invB, &mcb->activeImpedanceInvB[index&1][0], 6);
	    }
	   if (COC)
	    { *COC = mcb->activeImpedanceCOC[index&1];
	    }
	 }
	while ((index&~1) != (mcb->pathsegIndex&~1));
	return (1);
}
