/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1991 by John E. Lloyd.				    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#if defined(Sun4sol_TARGET) || defined(X86sol_TARGET)

#include <stdio.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/lock.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <signal.h>
#include <errno.h>
#include <sys/priocntl.h>
#include <sys/lwp.h>

#include "../h/cdefs.h"	
#include "rciConfig.h"
#include "../h/rciTask.h"
#include "../h/errorCodes.h"
#include "../h/errorStack.h"

#include "rciioctl.h"

#include "../rci/link.h"
#include "../rci/rciAutoAbort.h"

#include "../rci/sun4solCtrlTask.h"

/* #define VME_A16NPBASE	0xbd000000	/* XXX XXX XXX */
/* #define VME_A16NPBASE	0xb7c10000	/* XXX XXX XXX */
#define VME_A16NPBASE   HOST_IOPAGE_BASE
#define VME_A16NPSIZE	0x10000

/* #define VME_A24NPBASE	0xb2000000 */
#define VME_A24NPBASE   HOST_BUS_BASE
#define VME_A24NPSIZE	0x1000000

#define VME_A24RCIBASE	(VME_A24NPBASE+0x800000)
#define VME_A24RCISIZE	0x800000

#define NUM_GENRCI_DEVS	4

extern int rciIsOpen;
extern RciProcId rciCtrlTaskId;
extern RciProcId rciMainTaskId;
extern lwpid_t rciCtrlTaskLwp;
extern lwpid_t rciMainTaskLwp;
extern int rciClockFd;
extern int rciPlanCtrlFds[2];
extern int rciCtrlPlanFds[2];

extern FILE *rciStdout;

extern fd_set intrDevFds;
extern IntrDevDesc *intrDevList;

static char *memdev = "/dev/mmem";
static int memFd;

#if USE_RAW_LWP
static ucontext_t *rciCtrlTaskCtx;
#endif

extern unsigned char *rciBus;
extern unsigned char *rciIOpage;
extern int rciCpu;
extern int rciSimOnly;

#if USE_RAW_LWP
extern void rciCtrlTask();
#else
extern void *rciCtrlTask();
#endif

static int rciClkFxnHack (int genFd);

/* ----- Error Messages. Defined externally to save space ----- */

static char *nullMsg		= "%s --\n";
static char *cantOpenRciDev	= "%s -- can't open %s, errno %d\n";
static char *noAvailRciDevs	= "%s -- no generic RCI devices available\n";
static char *badIoctlMsg	= "%s -- %s ioctl failed, errno %d\n";
static char *outOfMemoryMsg	= "%s -- out of memory\n";
static char *alreadyAttachedMsg	= "%s -- device already attached\n";
static char *nullCallMsg = "probable call through NULL at control level\n";
static char *noResponseMsg	= "%s -- no response from control task\n";
static char *cantMapMemMsg	= "%s -- can't map %s, errno %d\n";
static char *notSupportedMsg    = "%s -- not supported on this system\n";

static SIGHAN_RETURN onSigXcpu();
static doHostClose();
static IntrDevDesc *devLocateById();
static touchMemory();

static initVariables()
{
	rciIsOpen = 0;
	rciCtrlTaskId = 0;
	rciMainTaskId = 0;
	rciClockFd = -1;
	rciPlanCtrlFds[0] = -1;
	rciPlanCtrlFds[1] = -1;
	rciCtrlPlanFds[0] = -1;
	rciCtrlPlanFds[1] = -1;
	memFd = -1;
	rciStdout = 0;
	rciBus = (unsigned char*)0;
	rciIOpage = (unsigned char*)0;
	FD_ZERO(&intrDevFds);
}

#define RET_ERROR	doHostClose(); return (-1);

static doSimOpen()
{
	int fd;

	rciBus = (unsigned char*)0;
	rciIOpage = (unsigned char*)0;
	rciIsOpen = 1;
	/* create a duplicate of stdout for the control task ... */
	if ((fd = dup(1)) < 0)
	 { perror ("doSimOpen: dup");
	   exit (1);
	 }	
	rciStdout = fdopen (fd, "w");
	rciMainTaskId = GET_TASK_ID();
}

rciHostOpen(sig, period)
int sig;
int *period;
{
	char *fxnName = "rciHostOpen()";
	void *mem;
	int numBusy;
	char com;
	int fd;
	int k;

	initVariables();

	if (rciSimOnly == 1)
	 { doSimOpen();
	   return (1);
	 }

	/* try and open a generic RCI device */

	numBusy = 0;
	for (k=0; k<NUM_GENRCI_DEVS; k++)
	 { char devName[32];

	   sprintf (devName, "/dev/genrci%d", k);
	   if ((rciClockFd = open (devName, O_RDWR)) < 0)
	    { if (errno != EBUSY && errno != ENXIO && errno != ENOENT)
	       { errorMsg (EBadAccess, cantOpenRciDev, fxnName, 
						       devName, errno);
		 RET_ERROR;
	       }
	      if (errno == EBUSY)
	       { numBusy++;
	       }
	    }
	   else
	    { break;
	    }
	 }
	if (k == NUM_GENRCI_DEVS)
	 { if (numBusy)
	    { errorMsg (ENotAvail, noAvailRciDevs, fxnName);
	      RET_ERROR;
	    }
	   else
	    { /* If we can't open any generic RCI devices, then
		 force us into SIMONLY mode. */
printf ("rciHostOpen(): forcing sim only\n");
	      rciSimOnly = 1;
	      doSimOpen();
	      return (1);
	    }
	 }
	if (ioctl (rciClockFd, RCIIOC_CONFIRM, 0) < 0)
	 { sysErrorMsg (EBadObj, fxnName, "ioctl CONFIRM");
	   RET_ERROR;
	 }
	if (ioctl (rciClockFd, RCIIOC_TIMER_ENABLE, 1) < 0)
	 { sysErrorMsg (EBadObj, fxnName, "ioctl TIMER ENABLE");
	   RET_ERROR;
	 }
#ifdef X86sol_TARGET
	/* On Intel PC machines, we need to explictly enable access
	   to the I/O ports from user mode. On Solaris X86 machines, 
	   we do this by setting bits 0x3000 in the EFLAGS register,
	   using a special ioctl */
	if (ioctl (rciClockFd, RCIIOC_ENABLE_IOPORTS, 1) < 0 ||
	    ioctl (rciClockFd, RCIIOC_ENABLE_IOPORTS, 0) < 0)
	 { sysErrorMsg (EBadObj, fxnName, "ioctl ENABLE IOPORTS");
	   RET_ERROR;
	 }
#endif
	FD_ZERO(&intrDevFds);

#if defined(UVax_TARGET) || defined(Vxw_TARGET)
#define ARB_MALLOC_SIZE 0x20000 /* 128K */
#else
#define ARB_MALLOC_SIZE	0x40000	/* 256K */
#endif
	if (localMemoryAlloc (ARBITER_CPU, &mem, ARB_MALLOC_SIZE) < 0)
	 { errorMsg (EMemory, nullMsg, fxnName);
	   RET_ERROR;
	 }
	rciMallocInit (NULLF, 0, mem, ARB_MALLOC_SIZE);

#if 0
	if ((memFd = open(memdev,O_RDWR,0)) < 0)
	 { errorMsg (EBadAccess, cantOpenRciDev, fxnName, memdev, errno);
	   RET_ERROR;
	 }
	rciIOpage = (unsigned char*)mmap 
			(0, VME_A16NPSIZE, PROT_READ|PROT_WRITE,
				    MAP_PRIVATE, memFd, VME_A16NPBASE);
	if (rciIOpage == (unsigned char*)-1)
	 { sysErrorMsg (EBadAccess, fxnName, "mmap for rciIOpage:");
	   RET_ERROR;
	 }
	rciBus = (unsigned char*)mmap 
			(0, VME_A24RCISIZE, PROT_READ|PROT_WRITE,
				    MAP_PRIVATE, memFd, VME_A24RCIBASE);
	if (rciBus == (unsigned char*)-1)
	 { sysErrorMsg (EBadAccess, fxnName, "mmap for rciBus:");
	   RET_ERROR;
	 }
#endif
	if ((*period = rciArbClkOpen()) <= 0)
	 { 
	   if (*period < 0)
	    { errorMsg (getErrorCode(0), nullMsg, fxnName);
	      return (-1);
	    }
	   *period = 10;	/* 10 milliseconds default */	
	 }
	rciClkFxnHack(rciClockFd);

	/* now start up the control task ... */

	/* lock down memory and page it in */

#ifndef REALTIME_DISABLE
	if (mlockall (MCL_CURRENT | MCL_FUTURE) < 0)
	 {  perror ("mlockall:");
	    exit (1);
	 }
#endif
#if 0
	/* UGHH ... page it all in 'cause plock won't. */
	 { extern etext, edata, _ftext, _fdata;

	   touchMemory ((char*)&_ftext, (char*)&etext);
	   touchMemory ((char*)&_fdata, (char*)&edata);
	 }
#endif
	/* open two pipes so that the main task and the control
	   task can communicate ... */

	if ((pipe (rciCtrlPlanFds) < 0) ||
	    (pipe (rciPlanCtrlFds) < 0))
	 { sysErrorMsg (EInternal, fxnName, "pipe");
	   RET_ERROR;
	 }

	/* create a duplicate of stdout for the control task to use ... */

	if ((fd = dup(1)) < 0)
	 { sysErrorMsg (EInternal, fxnName, "dup");
	   RET_ERROR;
	 }	
	rciStdout = fdopen (fd, "w");

#if 0
	/* Arrange for the planning and control tasks to send
	   themselves a SIGXCPU when they exit. */

	prctl (PR_SETEXITSIG, SIGXCPU);
#endif

	/* find out how many CPUs the system has, and the constrain
	   us to run on only one of them. We *should* check to see
	   if the processor is actually available for operation,
	   but we don't worry about that now */
#if 0
	k = sysmp (MP_NPROCS);
	if (k > 1)
	 { if (sysmp (MP_MUSTRUN, 0) < 0)
	    { sysErrorMsg (EInternal, fxnName, "sproc");
	      errorMsg (EInternal, 
"%s -- can't make RCI application run on cpu number 1\n", fxnName);
	      RET_ERROR;
	    }
	 }
#endif
	/* Now create the control task itself. */

	rciMainTaskId = GET_TASK_ID();
#if USE_RAW_LWP
	rciCtrlTaskCtx = (ucontext_t *)malloc(sizeof(ucontext_t));
	 { char *stack;

	   stack = (char*)malloc(CTRL_STACK_SIZE);
	   sigfillset (&rciCtrlTaskCtx->uc_sigmask);
	   _lwp_makecontext (rciCtrlTaskCtx, rciCtrlTask, NULL, NULL, 
			     stack, CTRL_STACK_SIZE);
	   if (_lwp_create(rciCtrlTaskCtx, /*flags=*/0, &rciCtrlTaskId) != 0)
	    { perror ("lwp_create");
	      exit (1);
	    }
	 }
#else
	rciMainTaskLwp = _lwp_self();
	if (thr_create ((void*)0, 0, rciCtrlTask, (void*)0, 
			 THR_BOUND, &rciCtrlTaskId) != 0)
	 { perror ("thr_create");
	   exit (1);
	 }
#endif
	if ((read_restart (rciCtrlPlanFds[0], &com, 1) != 1) ||
	    (com != RCI_CTRL_ACK))
	 { errorMsg (EInternal, noResponseMsg, fxnName);
	   RET_ERROR;
	 }
	rciIsOpen = 1;
	return (1);
}

static SIGHAN_RETURN onSigXcpu(sig)
int sig;
{
	rciCtrlTaskId = 0;
}

static doHostClose()
{
	SIGHAN_RETURN (*oldHandler)();

	if (rciCtrlTaskId)
	 { oldHandler = signal(SIGXCPU, onSigXcpu);
	   kill (rciCtrlTaskId, SIGXCPU);
	   while (rciCtrlTaskId)
	    { nap (0.01);
	    }
	   signal (SIGXCPU, oldHandler);
	 }
	if (rciStdout)
	 { fclose (rciStdout);
	 }
	if (rciCtrlPlanFds[0] != -1)
	 { close (rciCtrlPlanFds[0]);
	 }
	if (rciCtrlPlanFds[1] != -1)
	 { close (rciCtrlPlanFds[1]);
	 }
	if (rciPlanCtrlFds[0] != -1)
	 { close (rciPlanCtrlFds[0]);
	 }
	if (rciPlanCtrlFds[1] != -1)
	 { close (rciPlanCtrlFds[1]);
	 }
	if (rciBus != (unsigned char*)0 && rciBus != (unsigned char*)-1)
	 { munmap ((char*)rciBus, VME_A24RCISIZE);
	 }
	if (rciIOpage != (unsigned char*)0 && rciIOpage != (unsigned char*)-1)
	 { munmap ((char*)rciIOpage, VME_A16NPSIZE);
	 }
	if (memFd != -1)
	 { close (memFd);
	 }
	if (rciClockFd != -1)
	 { close (rciClockFd);
	 }
	rciArbClkClose();
	initVariables();

	return (1);
}

rciHostClose()
{
	doHostClose();
}

static IntrDevDesc *devLocateById(fd)
int fd;
{
	IntrDevDesc *d;

	for (d=intrDevList; d; d=d->next)
	 { if (d->fd == fd)
	    { return (d);
	    }
	 }
	return ((IntrDevDesc*)NULL);
}


rciInterruptOpen (intr, vector)
IO_INTERRUPT *intr;
int vector;
{
	char *fxnName = "rciInterruptOpen()";

	errorMsg (ENotSupported, notSupportedMsg, fxnName);
	RETURN_ERROR (-1);	
}

rciInterruptClose(vector)
int vector;
{
	return (0);
}

rciIntrAttach(fd, fxn, arg)
int fd;
int (*fxn)();
int arg;
{
	char *fxnName = "rciIntrAttach()";

	pcparms_t schedctx;
	IntrDevDesc *d;

	if (devLocateById(fd))
	 { errorMsg (EBusy, alreadyAttachedMsg, fxnName);
	   return (-1);
	 }
	if (ioctl (fd, RCIIOC_CONFIRM, 0) < 0)
	 { errorMsg (EBadDev, badIoctlMsg, fxnName, "CONFIRM", errno);
	   return (-1);
	 }
	if ((d = ALLOC(IntrDevDesc)) == NULL)
	 { errorMsg (EMemory, outOfMemoryMsg, fxnName, errno);
	   return (-1);
	 }
	d->fd = fd;
	d->fxn = fxn;
	d->arg = arg;
	if (ioctl (fd, RCIIOC_INTR_ENABLE, 0xff) < 0)
	 { errorMsg (EInternal, badIoctlMsg, fxnName, "INTR_ENABLE 0xff", 
		     errno);
	   FREE (d);
	   return (-1);
	 }
	rciSetRealTime (rciGetMaxRTpri()-1, 1, 0, &schedctx);
	LINK_H (intrDevList, d);
	FD_SET (d->fd, &intrDevFds);
	rciClearRealTime (&schedctx);
	return (1);
}

rciIntrDetach(fd)
int fd;
{
	char *fxnName = "rciIntrDetach()";

	pcparms_t schedctx;
	IntrDevDesc *d;
	int rv;

	if ((d = devLocateById(fd)) == NULL)
	 { return (0);
	 }
	d->fxn = 0;
	rv = 1;
	if (ioctl (fd, RCIIOC_INTR_ENABLE, 0) < 0)
	 { errorMsg (EInternal, badIoctlMsg, fxnName, "INTR_ENABLE 0", errno);
	   rv = -1;
	 }
	rciSetRealTime (rciGetMaxRTpri()-1, 1, 0, &schedctx);
	FD_CLR (d->fd, &intrDevFds);
	UNLINK_H (intrDevList, d);
	rciClearRealTime (&schedctx);
	FREE ((char*)d);
	return (rv);
}

rciCrashInfo()
{
	return (0);		
}

static touchMemory(addr1, addr2)
char *addr1, *addr2;
{
	/* Touch all memory in the range [addr1, addr2) */
	int pagesize;
	char *cp;
	volatile char c;

	pagesize = getPageSize();
	for (cp=addr1; cp<addr2; cp+=pagesize)
	 { c = *cp;
	 }
}

static char *clkDevName = "/dev/rciclk";
static int clkFd = UNDEF;

/* IRIX: don't need any clock stuff */
rciArbClkOpen()
{
	char *fxnName = "rciArbClkOpen()";
	int period;

#if defined(SIMONLY_CFG)
	return (0);
#else
	if (clkFd == UNDEF)
	 { if ((clkFd = open (clkDevName, O_RDWR)) < 0)
	    { clkFd = UNDEF;
	      if (errno != ENXIO && errno != ENOENT)
	       { sysErrorMsg (EBadDev, fxnName, "open");
		 return (-1);
	       }
	      else
	       { return (0);
	       }
	    }
	 }
	if (ioctl (clkFd, RCIIOC_GETCLKRATE, &period) < 0)
	 { sysErrorMsg (EBadIoctl, fxnName, "ioctl GETCLKRATE");
	   rciArbClkClose();
	   return (-1);
	 }
	return(period);
#endif
}

static int rciClkFxnHack (int genFd)
{
	char *fxnName = "rciClkFxnHack()";

	RciFxnInterface fxni;

	if (clkFd != UNDEF)
	 { 
	   if (ioctl (genFd, RCIIOC_GET_FXNS, &fxni) < 0)
	    { sysErrorMsg (EBadIoctl, fxnName, "ioctl GET_FXNS");
	      return (-1);
	    }
	   if (ioctl (clkFd, RCIIOC_SET_FXNS, &fxni) < 0)
	    { sysErrorMsg (EBadIoctl, fxnName, "ioctl SET_FXNS");
	      return (-1);
	    }
	 }
	return (1);
}

rciArbClkClose()
{
	if (clkFd != UNDEF)
	 { close (clkFd);
	   clkFd = UNDEF;
	   return (1);
	 }
	else
	 { return (0);
	 }
}

static clockCtrl(fxnName, req, reqName, argp)
char *fxnName;
unsigned long req;
char *reqName;
char *argp;
{
	int rv;

#if defined(SIMONLY_CFG)
	return (0);
#else
	if (rciSimOnly)
	 { return (0);
	 }
	if (clkFd == UNDEF)
	 { if ((clkFd = open (clkDevName, O_RDWR)) < 0)
	    { if (errno != ENXIO && errno != ENOENT)
	       { sysErrorMsg (EBadDev, fxnName, "open");
		 return (-1);
	       }
	      else
	       { return (0);
	       }
	    }
	 }
	rv = 1;
	if (ioctl (clkFd, req, argp) < 0)
	 { sysErrorMsg (EBadIoctl, fxnName, "ioctl");
	   rv = -1;
	 }
	return (rv);
#endif
}

rciArbClkSetPeriod(period)
int period;
{
	char *fxnName = "rciArbClkSetPeriod()";

#if defined(SIMONLY_CFG)
	return (0);
#else
	return (clockCtrl (fxnName, RCIIOC_SETCLKRATE,"SETCLKRATE",&period));
#endif
}

rciArbClkGetPeriod(period)
int *period;
{
	char *fxnName = "rciArbClkGetPeriod()";

#if defined(SIMONLY_CFG)
	return (0);
#else
	return (clockCtrl (fxnName, RCIIOC_GETCLKRATE, "GETCLKRATE", period));
#endif
}

rciArbClkSetDebug(arg)
int arg;
{
	char *fxnName = "rciArbClkSetDebug()";

#if defined(SIMONLY_CFG)
	return (0);
#else
	return (clockCtrl (fxnName, RCIIOC_SETCLKDEBUG, "SETCLKDEBUG", &arg));
#endif
}

rciArbClkEnable()
{
	char *fxnName = "rciArbClkEnable()";

#if defined(SIMONLY_CFG)
	return (0);
#else
	return (clockCtrl (fxnName, RCIIOC_CLKENABLE, "CLKENABLE", NULL));
#endif
}

rciArbClkDisable()
{
	char *fxnName = "rciArbClkDisable()";

#if defined(SIMONLY_CFG)
	return (0);
#else
	return (clockCtrl (fxnName, RCIIOC_CLKDISABLE, "CLKDISABLE", NULL));
#endif
}

rciArbClkStatus(arg)
int *arg;
{
	char *fxnName = "rciArbClkStatus()";

#if defined(SIMONLY_CFG)
	return (0);
#else
	return (clockCtrl (fxnName, RCIIOC_CLKSTATUS, "CLKSTATUS", arg));
#endif
}

#else /* !defined(Sun4sol_TARGET) && !defined(X86sol_TARGET) */

static int keepRanlibHappy;

#endif
