/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#include <stdio.h>

#include "../h/cdefs.h"
#include "../h/errorCodes.h"
#include "../h/errorStack.h"
#include "../h/robot_jls.h"
#include "../h/utilFxns.h"

#define MAX_NAMELEN	63

extern char *rciGetenv();

static char robotName[MAX_NAMELEN+1] = {'\0'};
static int initDone = 0;

/* ----- Error Messages. Defined externally to save space ----- */

static char *noEnvMsg		= "%s -- environment variable %s not set\n";
static char *cantAccessFileMsg	= "%s -- can't access file %s\n";
static char *cantReadFileMsg	= "%s -- can't read file %s\n";

char *getDefaultRobot ()
{	
	char *fxnName = "defaultRobot()";

	char *fileName = NULL;
	static char baseName[] = "defaultRobot.cfg";

	char *confDir;

	if (!planningLevelCheck(fxnName))
	 { return (NULL);
	 }
	if (!initDone)
	 { FILE *fp;

	   if ((fp = locateFile (&fileName, baseName)) == NULL)
	    { errorMsg (EBadAccess, cantAccessFileMsg, fxnName, baseName);
	      return (NULL);
	    }
	   if (fscanf (fp, "%63s", robotName) != 1)
	    { errorMsg (EBadObj, cantReadFileMsg, fxnName, fileName);
	      free (fileName);
	      fclose (fp);
	      return (NULL);
	    }
	   robotName[MAX_NAMELEN] = '\0';
	   free (fileName);
	   fclose (fp);
	   initDone = 1;
	   requestFlagInit(&initDone);
	 }
	return (robotName);
}
