/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1991 by John E. Lloyd.				    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#ifdef SUNVIEW_GRAPHICS

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>

#include <suntool/sunview.h>
#include <suntool/canvas.h>
#include <suntool/panel.h>

#include "../h/cdefs.h"
#include "../h/timeval.h"
#include "../h/grf4.h"
#include "grf4Internal.h"

typedef struct pixrect PIXRECT;

static Frame frame;
static int screenHeight;		/* height of screen */
static frameMargX;
static frameMargY;
static PIXRECT *screenBuf;
static Pixfont *font;

static int winPipe[2] = {-1, -1 };

static void svPutText();
static void svGetTextExtent();
static void svGetButtonExtent();
static void svDrawButton();
static void svGetWindowSize();
static void svClear();
static void svColor();
static void svFlush();
static void svDraw();
static void svPoint();
static int svGetEventFd ();
static int svEnableEvent ();
static int svDisableEvent ();
static int svPollEvent ();
static int svReadEvent ();
static void canvasRepaint ();
static void canvasResize ();
static Notify_value frameEventCatch();
static eventProc();

typedef struct {
	Canvas id;	/* id returned from window_create() */
	Pixwin *pix;	/* pixrect for the window */
	long x, y;	/* position of the window (WRT the frame) */
	long w, h;	/* height and width */
	int active;	/* true if the window is active */
	unsigned long eventMask; /* selected events */
	int cx, cy;	/* cursor position */
	int inFocus;	/* window is in focus */
} WinDesc;

static WinDesc wdesc[NUMWINS];
static int eventRef[GRF_MAX_EVENT_TYPES+1];

#ifdef CTRL	/* sun seems to define this somewhere */
#undef CTRL
#endif

#define MAIN	wdesc[MAINWIN]
#define CTRL	wdesc[CTRLWIN]

#define BUTTON1			1
#define BUTTON_BORDER_WIDTH	2

static void svPutText (str, x, y)
char *str;
int x, y;
{
	pw_ttext (MAIN.pix, x, (MAIN.h-1-y), PIX_SET, font, str);
}

static void svGetTextExtent (xp, yp, wp, hp, str)
int *xp, *yp, *wp, *hp;
char *str;
{
	struct pr_subregion bound;

	pf_textbound (&bound, str, font, str);
	if (xp) *xp = - bound.pos.x;
	if (yp) *yp = ((bound.size.y-1)+bound.pos.y);
	if (wp) *wp = bound.size.x;
	if (hp) *hp = bound.size.y;
}

static void svGetButtonExtent(b)
BUTTON_DESC *b;
{
	int x, y, w, h;
	int margin;

	svGetTextExtent (&x, &y, &w, &h, b->name);
	margin = 1 + BUTTON_BORDER_WIDTH;
	b->coffX = margin + x;
	b->coffY = margin + y;
	b->minw = 2*margin + w;
	b->minh = 2*margin + h;
}

static void svDrawButton(b)
BUTTON_DESC *b;
{
	int baseX, baseY;
	int cbaseX, cbaseY;

	baseX = b->x;
	baseY = (CTRL.h-1)-b->y-b->h+1;
	cbaseX = baseX + b->coffX + (b->w-b->minw)/2;
	cbaseY = (CTRL.h-1)-b->y-b->coffY - (b->h-b->minh)/2;

	if (!(b->flags & BF_NOTBOXED))
	 { pw_rop(CTRL.pix,baseX,baseY,b->w,b->h,
		  PIX_SET, (PIXRECT*)0, 0,0);
	 }
	else
	 { pw_rop(CTRL.pix,baseX,baseY,b->w,b->h,
		  PIX_CLR, (PIXRECT*)0, 0,0);
	 }
	if (b->flags & BF_ILLUMINATED)
	 { pw_ttext (CTRL.pix, cbaseX, cbaseY, PIX_CLR, font, b->name);
	 }
	else
	 { pw_rop(CTRL.pix,baseX+BUTTON_BORDER_WIDTH,
		       baseY+BUTTON_BORDER_WIDTH,
		       b->w-2*BUTTON_BORDER_WIDTH,
		       b->h-2*BUTTON_BORDER_WIDTH,
		  PIX_CLR, (PIXRECT*)0, 0,0);
	   pw_ttext (CTRL.pix, cbaseX, cbaseY, PIX_SET, font, b->name);
	 }
}

grf4SvInit(label, fgcolor, bgcolor, placing, flags)
char *label;
int fgcolor;
int bgcolor;
WinPlaceDesc *placing;
int flags;
{
	Rect *scrRect;
	int rootw, rooth;
	int canvw, canvh;
	int x, y, w, h;

	font = pf_default();

	buttonInit (&CTRL.w, &CTRL.h);
	CTRL.active = (CTRL.w > 0);

	/* set up default window sizes */

	if (!placing->placed)
	 { x = y = 200;
	   w = h = 300;
	 }
	w = MAX(w,placing->minW);
	h = MAX(h,placing->minH);

	/* Can't implement run time mimimum spacing at the moment */

	/* create root window frame, and canvas subwindow */
	
	rootw = w+CTRL.w+1;	/* the extra 1 is for the dividing line */
	rooth = MAX(h,CTRL.h);

	frame = window_create (NULL, FRAME, 
			       FRAME_LABEL, label, 
			       FRAME_SUBWINDOWS_ADJUSTABLE, FALSE,
			       WIN_X, x,
			       WIN_Y, y,
			       0);
	if (frame == 0)
	 { return (-1);
	 }

	scrRect = (Rect*)window_get (frame, WIN_SCREEN_RECT);

	wdesc[NULLWIN].x = -x;	/* position is WRT the frame */
	wdesc[NULLWIN].y = -y;
	wdesc[NULLWIN].w = scrRect->r_width;
	wdesc[NULLWIN].h = scrRect->r_height;
	wdesc[NULLWIN].cx = -1;
	wdesc[NULLWIN].cx = -1;
	screenHeight = scrRect->r_height;

	MAIN.id = window_create (frame, CANVAS,
				CANVAS_RETAINED, FALSE,
				CANVAS_FIXED_IMAGE, FALSE,
				CANVAS_REPAINT_PROC, canvasRepaint,
				CANVAS_RESIZE_PROC, canvasResize,
				CANVAS_FAST_MONO, TRUE,
				WIN_X, 0,
				WIN_Y, 0,
				WIN_WIDTH, w,
				WIN_HEIGHT, MAX(h,CTRL.h),
				WIN_EVENT_PROC, eventProc, 0,
				0);

	if (MAIN.id == 0)
	 { fprintf(stderr,"grf4SvInit(): can't create MAIN.id\n");
	   exit(1);
	 }

	MAIN.w = w;
	MAIN.h = rooth;
	MAIN.x = 0;
	MAIN.y = 0;

	if (MAIN.w > 0)
	 { screenBuf = mem_create(MAIN.w,MAIN.h,1);
	   if (screenBuf == 0)
	    { fprintf(stderr,"Can't create draw buffer\n");
	      exit(1);
	    }
	   pr_rop (screenBuf, 0,0,MAIN.w,MAIN.h, PIX_CLR,(PIXRECT *)0,0,0);
	 }
	else
	 { screenBuf = 0;
	 }
	MAIN.pix = canvas_pixwin(MAIN.id);
	if (bgcolor)
	 { pw_blackonwhite(MAIN.pix,0,1);
	 }
	else
	 { pw_whiteonblack(MAIN.pix,0,1);
	 }

	if (CTRL.active)
	 { CTRL.id = window_create (frame, CANVAS,
				CANVAS_RETAINED, FALSE,
				CANVAS_FIXED_IMAGE, FALSE,
				CANVAS_REPAINT_PROC, canvasRepaint,
				CANVAS_RESIZE_PROC, canvasResize,
				CANVAS_FAST_MONO, TRUE,
				WIN_X, MAIN.w+2,
				WIN_Y, 0,
				WIN_WIDTH, CTRL.w,
				WIN_HEIGHT, CTRL.h,
				WIN_EVENT_PROC, eventProc, 0,
				0);
	   if (CTRL.id == 0)
	    { fprintf(stderr,"grf4SvInit(): can't create CTRL.id\n");
	      exit(1);
	    }

	   CTRL.pix = canvas_pixwin(CTRL.id);
	   if (bgcolor)
	    { pw_blackonwhite(CTRL.pix,0,1);
	    }
	   else
	    { pw_whiteonblack(CTRL.pix,0,1);
	    }
	   CTRL.x = MAIN.w+1;
	   CTRL.y = 0;
	 }
	window_set (frame, WIN_SHOW, TRUE, 0);
	window_fit (frame);

	/* It turns out that the frame needs a bit of extra space around
	   the edges to fit everything in. So we try to determine what
	   this space is for when we resize things ourselves. */

	frameMargX = (int)window_get(frame, WIN_WIDTH) - (MAIN.w+CTRL.w+2);
	frameMargY = (int)window_get(frame, WIN_HEIGHT) - (MAIN.h);

	notify_interpose_event_func (frame, frameEventCatch, NOTIFY_SAFE);

	pw_rop(MAIN.pix,0,0,MAIN.w,MAIN.h, PIX_CLR,(PIXRECT *)0,0,0);
	if (CTRL.active)
	 { pw_rop(CTRL.pix,0,0,CTRL.w,CTRL.h, PIX_CLR,(PIXRECT *)0,0,0);
	 }

	pipe(winPipe);
	fcntl (winPipe[0], F_SETFL, fcntl(winPipe[0],F_GETFL,0) | FNDELAY);
	notify_do_dispatch();

	return(1);
}

/* TODO: get rid of the window size functions */

static void svGetWindowSize(w,h,directionOfY)
int *w, *h;
int *directionOfY;
{
	*w = MAIN.w;
	*h = MAIN.h;
	*directionOfY = -1;
}

static void svClear(x,y,w,h)
int x,y,w,h;
{
	if (screenBuf)
	 { pr_rop (screenBuf, x, y, w, h, PIX_CLR,(PIXRECT *)0,0,0);
	 }
}

static void svColor(color)
int color;
{
}

static void svFlush(x,y,w,h)
int x,y,w,h;
{
	if (screenBuf)
	 { pw_rop(MAIN.pix,x,y,w,h,PIX_SRC, screenBuf, x,y);
	 }
}
	 
static void svDraw(x1,y1,x2,y2)
int x1, y1, x2, y2;
{
	if (screenBuf)
	 { pr_vector(screenBuf,x1,y1,x2,y2,(PIX_SET/*|PIX_DONTCLIP*/),1);
	 }
}

static void svPoint(x1,y1)
int x1, y1;
{
	if (screenBuf)
	 { pr_put(screenBuf,x1,y1,~0);
	 }
}

static int winIdToNum (id)
Canvas id;
{
	if (id == MAIN.id)
	 { return (MAINWIN);
	 }
	else if (id == CTRL.id)
	 { return (CTRLWIN);
	 }
	else
	 { return (NULLWIN);
	 }
}

static Notify_value frameEventCatch (frame, event, arg, type)
Frame frame;
Event *event;
Notify_arg arg;
Notify_event_type type;
{
	Notify_value val;
	int w, h;
	int mw, mh;

	val = notify_next_event_func (frame, event, arg, type);
	if (!(int)window_get (frame, FRAME_CLOSED))
	 { if (event_action(event) == WIN_RESIZE)
	    { 
	      w = (int)window_get (frame, WIN_WIDTH);
	      h = (int)window_get (frame, WIN_HEIGHT);

	      mw = w - frameMargX - CTRL.w - 2;
	      mh = h - frameMargY;

	      if (mw != MAIN.w || mh != MAIN.h)
	       { window_set (MAIN.id, 
		   WIN_WIDTH,	MAX(0,mw),
		   WIN_HEIGHT,	MAX(0,mh),
		   0);
		 window_set (CTRL.id,
		   WIN_X,	MAX(0,mw+2),
		   WIN_Y,  0,
		   0);

		 /* canvasResize doesn't seem to get called automatically */
		 canvasResize (MAIN.id, mw, mh);
	       }
	    }
	 }
	return (val);
}


static void canvasRepaint (canvas, pw, repaint_area)
Canvas canvas;
Pixwin *pw;
Rectlist *repaint_area;
{
	/* nothing to do, actually ... */
}

static void canvasResize (canvas, width, height)
Canvas canvas;
int width;
int height;
{
	int win;

	win = winIdToNum (canvas);
	wdesc[win].w = width;
	wdesc[win].h = height;

	if (canvas == MAIN.id)
	 { if (screenBuf)
	    { mem_destroy(screenBuf);
	    }
	   if (MAIN.w > 0)
	    { screenBuf = mem_create(MAIN.w,MAIN.h,1);
	    }
	 }
}

static int svGetEventFd ()
{
/*	notify_dispatch(); */
	return (winPipe[0]);
}

static void connectEvent (win, type, enable)
int win;
int type;
int enable;
{
	switch (type)
	 { default:
	    { grf4Fatal ("SV connectEvent: bogus type %d\n", type);
	    }
	   case GRF_MOTION_EVENT:
	   case GRF_BUTTON_EVENT:
	   case GRF_RESIZE_EVENT:
	   case GRF_REDRAW_EVENT:
	   case GRF_FOCUS_EVENT:
	    { break;		/* these are always enabled */
	    }
	   case GRF_KEY_EVENT:
	    { if (enable)
	       { window_set (wdesc[win].id, WIN_CONSUME_KBD_EVENT, 
					    WIN_ASCII_EVENTS, 0);
	       }
	      else
	       { window_set (wdesc[win].id, WIN_IGNORE_KBD_EVENT, 
					    WIN_ASCII_EVENTS, 0);
	       }
	      break;
	    }
	 }
}

static int svEnableEvent (win, type)
int win;
int type;
{
	if (wdesc[win].eventMask & (1<<type))		/* is already enabled */
	 { return (1);
	 }
	else
	 { eventRef[type]++;
	   if (eventRef[type] == 1)
	    { connectEvent (win, type, 1);
	    }
	   wdesc[win].eventMask |= (1<<type);
	   return (0);
	 }
}

static int svDisableEvent (win, type)
int win;
int type;
{
	if (!(wdesc[win].eventMask & (1<<type)))	/* is not enabled */
	 { return (0);
	 }
	else
	 { eventRef[type]--;
	   if (eventRef[type] == 0)
	    { connectEvent (win, type, 0);
	    }
	   if (eventRef[type] < 0)
	    { grf4Fatal ("svDisablebEvent: ref count for type %d < 0\n", type);
	    }
	   wdesc[win].eventMask &= ~(1<<type);
	   return (1);
	 }
}

static computeGrfCoords (xp, yp, win)
int *xp, *yp;
int win;
{
	int xframe, yframe;
	int xwin, ywin;
	int height;

	xframe = (int)window_get(frame, WIN_X);
	yframe = (int)window_get(frame, WIN_Y);
	xwin = (int)window_get(wdesc[win].id, WIN_X);
	ywin = (int)window_get(wdesc[win].id, WIN_Y);
	
	*xp = xframe + xwin;
	*yp = (screenHeight-wdesc[win].h) - (yframe+ywin);	
}

static int svPollEvent (win, code, ev)
int win;
int code;
Grf4Event *ev;
{
	int val1 = -1;
	int val2 = -1;

	switch (code)
	 { default:
	    { break;
	    }
	   case GRF_CURSOR:
	    { val1 = wdesc[win].cx;
	      val2 = wdesc[win].cy;
	      /* LATER: fix for NULL window, so we don't get -1 */
	      break;
	    }
	   case GRF_LEFT_MOUSE:
	    { val1 = (int)window_get(wdesc[win].id,WIN_EVENT_STATE,MS_LEFT);
	      break;
	    }
	   case GRF_MIDDLE_MOUSE:
	    { val1 = (int)window_get(wdesc[win].id,WIN_EVENT_STATE,MS_MIDDLE);
	      break;
	    }
	   case GRF_RIGHT_MOUSE:
	    { val1 = (int)window_get(wdesc[win].id,WIN_EVENT_STATE,MS_RIGHT);
	      break;
	    }
	   case GRF_FOCUS_CHANGE:
	    { val1 = wdesc[win].inFocus;
	      break;
	    }
	   case GRF_REDRAW:
	    { computeGrfCoords (&val1, &val2, win);
	      break;
	    }
	   case GRF_RESIZE:
	    { val1 = wdesc[win].w;
	      val2 = wdesc[win].h;
	      break;
	    }
	 }
	ev->code = code;
	ev->val1 = val1;
	ev->val2 = val2;
	return (1);
}

static int svReadEvent (ev)
Grf4Event *ev;
{
	Grf4Event desc;
	int n;
	
	n = read (winPipe[0], &desc, sizeof(desc));
	if (n == -1 && errno == EWOULDBLOCK)
	 { return (0);
	 }
	else if (n == sizeof(desc))
	 { ev->code = desc.code;
	   ev->val1 = desc.val1;
	   ev->val2 = desc.val2;
	   ev->win = desc.win;
	   return (1);
	 }
	else
	 { fprintf (stderr, "svReadEvent: incomplete read (%d bytes)\n", n);
	   if (n<0)
	    { perror("read");
	    }
	   return (0);
	 }
}

static int eventProc(wid, e, arg)
Canvas wid;
Event *e;
caddr_t arg;
{
	Grf4Event desc;
	int win;

	desc.code = GRF_NULLEVENT;
	desc.val1 = desc.val2 = -1;
	desc.win = win = winIdToNum (wid);

	switch (event_action(e))
	 { default:
	    { break;
	    }
	   case LOC_DRAG:
	   case LOC_MOVE:
	    { wdesc[win].cx = event_x(e);
	      wdesc[win].cy = (wdesc[win].h-1) - event_y(e);
	      if (wdesc[win].eventMask & GRF_MOTION_MASK)
	       { desc.val1 = wdesc[win].cx;
		 desc.val2 = wdesc[win].cy;
		 desc.code = GRF_CURSOR;
	       }
	      break;
	    }
	   case MS_LEFT:
	    { if (wdesc[win].eventMask & GRF_BUTTON_MASK)
	       { desc.val1 = event_is_down(e);
		 desc.code = GRF_LEFT_MOUSE;
	       }
	      break;
	    }
	   case MS_MIDDLE:
	    { if (wdesc[win].eventMask & GRF_BUTTON_MASK)
	       { desc.val1 = event_is_down(e);
		 desc.code = GRF_MIDDLE_MOUSE;
	       }
	      break;
	    }
	   case MS_RIGHT:
	    { if (wdesc[win].eventMask & GRF_BUTTON_MASK)
	       { desc.val1 = event_is_down(e);
		 desc.code = GRF_RIGHT_MOUSE;
	       }
	      break;
	    }
	   case WIN_RESIZE:
	    { desc.val1 = wdesc[win].w;
	      desc.val2 = wdesc[win].h;
	      desc.code = GRF_RESIZE;
	      break;
	    }
	   case WIN_REPAINT:
	    { computeGrfCoords (&desc.val1, &desc.val2, win);
	      desc.code = GRF_REDRAW;
	      break;
	    }
	   case LOC_RGNEXIT:
	    { wdesc[win].cx = event_x(e);
	      wdesc[win].cy = (wdesc[win].h-1) - event_y(e);
	      wdesc[win].inFocus = 0;
	      wdesc[NULLWIN].inFocus = 1;
	      if (wdesc[win].eventMask & GRF_FOCUS_MASK)
	       { desc.code = GRF_FOCUS_CHANGE;
		 desc.val1 = NULLWIN;
	       }
	      break;
	    }
	   case LOC_RGNENTER:
	    { wdesc[win].cx = event_x(e);
	      wdesc[win].cy = (wdesc[win].h-1) - event_y(e);
	      wdesc[win].inFocus = 1;
	      wdesc[NULLWIN].inFocus = 0;
	      if (wdesc[win].eventMask & GRF_FOCUS_MASK)
	       { desc.code = GRF_FOCUS_CHANGE;
		 desc.val1 = win;
	       }
	      break;
	    }
	 }
	if (desc.code != GRF_NULLEVENT)
	 { write (winPipe[1], &desc, sizeof(desc));
	 }
	return (1);
}

grf4SvSetupFxns (ftab)
struct grf4fxnTable *ftab;
{
	ftab->getWindowSize = svGetWindowSize;
	ftab->color = svColor;

	ftab->putText = svPutText;
	ftab->getTextExtent = svGetTextExtent;
	ftab->drawButton = svDrawButton;
	ftab->getButtonExtent = svGetButtonExtent;

	ftab->pxlDraw = svDraw;
	ftab->pxlPoint = svPoint;
	ftab->pxlFlush = svFlush;
	ftab->pxlClear = svClear;
	
	ftab->getEventFd = svGetEventFd;
	ftab->enableEvent = svEnableEvent;
	ftab->disableEvent = svDisableEvent;
	ftab->pollEvent = svPollEvent;
	ftab->readEvent = svReadEvent;

	return (1);
}

#else

#include <stdio.h>

grf4SvInit ()
{
	return(-1);
}

grf4SvSetupFxns()
{
	fprintf (stderr, "grf4Init(): Sunview support not available\n");
	return (-1);
}

#endif
