#ifdef X11_GRAPHICS

/* Check ~mouse/X/Xlib.doc */

#include <stdio.h>
#include <errno.h>
#include <ctype.h>	
#include <sys/types.h>
#include <sys/time.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xresource.h>
#include <X11/keysym.h>

#include "../h/cdefs.h"
#include "../h/timeval.h"
#include "../h/grf4.h"
#include "grf4Internal.h"

extern void *malloc();
char *getenv();

static XrmDatabase db;
static char *defaults = "\
*Foreground: white\n\
*Background: black\n\
*BorderColor: white\n\
*BorderWidth: 1\n\
*BorderMargin: 0\n\
*Font: fixed\n\
*Name: robotsim\n\
*IconName: robotsim\n\
";

static char *displayname;
static char *geometryspec;
static char *foreground;
static char *background;
static char *bordercstr;
static char *borderwstr;
static char *bordermstr;
static char *margincstr;
static char *name;
static char *iconname;
static char *fontname;

static int argc;
static char **argv;

static Display *disp;
static Screen *scr;
static int width;
static int height;
static int depth;
static Window rootwin;
static Colormap defcmap;
static Visual *visual;

static XColor fgcolor;
static XColor bgcolor;
static XColor bdcolor;
static XColor mgcolor;

static int reverse = 0;

static GC wingc;
static GC revgc;
static Window topwin;
static Colormap wincmap;
static int topx;
static int topy;
static int topw;
static int toph;
static int margin;
static int borderwidth;
static XFontStruct *font;
static long eventMask;

static Pixmap pixmap = None;

static saveargv();
static handleargs();
static maybeset();
static setup_db();
static char *get_default_value();
static char *defaultValue();
static setup_color();
static setup_font();
static setup_colors();
static setup_numbers();
static setup_windows();
static setup_pixmap();
static resize();
static redisplay();

static void x11PutText();
static void x11GetTextExtent();
static void x11GetButtonExtent();
static void x11DrawButton();
static void x11Clear();
static void x11Draw();
static void x11Point();
static void x11Flush();
static void x11GetWindowSize();
static void x11Color();
static int x11GetEventFd();
static int x11EnableEvent();
static int x11DisableEvent();
static int x11PollEvent();
static int x11ReadEvent();
static void x11InitDeviceMaps();

static GC topWinGc;
static GC panWinGc;
static GC panRevGc;

typedef struct {
	Window id;	/* id returned from XCreateWindow() */
	long x, y;	/* position of the window (WRT topwin) */
	long w, h;	/* height and width */
	int active;	/* true if the window is active */
	unsigned long eventMask; /* selected events */
} WinDesc;

#if 0
#define TEXTREQ_BSIZE	16

typedef struct textReq {
	struct textReq *next;
	char *str;
	int x, y;
	int len;
	char buf[TEXTREQ_BSIZE];
} TextReq;

TextReq *textReqList = NULL;
TextReq *textReqFree = NULL;
#endif

static WinDesc wdesc[NUMWINS];

#define MAIN	wdesc[MAINWIN]
#define CTRL	wdesc[CTRLWIN]

static saveargv(ac,av)
int ac;
char **av;
{
	int i;
	int nc;
	char *abuf;

	argc = ac;
	argv = (char **)malloc((ac+1)*sizeof(char *));
	nc = 1;
	for (i=0;i<ac;i++)
	 { nc += strlen(av[i]) + 1;
	 }
	abuf = (char *) malloc(nc*sizeof(char));
	for (i=0;i<ac;i++)
	 { argv[i] = abuf;
	   strcpy(abuf,av[i]);
	   while (*abuf) abuf ++;
	   abuf ++;
	 }
	*abuf = '\0';
	argv[ac] = 0;
}

static handleargs(ac,av)
int ac;
char **av;
{
	int skip;
	int errs;

	skip = 0;
	errs = 0;
	for (ac--,av++;ac;ac--,av++)
	 { if (skip > 0)
	    { skip --;
	      continue;
	    }
	   if (**av != '-')
	    { fprintf(stderr,"unrecognized argument `%s'\n",*av);
	      errs ++;
	      continue;
	    }
	   if (0)
	    {
needarg:;
	      fprintf(stderr,"%s needs a following argument\n",*av);
	      errs ++;
	      continue;
	    }
#define WANTARG() do { if (++skip >= ac) goto needarg; } while (0)
	   if (!strcmp(*av,"-display"))
	    { WANTARG();
	      displayname = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-geometry"))
	    { WANTARG();
	      geometryspec = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-foreground") || !strcmp(*av,"-fg"))
	    { WANTARG();
	      foreground = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-background") || !strcmp(*av,"-bg"))
	    { WANTARG();
	      background = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-bordercolor") || !strcmp(*av,"-bd"))
	    { WANTARG();
	      bordercstr = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-borderwidth") || !strcmp(*av,"-bw"))
	    { WANTARG();
	      borderwstr = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-bordermargin") || !strcmp(*av,"-bm"))
	    { WANTARG();
	      bordermstr = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-margincolor") || !strcmp(*av,"-mc"))
	    { WANTARG();
	      margincstr = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-name"))
	    { WANTARG();
	      name = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-iconname"))
	    { WANTARG();
	      iconname = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-font") || !strcmp(*av,"-fn"))
	    { WANTARG();
	      fontname = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-reverse"))
	    { reverse = 1;
	      continue;
	    }
#undef WANTARG
	   fprintf(stderr,"unrecognized option `%s'\n",*av);
	   errs ++;
	 }
	if (errs)
	 { exit(1);
	 }
}

static maybeset(strp,str)
char **strp;
char *str;
{
	if (str && !*strp) *strp = str;
}

static setup_db()
{
	char *str;
	char *home;
	XrmDatabase db2;

	db = XrmGetStringDatabase(defaults);
	str = XResourceManagerString(disp);
	if (str)
	 { db2 = XrmGetStringDatabase(str);
	   XrmMergeDatabases(db2,&db);
	 }
	else
	 { home = getenv("HOME");
	   if (home)
	    { str = (char*)malloc(strlen(home)+1+10+1);
	      sprintf(str,"%s/.Xdefaults",home);
	      db2 = XrmGetFileDatabase(str);
	      if (db2)
	       { XrmMergeDatabases(db2,&db);
	       }
	      free(str);
	    }
	 }
}

static char *get_default_value(name,class)
char *name;
char *class;
{
	char *type;
	XrmValue value;

	if (XrmGetResource(db,name,class,&type,&value) == False) return(0);
	return(value.addr);
}

static char *defaultValue(name,prop,Name,class)
char *name;
char *prop;
char *Name;
char *class;
{
	char nbuf[256];
	char cbuf[256];

	sprintf (nbuf, "%s.%s", name, prop);
	sprintf (cbuf, "%s.%s", Name, class);
	return (get_default_value(nbuf,cbuf));
}

static setup_color(str,col)
char *str;
XColor *col;
{
	if (XParseColor(disp,wincmap,str,col) == 0)
	 { fprintf(stderr,"bad color `%s'\n",str);
	   exit(1);
	 }
	while (1)
	 { if (XAllocColor(disp,wincmap,col) == 0)
	    { if (wincmap != defcmap)
	       { fprintf(stderr,"can't allocate colormap cell for color `%s'\n",str);
		 exit(1);
	       }
	      wincmap = XCopyColormapAndFree(disp,wincmap);
	      continue;
	    }
	   break;
	 }
}

static setup_font()
{
	font = fontname ? XLoadQueryFont(disp,fontname) : 0;
	if (fontname && !font)
	 { fprintf(stderr,"can't load font %s, using server default\n",fontname);
	 }
}

static setup_colors()
{
	wincmap = defcmap;
	setup_color(foreground,&fgcolor);
	setup_color(background,&bgcolor);
	setup_color(bordercstr,&bdcolor);
	setup_color(margincstr,&mgcolor);
	if (reverse)
	 { XColor ct;
	   ct = fgcolor;
	   fgcolor = bgcolor;
	   bgcolor = ct;
	 }
}

static setup_numbers()
{
	if (bordermstr) margin = atoi(bordermstr);
	if (borderwstr) borderwidth = atoi(borderwstr);
}

static setup_windows(Name,name,x,y,w,h,minW,minH)
char *Name;
char *name;
int x,y;
unsigned int w,h;
int minW, minH;
{
	int bits;
	unsigned long int attrmask;
	unsigned long int gcvalmask;
	XSetWindowAttributes attr;
	XWindowAttributes aget;
	XGCValues gcval;
#ifdef X11R3
	XSizeHints sizehints;
#else
	XTextProperty wn_prop;
	XTextProperty in_prop;
	XSizeHints normal_hints;
	XWMHints wm_hints;
	XClassHint class_hints;
#endif

	minW = MAX(minW,1);
	minH = MAX(minH,1);

	bits = XParseGeometry(geometryspec,&x,&y,&w,&h);
	if (bits & XNegative) x = width + x - w;
	if (bits & YNegative) y = height + y - h;
	attrmask = 0;
	attr.background_pixel = mgcolor.pixel;
	attrmask |= CWBackPixel;
	attr.border_pixel = bdcolor.pixel;
	attrmask |= CWBorderPixel;
	attr.backing_store = NotUseful;
	attrmask |= CWBackingStore;
	attr.event_mask = StructureNotifyMask|ExposureMask;
	attrmask |= CWEventMask;
	w -= 2 * borderwidth;
	h -= 2 * borderwidth;
	topwin = XCreateWindow(disp,rootwin,x,y,w,h,borderwidth,depth,InputOutput,CopyFromParent,attrmask,&attr);
#ifdef X11R3
	sizehints.flags = 0;
	if (CTRL.active)
	 { sizehints.min_width = (CTRL.w+1) + minW;
	   sizehints.min_height = MAX(minH,CTRL.h);
	 }
	else
	 { sizehints.min_width = minW;
	   sizehints.min_height = minH;
	 }
	sizehints.flags |= PMinSize;
	sizehints.width = w;
	sizehints.height = h;
	sizehints.flags |= (bits & (WidthValue|HeightValue)) ? USSize : PSize;
	if (bits & (XValue|YValue))
	 { sizehints.x = x;
	   sizehints.y = y;
	   sizehints.flags |= USPosition;
	 }
	XSetStandardProperties(disp,topwin,Name,name,None,argv,argc,&sizehints);
#else	
	wn_prop.value = (unsigned char *)Name;
	wn_prop.encoding = XA_STRING;
	wn_prop.format = 8;
	wn_prop.nitems = strlen((char *)wn_prop.value);
	in_prop.value = (unsigned char *)name;
	in_prop.encoding = XA_STRING;
	in_prop.format = 8;
	in_prop.nitems = strlen((char *)in_prop.value);
	normal_hints.flags = PMinSize | PResizeInc;
	normal_hints.x = x;
	normal_hints.y = y;
	normal_hints.flags |= (bits & (XValue|YValue)) ? USPosition : PPosition;
	normal_hints.width = w;
	normal_hints.height = h;
	normal_hints.flags |= (bits & (WidthValue|HeightValue)) ? USSize : PSize;
	if (CTRL.active)
	 { normal_hints.min_width = (CTRL.w+1) + minW;
	   normal_hints.min_height = MAX(minH,CTRL.h);
	 }
	else
	 { normal_hints.min_width = minW;
	   normal_hints.min_height = minH;
	 }
	normal_hints.width_inc = 1;
	normal_hints.height_inc = 1;
	wm_hints.flags = InputHint;
	wm_hints.input = True;
	class_hints.res_name = name;
	class_hints.res_class = Name;
	XSetWMProperties(disp,topwin,&wn_prop,&in_prop,argv,argc,&normal_hints,&wm_hints,&class_hints);
#endif
	if (wincmap != defcmap) XSetWindowColormap(disp,topwin,wincmap);
	attrmask = 0;
	attr.background_pixel = bgcolor.pixel;
	attrmask |= CWBackPixel;
	attr.backing_store = NotUseful;
	attrmask |= CWBackingStore;
	eventMask = 0;
/*	      ButtonMotionMask | ButtonPressMask |  ButtonReleaseMask); */
	attr.event_mask = eventMask;
	attrmask |= CWEventMask;
	MAIN.id = XCreateWindow(disp,topwin,0,0,1,1,0,depth,InputOutput,
		     CopyFromParent,attrmask,&attr);
	XMapRaised(disp,MAIN.id);
	gcvalmask = 0;
	if (font)
	 { gcval.font = font->fid;
	   gcvalmask |= GCFont;
	 }
	gcval.foreground = fgcolor.pixel;
	gcvalmask |= GCForeground;
	gcval.background = bgcolor.pixel;
	gcvalmask |= GCBackground;
	wingc = XCreateGC(disp,MAIN.id,gcvalmask,&gcval);
	revgc = XCreateGC(disp,MAIN.id,gcvalmask,&gcval);
	XSetForeground(disp,revgc,bgcolor.pixel);
	XSetBackground(disp,revgc,fgcolor.pixel);
	XMapRaised(disp,topwin);
	topWinGc = XCreateGC(disp,topwin,gcvalmask,&gcval);

 	if (CTRL.active)
	 { long mask;

	   CTRL.id = XCreateWindow(disp,topwin,0,0,1,1,0,depth,InputOutput,
				  CopyFromParent,attrmask,&attr);
	   XMapRaised(disp,CTRL.id);
	   panWinGc = XCreateGC(disp,CTRL.id,gcvalmask,&gcval);
	   panRevGc = XCreateGC(disp,CTRL.id,gcvalmask,&gcval);
	   XSetForeground(disp,panRevGc,bgcolor.pixel);
	   XSetBackground(disp,panRevGc,fgcolor.pixel);
	   mask = attr.event_mask;
	   mask |= EnterWindowMask|LeaveWindowMask;
	   mask |= ButtonPressMask|ButtonReleaseMask;
	   mask |= PointerMotionMask;
	   XSelectInput (disp, CTRL.id, mask);
	 }
	MAIN.x = margin;	/* this won't change */
	MAIN.y = margin;

	XGetWindowAttributes (disp, topwin, &aget);
	topx = aget.x;
	topy = aget.y;
	topw = w;
	toph = h;

	resize(w,h);
}

static setup_pixmap()
{
	if (pixmap != None)
	 { XFreePixmap (disp, pixmap);
	 }
	if (MAIN.w > 0)
	 { pixmap = XCreatePixmap (disp, MAIN.id, MAIN.w, MAIN.h, depth);
	   XSetForeground(disp,wingc,bgcolor.pixel);
	   XFillRectangle(disp,pixmap,wingc,0,0,MAIN.w,MAIN.h);
	   XSetForeground(disp,wingc,fgcolor.pixel);
	 }
	else
	 { pixmap = None;
	 }
}

#define BUTTON_BORDER_WIDTH	2


static void x11PutText (str, x, y)
char *str;
int x, y;
{
#if 0
	TextReq *req;

	if ((req = textReqFree) == NULL)
	 { req = (TextReq*)malloc(sizeof(TextReq));
	 }
	else
	 { textReqFree = req->next;
	 }
	req->next = textReqList;
	textReqList = req;
	req->x = x;
	req->y = y;
	req->len = strlen(str);
	if (req->len > TEXTREQ_BSIZE-1)
	 { req->str = (char*)malloc(req->len+1);
	 }
	else
	 { req->str = req->buf;
	 }
	strcpy (req->str, str);
#else
	XDrawImageString(disp, pixmap, topWinGc,
			 x, MAIN.h-1-y, str, strlen(str));
#endif
}

static void x11GetTextExtent (xp, yp, wp, hp, str)
int *xp, *yp, *wp, *hp;
char *str;
{
	int direction;
	int fontAscent;
	int fontDescent;
	XCharStruct cs;

	XTextExtents (font, str, strlen(str), 
		      &direction,
		      &fontAscent,
		      &fontDescent,
		      &cs);
	if (xp) *xp = -cs.lbearing;
	if (yp) *yp = MAX(cs.descent,fontDescent);
	if (wp) *wp = cs.rbearing-cs.lbearing;
	if (hp) *hp = MAX(cs.ascent,fontAscent) + MAX(cs.descent,fontDescent);
}

static void x11GetButtonExtent(b)
BUTTON_DESC *b;
{
	int x, y, w, h;
	int margin;

	x11GetTextExtent (&x, &y, &w, &h, b->name);
	margin = 1 + BUTTON_BORDER_WIDTH;
	b->coffX = margin + x;
	b->coffY = margin + y;
	b->minw = 2*margin + w;
	b->minh = 2*margin + h;
}

static void x11DrawButton(b)
BUTTON_DESC *b;
{
	int baseX, baseY;
	int cbaseX, cbaseY;

	baseX = b->x;
	baseY = (CTRL.h-1) - b->y - b->h + 1;
	cbaseX = baseX + b->coffX + (b->w - b->minw)/2;
	cbaseY = (CTRL.h-1) - b->y - b->coffY - (b->h - b->minh)/2;

	if (!(b->flags & BF_NOTBOXED))
	 { XFillRectangle(disp,CTRL.id,panWinGc,baseX,baseY,b->w,b->h);
	 }
	else
	 { XFillRectangle(disp,CTRL.id,panRevGc,baseX,baseY,b->w,b->h);
	 }
	if (b->flags & BF_ILLUMINATED)
	 { XFillRectangle (disp,CTRL.id,panWinGc,
			     baseX+BUTTON_BORDER_WIDTH,
			     baseY+BUTTON_BORDER_WIDTH,
			     b->w-2*BUTTON_BORDER_WIDTH,
			     b->h-2*BUTTON_BORDER_WIDTH);
	   XDrawImageString(disp, CTRL.id, panRevGc, cbaseX, cbaseY,
			    b->name, strlen(b->name));
	 }
	else
	 { XFillRectangle (disp,CTRL.id,panRevGc,
			     baseX+BUTTON_BORDER_WIDTH,
			     baseY+BUTTON_BORDER_WIDTH,
			     b->w-2*BUTTON_BORDER_WIDTH,
			     b->h-2*BUTTON_BORDER_WIDTH);
	   XDrawImageString(disp, CTRL.id, panWinGc, cbaseX, cbaseY,
			    b->name, strlen(b->name));
	 }
}

static resize(tw,th)
int tw;
int th;
{
	int x, y, w, h;

	x = MAIN.x;
	y = MAIN.y;

	w = tw - (2*margin);
	h = th - (2*margin);
	if (CTRL.active)
	 { w -= (CTRL.w+1);	/* extra 1 to give space for dividing line */
	 }
	XMoveResizeWindow(disp,MAIN.id,x,y,w,h);
	XClearArea(disp,MAIN.id,0,0,0,0, False/*(no "expose" event)*/);
	MAIN.w = w;
	MAIN.h = h;
	if (CTRL.active)
	 { XMoveResizeWindow(disp,CTRL.id,x+w+1,y,CTRL.w,CTRL.h);
	   XClearArea(disp,CTRL.id,0,0,0,0, False/*(no "expose" event)*/);
	   CTRL.x = x+w+1;
	   CTRL.y = y;
	 }
	setup_pixmap();
}

static redisplay(x,y,w,h)
int x;
int y;
int w;
int h;
{
#if 0
	TextReq *req, *nextReq;
#endif
	if (pixmap != None && w > 0)
	 { XCopyArea (disp, pixmap, MAIN.id, wingc, x, y, w, h, x, y);
	 }
#if 0
	for (req=textReqList; req; req=nextReq)
	 { nextReq = req->next;
	   XDrawImageString(disp, MAIN.id, panWinGc,
			 req->x, (MAIN.h-1-req->y), req->str, req->len);
	   if (req->str != req->buf)
	    { free (req->str);
	    }
	   req->next = textReqFree;
	   textReqFree = req;
	 }
	textReqList = NULL;
#endif
}

static void x11Clear (x,y,w,h)
int x, y, w, h;
{
	/* Clear the memory image prior to drawing */

	if (pixmap != None)
	 { XSetForeground(disp,wingc,bgcolor.pixel);
	   XFillRectangle(disp,pixmap,wingc,x,y,w,h);
	   XSetForeground(disp,wingc,fgcolor.pixel);
	 }
}

static void x11Draw (x1,y1,x2,y2)
int x1,y1,x2,y2;
{
	/* Draw a line into the memory image */

	if (pixmap != None)
	 { XDrawLine(disp,pixmap,wingc,x1,y1,x2,y2);
	 }
}

static void x11Point (x1,y1)
int x1,y1;
{
	/* Draw a point into the memory image */

	if (pixmap != None)
	 { XDrawPoint(disp,pixmap,wingc,x1,y1);
	 }
}

static void x11Flush (x, y, w, h)
int x, y, w, h;
{
	/* Flush the memory image to the output image. */

	redisplay (x, y, w, h);
	XFlush (disp);
}

static void x11GetWindowSize (w, h, directionOfY)
int *w, *h;
int *directionOfY;
{
	*w = MAIN.w;
	*h = MAIN.h;
	*directionOfY = -1;
}

static void x11Color(color)
unsigned long color;
{
}

grf4X11Init (label, fgcolor, bgcolor, placing, flags)
char *label;
int fgcolor;
int bgcolor;
WinPlaceDesc *placing;
int flags;
{
	int ac = 1;
	char *dummyProgName = "dummyProgName";
	char **av = &dummyProgName;

	char Nbuf[256];
	char nbuf[256];

	int x, y, w, h;

	strcpy(nbuf, label);
	strcpy(Nbuf, label);
	nbuf[0] = tolower(nbuf[0]);
	Nbuf[0] = toupper(Nbuf[0]);

	saveargv(ac,av);
	handleargs(ac,av);
	if ((disp = XOpenDisplay(displayname)) == 0)
	 { return (-1);
	 }

	scr = XDefaultScreenOfDisplay(disp);
	width = XWidthOfScreen(scr);
	height = XHeightOfScreen(scr);
	depth = XDefaultDepthOfScreen(scr);
	rootwin = XRootWindowOfScreen(scr);
	defcmap = XDefaultColormapOfScreen(scr);
	visual = XDefaultVisualOfScreen(scr);
	setup_db();

	maybeset(&geometryspec, defaultValue(
		 nbuf, "geometry", Nbuf, "Geometry"));
	maybeset(&fontname, defaultValue(
		 nbuf, "font", Nbuf, "Font"));
	maybeset(&foreground, defaultValue(
		 nbuf, "foreground", Nbuf, "Foreground"));
	maybeset(&background, defaultValue(
		 nbuf, "background", Nbuf, "Background"));
	maybeset(&bordercstr, defaultValue(
		 nbuf, "borderColor", Nbuf, "BorderColor"));
	maybeset(&borderwstr, defaultValue(
		 nbuf, "borderWidth", Nbuf, "BorderWidth"));
	maybeset(&bordermstr, defaultValue(
		 nbuf, "borderMargin", Nbuf, "BorderMargin"));
	maybeset(&margincstr, defaultValue(
		 nbuf, "marginColor", Nbuf, "Background"));
	maybeset(&name, defaultValue(
		 nbuf, "name", Nbuf, "Name"));
	maybeset(&iconname, defaultValue(
		 nbuf, "iconName", Nbuf, "IconName"));

	setup_font();
	if (bgcolor)
	 { reverse = 1;
	 }
	setup_colors();
	setup_numbers();

	/* set up the null window with info appropriate to the screen */

	wdesc[NULLWIN].x = 0;
	wdesc[NULLWIN].y = 0;
	wdesc[NULLWIN].w = width;
	wdesc[NULLWIN].h = height;

	buttonInit (&CTRL.w, &CTRL.h);
	CTRL.active = (CTRL.w > 0);

	if (!placing->placed)
	 { x = y = 200;
	   w = h = 300;
	 }
	else
	 { x = placing->x;
	   y = placing->y;
	   w = placing->w;
	   h = placing->h;
	 }

	w = MAX(w,placing->minW);
	h = MAX(h,placing->minH);

	setup_windows(Nbuf,nbuf,x,y,
		      2*margin+w+CTRL.w,2*margin+MAX(h,CTRL.h),
		      placing->minW, placing->minH);

	if (CTRL.active)
	 { XDrawLine(disp,topwin,topWinGc,MAIN.x+MAIN.w,0,
					  MAIN.x+MAIN.w,toph-1);
	 }
	x11InitDeviceMaps();
/*	XAutoRepeatOff (disp); 	This seems to be sticky, at least on SGI's */
	return (1);
}

grf4X11SetupFxns (ftab)
struct grf4fxnTable *ftab;
{
	ftab->getWindowSize = x11GetWindowSize;
	ftab->color = x11Color;

	ftab->getTextExtent = x11GetTextExtent;
	ftab->drawButton = x11DrawButton;
	ftab->getButtonExtent = x11GetButtonExtent;

	ftab->pxlPutText = x11PutText;
	ftab->pxlDraw = x11Draw;
	ftab->pxlPoint = x11Point;
	ftab->pxlFlush = x11Flush;
	ftab->pxlClear = x11Clear;

	ftab->getEventFd = x11GetEventFd;
	ftab->enableEvent = x11EnableEvent;
	ftab->disableEvent = x11DisableEvent;
	ftab->pollEvent = x11PollEvent;
	ftab->readEvent = x11ReadEvent;

	return (1);
}

static AlistElem alistBase[] = { 
	GRF_KB_A,		XK_a,
	GRF_KB_B,		XK_b,
	GRF_KB_C,		XK_c,
	GRF_KB_D,		XK_d,
	GRF_KB_E,		XK_e,
	GRF_KB_F,		XK_f,
	GRF_KB_G,		XK_g,
	GRF_KB_H,		XK_h,
	GRF_KB_I,		XK_i,
	GRF_KB_J,		XK_j,
	GRF_KB_K,		XK_k,
	GRF_KB_L,		XK_l,
	GRF_KB_M,		XK_m,
	GRF_KB_N,		XK_n,
	GRF_KB_O,		XK_o,
	GRF_KB_P,		XK_p,
	GRF_KB_Q,		XK_q,
	GRF_KB_R,		XK_r,
	GRF_KB_S,		XK_s,
	GRF_KB_T,		XK_t,
	GRF_KB_U,		XK_u,
	GRF_KB_V,		XK_v,
	GRF_KB_W,		XK_w,
	GRF_KB_X,		XK_x,
	GRF_KB_Y,		XK_y,
	GRF_KB_Z,		XK_z,
	GRF_KB_0,		XK_0,
	GRF_KB_1,		XK_1,
	GRF_KB_2,		XK_2,
	GRF_KB_3,		XK_3,
	GRF_KB_4,		XK_4,
	GRF_KB_5,		XK_5,
	GRF_KB_6,		XK_6,
	GRF_KB_7,		XK_7,
	GRF_KB_8,		XK_8,
	GRF_KB_9,		XK_9,
	GRF_KB_Break,		XK_Break,
	GRF_KB_Ctrl,		XK_Control_L,
	GRF_KB_CapsLock,	XK_Caps_Lock,
	GRF_KB_RightShift,	XK_Shift_R,
	GRF_KB_LeftShift,	XK_Shift_L,
	GRF_KB_Esc,		XK_Escape,
	GRF_KB_Tab,		XK_Tab,
	GRF_KB_Ret,		XK_Return,
	GRF_KB_Space,		XK_space,
	GRF_KB_Linefeed,	XK_Linefeed,
	GRF_KB_Backspace,	XK_BackSpace,
	GRF_KB_Del,		XK_Delete,
	GRF_KB_Semicolon,	XK_semicolon,
	GRF_KB_Period,		XK_period,
	GRF_KB_Comma,		XK_comma,
	GRF_KB_Quote,		XK_apostrophe,
	GRF_KB_Backquote,	XK_grave,
	GRF_KB_Minus,		XK_minus,
	GRF_KB_Slash,		XK_slash,
	GRF_KB_Backslash,	XK_backslash,
	GRF_KB_Equal,		XK_equal,
	GRF_KB_LeftBracket,	XK_bracketleft,
	GRF_KB_RightBracket,	XK_bracketright,
	GRF_KB_LeftArrow,	XK_Left,
	GRF_KB_DownArrow,	XK_Right,
	GRF_KB_RightArrow,	XK_Up,
	GRF_KB_UpArrow,		XK_Down,
	GRF_KB_Pad0,		XK_KP_0,
	GRF_KB_Pad1,		XK_KP_1,
	GRF_KB_Pad2,		XK_KP_2,
	GRF_KB_Pad3,		XK_KP_3,
	GRF_KB_Pad4,		XK_KP_4,
	GRF_KB_Pad5,		XK_KP_5,
	GRF_KB_Pad6,		XK_KP_6,
	GRF_KB_Pad7,		XK_KP_7,
	GRF_KB_Pad8,		XK_KP_8,
	GRF_KB_Pad9,		XK_KP_9,
	GRF_KB_PadF1,		XK_KP_F1,
	GRF_KB_PadF2,		XK_KP_F2,
	GRF_KB_PadF3,		XK_KP_F3,
	GRF_KB_PadF4,		XK_KP_F4,
	GRF_KB_PadPeriod,	XK_KP_Decimal,
	GRF_KB_PadMinus,	XK_KP_Subtract,
	GRF_KB_PadComma,	XK_KP_Separator,
	GRF_KB_PadEnter,	XK_KP_Enter,
	GRF_KB_LeftAlt,		XK_Alt_L,
	GRF_KB_RightAlt,	XK_Alt_R,
	GRF_KB_RightCtrl,	XK_Control_R,
	GRF_KB_F1,		XK_F1,
	GRF_KB_F2,		XK_F2,
	GRF_KB_F3,		XK_F3,
	GRF_KB_F4,		XK_F4,
	GRF_KB_F5,		XK_F5,
	GRF_KB_F6,		XK_F6,
	GRF_KB_F7,		XK_F7,
	GRF_KB_F8,		XK_F8,
	GRF_KB_F9,		XK_F9,
	GRF_KB_F10,		XK_F10,
	GRF_KB_F11,		XK_F11,
	GRF_KB_F12,		XK_F12,
	GRF_KB_PrintScreen,	XK_Print,			  
	GRF_KB_ScrollLock,	XK_Scroll_Lock,
	GRF_KB_Pause,		XK_Pause,
	GRF_KB_Insert,		XK_Insert,
	GRF_KB_Home,		XK_Home,
	GRF_KB_PageUp,		XK_Prior,
	GRF_KB_End,		XK_End,
	GRF_KB_PageDown,	XK_Next,
	GRF_KB_NumLock,		XK_Num_Lock,
	GRF_KB_PadDivide,	XK_KP_Divide,
	GRF_KB_PadMultiply,	XK_KP_Multiply,
	GRF_KB_PadPlus,		XK_KP_Add,
};

static unsigned char keyToCodeMap[256];
static unsigned char codeToKeyMap[256];

static void x11InitDeviceMaps()
{
	int i, key, code;

	for (i=0; i<ARRAY_SIZE(alistBase); i++)
	 { key = alistBase[i].left;
	   code = XKeysymToKeycode(disp, alistBase[i].right);
	   if (code < 0 || code > 255 || key <= 0 || key > 255)
	    { printf 
("Warning, X11 init, range troubke with key %d, code=%d\n", key, code);
	    }
	   else
	    { keyToCodeMap[key] = code;
	      if (code > 0)
	       { codeToKeyMap[code] = key;
	       }
	    }
	 }
}

static int x11GetEventFd ()
{
	return (XConnectionNumber(disp));
}

static unsigned long getX11EventMask(type)
{
	switch (type)
	 { default:
	    { grf4Fatal ("X11 getEventMask(): bougs type (%d)\n", type);
	    }
	   case GRF_KEY_EVENT:
	    { return (KeyPressMask|KeyReleaseMask);
	    }
	   case GRF_BUTTON_EVENT:
	    { return (ButtonPressMask|ButtonReleaseMask);
	    }
	   case GRF_MOTION_EVENT:
	    { return (PointerMotionMask);
	    }
	   case GRF_FOCUS_EVENT:
	    { return (EnterWindowMask|LeaveWindowMask);
	    }
	   case GRF_RESIZE_EVENT:
	   case GRF_REDRAW_EVENT:
	    { return (0);	/* these are always enabled */
	    }
	 }
}

static int x11EnableEvent (win, type)
int win;
int type;
{
	unsigned long mask;

	mask = getX11EventMask(type);
	if (mask & wdesc[win].eventMask)
	 { return (1);
	 }
	else
	 { wdesc[win].eventMask |= mask;
	   XSelectInput (disp, wdesc[win].id, wdesc[win].eventMask);
	   return (0);
	 }
}

static int x11DisableEvent (win, type)
int win;
int type;
{
	unsigned long mask;

	mask = getX11EventMask(type);
	if (!(mask & wdesc[win].eventMask))
	 { return (0);
	 }
	else
	 { wdesc[win].eventMask &= ~mask;
	   XSelectInput (disp, wdesc[win].id, wdesc[win].eventMask);
	   return (1);
	 }
}

static int winIdToNum (wid)
Window wid;
{
	if (wid == MAIN.id)
	 { return (MAINWIN);
	 }
	else if (wid == CTRL.id)
	 { return (CTRLWIN);
	 }
	else
	 { return (NULLWIN);
	 }
}

static int x11PollEvent (win, code, ev)
int win;
int code;
Grf4Event *ev;
{
	int val1 = -1;
	int val2 = -1;

	/* Return values from XQueryPointer */
	Window rootWin, childWin;
	int rx, ry, x, y;
	unsigned int mask;
	
	if (GRF_IS_KEY(code))
	 { unsigned char code;
	   char map[32];

	   code = keyToCodeMap[code];
	   if (code == 0)
	    { return (-1);
	    }
	   XQueryKeymap(disp, map);
	   val1 = map[code/8] & (1<<(code%8));		 
	 }
	else
	 { Window wid;

	   wid = (win==NULLWIN) ? MAIN.id : wdesc[win].id;

	   XQueryPointer (disp, wid, &rootWin, &childWin, 
				&rx, &ry, &x, &y, &mask);
	   
	   switch (code)
	    { default:
	       { return (-1);
		 break;
	       }
	      case GRF_LEFT_MOUSE:
	       { val1 = mask & Button1Mask;
		 break;
	       }
	      case GRF_MIDDLE_MOUSE:
	       { val1 = mask & Button2Mask;
		 break;
	       }
	      case GRF_RIGHT_MOUSE:
	       { val1 = mask & Button3Mask;
		 break;
	       }
	      case GRF_CURSOR:
	       { if (win == NULLWIN)
		  { val1 = rx;
		    val2 = (height-1)-ry;
		  }
		 else
		  { val1 = x;
		    val2 = (wdesc[win].h-1)-y;
       		  }
		 break;
	       }
	      case GRF_REDRAW:
	       { if (win == NULLWIN)
		  { val1 = 0;
		    val2 = 0;
		  }
		 else
		  { XWindowAttributes aget;
		 
		    XGetWindowAttributes (disp, topwin, &aget);
		    val1 = aget.x+wdesc[win].x;
		    val2 = (height-wdesc[win].h) - (aget.y+wdesc[win].y);
		  }
		 break;
	       }
	      case GRF_RESIZE:
	       { val1 = wdesc[win].w;
		 val2 = wdesc[win].h;
		 break;
	       }
	      case GRF_FOCUS_CHANGE:
	       { val1 = winIdToNum(childWin);
		 break;
	       }
	    }
	 }
	ev->val1 = val1;
	ev->val2 = val2;
	ev->code = code;
	return (1);
}

/* The event stack is used to insert new events where necessary */

#define EVSTACKSIZE	10

static Grf4Event evStack[EVSTACKSIZE];
static int evStackp = 0;

static int pushEvent (code, val1, val2, win)
int code, val1, val2, win;
{
	if (evStackp < EVSTACKSIZE)
	 { evStack[evStackp].code = code;
	   evStack[evStackp].val1 = val1;
	   evStack[evStackp].val2 = val2;
	   evStack[evStackp].win  = win;
	   evStackp++;
	   return (1);
	 }
	else
	 { return (0);
	 }
}

static int popEvent (ev)
Grf4Event *ev;
{
	if (evStackp)
	 { *ev = evStack[--evStackp];
	   return (1);
	 }
	else
	 { return (0);
	 }
}

static int x11ReadEvent(ev)
Grf4Event *ev;
{
	int code = GRF_NULLEVENT;
	int val1 = -1;
	int val2 = -1;
	int win;
	XEvent e, eNext;

	if (popEvent (ev))
	 { return (1);
	 }
	XSync(disp,False);
	if (XQLength(disp) <= 0)
	 { return (0);
	 }
	XNextEvent (disp, &e);

	win = winIdToNum(e.xany.window);

	switch (e.type)
	 { default:
	    { break;
	    }
	   case ConfigureNotify:
	    { int tw, th;	/* width and height of top window */
	      int w, h, x, y;	/* GRF4-relative coords for main window */
	      int doResize = 0;

	      /* absorb following expose events */

	      while (XQLength(disp) > 0)
	       { XPeekEvent (disp, &eNext);
		 if (eNext.xany.window != e.xany.window ||
		     (eNext.type != Expose && 
		      eNext.type != GraphicsExpose) )		     
		  { break;
		  }
		 XNextEvent (disp, &eNext);
	       }
	      tw = e.xconfigure.width;
	      th = e.xconfigure.height;
	       x = e.xconfigure.x+margin;
	       y = (height-MAIN.h) - (e.xconfigure.y+margin);
	       w = tw - 2*margin - CTRL.w;
	       h = th - 2*margin;

	      if (tw != topw || th != toph)
	       { /* resize event */

		 resize (tw, th);
		 doResize = 1;
	       }

	      topw = tw;
	      toph = th;
	      topx = e.xconfigure.x;
	      topy = e.xconfigure.y;

	      pushEvent (GRF_REDRAW, x+CTRL.w, y, CTRLWIN);
	      pushEvent (GRF_REDRAW, x, y, MAINWIN);

	      if (doResize)
	       { pushEvent (GRF_RESIZE, w, h, MAINWIN);
	       }
	      if (CTRL.active)
	       { XDrawLine(disp,topwin,topWinGc,MAIN.x+MAIN.w,0,
						MAIN.x+MAIN.w,toph-1);
	       }
	      break;
	    }
	   case Expose:
	    { /* XExposeEvent - xexpose */
	      if (e.xexpose.window==topwin)
	       { int x, y;

		 x = topx+margin;
		 y = (height-MAIN.h) - (topy+margin);
		 pushEvent (GRF_REDRAW, x+CTRL.w, y, CTRLWIN);
		 pushEvent (GRF_REDRAW, x, y, MAINWIN);
		 if (CTRL.active)
		  { XDrawLine(disp,topwin,topWinGc,MAIN.x+MAIN.w,0,
						MAIN.x+MAIN.w,toph-1);
		  }
	       }
	      break;
	    }
	   case GraphicsExpose:
	    { /* XGraphicsExposeEvent - xgraphicsexpose */
	      /* XXX - shouldn't depend on being able to cast a Window into a Drawable (but how else?) */
	      if (e.xgraphicsexpose.drawable==(Drawable)MAIN.id ||
		  e.xgraphicsexpose.drawable==(Drawable)CTRL.id)
	       { int x, y;

		 x = topx+margin;
		 y = (height-MAIN.h) - (topy+margin);
		 pushEvent (GRF_REDRAW, x+CTRL.w, y, CTRLWIN);
		 pushEvent (GRF_REDRAW, x, y, MAINWIN);
		 if (CTRL.active)
		  { XDrawLine(disp,topwin,topWinGc,MAIN.x+MAIN.w,0,
						MAIN.x+MAIN.w,toph-1);
		  }
	       }
	      break;
	    }
	   case MotionNotify:
	    { while (XQLength(disp) > 0)
	       { XPeekEvent (disp, &eNext);
		 if (eNext.type != MotionNotify || 
		     eNext.xany.window != e.xany.window)
		  { break;
		  }
		 XNextEvent (disp, &e);
	       }
	      code = GRF_CURSOR;
	      if (win == CTRLWIN)
	       { val1 = e.xmotion.x;
		 val2 = (CTRL.h-1)-e.xmotion.y;
	       }
	      else
	       { val1 = e.xmotion.x;
		 val2 = (MAIN.h-1)-e.xmotion.y;
	       }
	      break;
	    }
	   case ButtonPress:
	    { switch (e.xbutton.button)
	       { case Button1: code = GRF_LEFT_MOUSE;
		      break;
		 case Button2: code = GRF_MIDDLE_MOUSE;
		      break;
		 case Button3: code = GRF_RIGHT_MOUSE;
		      break;
	       }
	      val1 = 1;
	      break;
	    }
	   case ButtonRelease:
	    { switch (e.xbutton.button)
	       { case Button1: code = GRF_LEFT_MOUSE;
		      break;
		 case Button2: code = GRF_MIDDLE_MOUSE;
		      break;
		 case Button3: code = GRF_RIGHT_MOUSE;
		      break;
	       }
	      val1 = 0;
	      break;
	    }
	   case KeyPress:
	    { code = codeToKeyMap[e.xkey.keycode];
	      val1 = 1;
	      break;
	    }
	   case KeyRelease:
	    { code = codeToKeyMap[e.xkey.keycode];
	      val1 = 0;
	      break;
	    }
	   case EnterNotify:
	    { code = GRF_FOCUS_CHANGE;
	      val1 = win;
	      break;
	    }
	   case LeaveNotify:
	    { code = GRF_FOCUS_CHANGE;
	      val1 = NULLWIN;
	      break;
	    }
	 }
	ev->val1 = val1;
	ev->val2 = val2;
	ev->code = code;
	ev->win = win;
	return (1);
}

#else

#include <stdio.h>

grf4X11Init ()
{
	return(-1);
}

grf4X11SetupFxns()
{
	fprintf (stderr, "grf4Init(): X11 support not available\n");
	return (-1);
}

#endif
