/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1991 by John E. Lloyd.				    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
#include <stdio.h>
#include <math.h>

#include "../h/cdefs.h"
#include "../h/rccl.h"	/* XXX MUST include rccl.h before grf4.h */
#include "../h/stage_kynvar.h"

#include "../h/grf4.h"

#define NUM_CIRC_EDGES		12

#define J1	0
#define J2	1
#define J3	2
#define J4	3
#define J5	4

static int capacity;

#define BEIGE		RGBPACK(225,160, 90)
#define DARK_BROWN	RGBPACK( 85, 55, 35)
#define DARK_GREY	RGBPACK( 50, 50, 50)
#define BLACK		RGBPACK(  0,  0,  0)
#define WHITE		RGBPACK(255,255,255)
#define SILVER		RGBPACK(200,200,220)
#define AMBER		RGBPACK(255,125,  0)

/* Stage sim stuff */

#define SSIM_TRACK_DEPTH		 1.2
#define SSIM_TRACK_WIDTH		15.24
#define SSIM_TRACK_X_LEN		45.72
#define SSIM_PEDESTAL_X_DEPTH		 6.35
#define SSIM_TRACK_Y_LEN		35.56
#define SSIM_PEDESTAL_Y_DEPTH		10.35
#define SSIM_ROTOR_RADIUS		7.62
#define SSIM_ROTOR_DEPTH		3.9

static multA1();
static multA2();
static multA3();

drawStageHand ()
{
	grf4Move (0.0, 0.0, 0.0);
	grf4Draw (SSIM_ROTOR_RADIUS, 0.0, 0.0);
	grf4Move (0.0, 0.0, 0.0);
	grf4Draw (0.0, SSIM_ROTOR_RADIUS, 0.0);
}

Grf4Object *stageBaseX;
Grf4Object *stagePedestalX;
Grf4Object *stageBaseY;
Grf4Object *stagePedestalY;
Grf4Object *stageRotor;

defineStageSimLinks()
{
	void *bodyMat;
	void *rotorMat;

	rotorMat = grf4DefMaterial(SILVER, SILVER, 0, 0, 0.0, 1.0);
	bodyMat = grf4DefMaterial(DARK_GREY, DARK_GREY, 0,0x808080,50.0,1.0);

	stageBaseX = grf4GenBox (SSIM_TRACK_X_LEN, SSIM_TRACK_WIDTH, 
				 SSIM_TRACK_DEPTH, bodyMat);
	stagePedestalX = grf4GenBox (SSIM_TRACK_WIDTH, SSIM_TRACK_WIDTH,
				 SSIM_PEDESTAL_X_DEPTH, bodyMat);
	stageBaseY = grf4GenBox (SSIM_TRACK_Y_LEN, SSIM_TRACK_WIDTH,
				 SSIM_TRACK_DEPTH, bodyMat);
	stagePedestalY = grf4GenBox (SSIM_TRACK_WIDTH, SSIM_TRACK_WIDTH,
				 SSIM_PEDESTAL_Y_DEPTH, bodyMat);
	stageRotor = grf4GenDisc (12, SSIM_ROTOR_RADIUS, SSIM_ROTOR_DEPTH,
				rotorMat);
  	capacity = grf4Capacity();
}

drawStageSim (T, kyn, var, drawHand, handpos)
register TRSF *T;
KYN *kyn;
void *var;
int drawHand;
int handpos;
{
	/* clear the display and draw a robot, given a set of joint angles */

	STAGE_KYN *evar = (STAGE_KYN*)var;
	GEN_KYN *gen = kyn->gen;

	TRSF trans;

	/* Draw the 4 posts */
	grf4GetTrans (&trans);
	T[0] = trans;
	multTrsfXyz (&trans, -SSIM_TRACK_X_LEN/2, -SSIM_TRACK_WIDTH/2, 
			     -2*SSIM_TRACK_DEPTH-SSIM_PEDESTAL_X_DEPTH -
			      SSIM_PEDESTAL_Y_DEPTH-SSIM_ROTOR_DEPTH);
	grf4SetTrans (&trans);
	grf4DrawObject (stageBaseX);

	multA1 (&trans, &T[0], kyn, evar);
	T[1] = trans;
	multTrsfRpy (&trans, 0, -90, 0);
	multTrsfXyz (&trans, -SSIM_TRACK_WIDTH/2, -SSIM_TRACK_WIDTH/2, 
			     -SSIM_TRACK_DEPTH-SSIM_PEDESTAL_X_DEPTH -
			      SSIM_PEDESTAL_Y_DEPTH-SSIM_ROTOR_DEPTH);
	grf4SetTrans (&trans);
	grf4DrawObject (stagePedestalX);

	trans = T[1];
	multTrsfRpy (&trans, 0, -90, 0);
	multTrsfXyz (&trans, -SSIM_TRACK_Y_LEN/2, -SSIM_TRACK_WIDTH/2, 
			     -SSIM_TRACK_DEPTH-SSIM_PEDESTAL_Y_DEPTH
			     -SSIM_ROTOR_DEPTH);
	grf4SetTrans (&trans);
	grf4DrawObject (stageBaseY);

	multA2 (&trans, &T[1], kyn, evar);
	T[2] = trans;
	multTrsfXyz (&trans, -SSIM_TRACK_WIDTH/2, -SSIM_TRACK_WIDTH/2, 
			     -SSIM_PEDESTAL_Y_DEPTH-SSIM_ROTOR_DEPTH);
	grf4SetTrans (&trans);
	grf4DrawObject (stagePedestalY);

	trans = T[2];
	multTrsfXyz (&trans, 0, 0, -SSIM_ROTOR_DEPTH);
	grf4SetTrans (&trans);
	grf4DrawObject (stageRotor);

	multA3 (&trans, &T[2], kyn, evar);
	T[3] = trans;

	if (drawHand)
	 { grf4SetTrans (&trans);
	   drawStageHand ();
	 }
}

stageSimViewSize (r, b, h, kyn)
float *r, *b, *h;
KYN *kyn;
{
	GEN_KYN *gen = kyn->gen;

	*r = SSIM_TRACK_X_LEN;
	*b = SSIM_PEDESTAL_Y_DEPTH;
	*h = *r;
}

stageSimLinkFrames (T, B, kyn, var)
TRSF *T, *B;
KYN *kyn;
void *var;
{
	STAGE_KYN *evar = (STAGE_KYN*)var;

	T[0] = *B;
	multA1 (&T[1], &T[0], kyn, evar);
	multA2 (&T[2], &T[1], kyn, evar);
	multA3 (&T[3], &T[2], kyn, evar);
}

/* Routines for calculating the frame matrices for STAGE robots */

static multA1(tr, t1, kyn, var)
register TRSF *tr;
register TRSF *t1;
KYN *kyn;
STAGE_VAR *var;
{
	register GEN_KYN *gen = kyn->gen;
	float D;

	D = var->d[J1]/10;

	tr->n.x = -t1->a.x;
	tr->o.x = -t1->n.x;
	tr->a.x =  t1->o.x;
	tr->p.x =  t1->n.x*D + t1->p.x;

	tr->n.y = -t1->a.y;
	tr->o.y = -t1->n.y;
	tr->a.y =  t1->o.y;
	tr->p.y =  t1->n.y*D + t1->p.y;

	tr->n.z = -t1->a.z;
	tr->o.z = -t1->n.z;
	tr->a.z =  t1->o.z;
	tr->p.z =  t1->n.z*D + t1->p.z;
}

static multA2(tr, t1, kyn, var)
register TRSF *tr;
register TRSF *t1;
KYN *kyn;
STAGE_VAR *var;
{
	register GEN_KYN *gen = kyn->gen;
	float D;

	D = var->d[J2]/10;

	tr->n.x = -t1->o.x;
	tr->o.x =  t1->a.x;
	tr->a.x = -t1->n.x;
	tr->p.x =  t1->a.x*D + t1->p.x;

	tr->n.y = -t1->o.y;
	tr->o.y =  t1->a.y;
	tr->a.y = -t1->n.y;
	tr->p.y =  t1->a.y*D + t1->p.y;

	tr->n.z = -t1->o.z;
	tr->o.z =  t1->a.z;
	tr->a.z = -t1->n.z;
	tr->p.z =  t1->a.z*D + t1->p.z;
}

static multA3(tr, t1, kyn, var)
register TRSF *tr;
register TRSF *t1;
KYN *kyn;
STAGE_VAR *var;
{
	register GEN_KYN *gen = kyn->gen;
	register float C, S;

	C = var->c3;
	S = var->s3;

	tr->n.x =   C*t1->n.x + S*t1->o.x;
	tr->o.x =   C*t1->o.x - S*t1->n.x;
	tr->a.x =   t1->a.x;
	tr->p.x =   t1->p.x;

	tr->n.y =   C*t1->n.y + S*t1->o.y;
	tr->o.y =   C*t1->o.y - S*t1->n.y;
	tr->a.y =   t1->a.y;
	tr->p.y =   t1->p.y;

	tr->n.z =   C*t1->n.z + S*t1->o.z;
	tr->o.z =   C*t1->o.z - S*t1->n.z;
	tr->a.z =   t1->a.z;
	tr->p.z =   t1->p.z;
}
