#ifdef Tsp_TARGET         

#include <stdio.h>
#include <ctype.h>
#include <cdefs.h>        
#include <conc.h>
#include "tgen.h"
#include "tgenTsp.h"
#include "servo.h"
#include <vector.h>
#include <puma_kynvar.h>

extern double strtod();

#define SGN(x)	((x) >= 0 ? 1 : -1)

#define SEC_TO_TICKS    (1000000/64.)

/* x,y,x and roll,pitch,yaw terms defining the transformation ST
   from T6 to sensor coordinates */

#define ST_X		0.0
#define ST_Y		0.0
#define ST_Z   	      101.4
#define ST_ROLL	        0.0
#define ST_PITCH	0.0
#define ST_YAW		0.0

/* x,y,x and roll,pitch,yaw terms defining the transformation JT
   from sensor coordinates to jig coordinates */

#define JT_X		0.0
#define JT_Y		0.0
#define JT_Z   	       14.08
#define JT_ROLL	       45.0
#define JT_PITCH	0.0
#define JT_YAW		0.0

/* Diameter of pin holding weight in place */

#define PIN_DIAMETER	4.66

TRSF ST, JT;

#define NSAMP	10	/* number od samples to collect */

Tgen *tg;
TspTgenCtrl *ctrl;
FILE *fp;
int dataCount = 0;

writeDataFile (fp, data, force)
FILE *fp;
double *data;
FORCE *force;
{
	if (dataCount == 0)
	 { fprintf (fp, "    f = [");
	   fprintVf (fp, "%9.4f", (float*)force, 6);
	   fprintf (fp, "]';\n");
	   fprintf (fp, "    d = [");
	   fprintVd (fp, "%9.6f", data, 6);
	   fprintf (fp, "]';\n");
	   fflush (fp);
	 }
	else
	 { fprintf (fp, "f = [f, [");
	   fprintVf (fp, "%9.4f", (float*)force, 6);
	   fprintf (fp, "]'];\n");
	   fprintf (fp, "d = [d, [");
	   fprintVd (fp, "%9.6f", data, 6);
	   fprintf (fp, "]'];\n");
	   fflush (fp);
	 }
	dataCount++;
}

getRawData (rawVals)
int *rawVals;
{
	tgenLock (tg);
	copyVi (rawVals, ctrl->rawForceData, 6);
	tgenUnlock (tg);
}

getUnbiasedData (vals)
double *vals;
{
	int rawVals[6];

	tgenLock (tg);
	copyVi (rawVals, ctrl->rawForceData, 6);
	tgenUnlock (tg);
	subtractVi (rawVals, rawVals, ctrl->rawForceBias, 6);
	copyVid (vals, rawVals, 6);
	scaleVd (vals, 1/512.0, vals, 6);
}

getRealT6 (T)
TRSF *T;
{
	tgenLock (tg);
	*T = tg->T6real;
	tgenUnlock (tg);
}

getRealJ6 (j6)
JntVec *j6;
{
	tgenLock (tg);
	*j6 = tg->j6real;
	tgenUnlock (tg);
}

getWord (buf)
char *buf;
{
	int c;

	do
	 { c = getchar();
	 }
	while (isspace(c) && c != EOF && c != '\n');
	if (c == EOF || c == '\n')
	 { return (0);
	 }
	ungetc (c, stdin);
	scanf ("%s", buf);
	return (1);
}

getInt (ip)
int *ip;
{
	char buf[128];
	char *check;

	if (getWord (buf) == 0)
	 { return (0);
	 }
	*ip = strtol(buf, &check, 10);
	return (!(check == buf));
}

getDouble (dp)
double *dp;
{
	char buf[128];

	char *check;

	if (getWord (buf) == 0)
	 { return (0);
	 }
	*dp = strtod(buf, &check);
	return (!(check == buf));
}

waitForMotionCompletion ()
{
	do
	 { ProcWait (0.2*SEC_TO_TICKS);
	 }
	while (tgenIsBusy(tg));
	ProcWait (1.0*SEC_TO_TICKS);
	ctrl->resetBias = 1;
	do
	 { ProcWait (0.2*SEC_TO_TICKS);
	 }
	while (ctrl->resetBias);
}

main()
{
	int rawVals[6];
	int unbiasedVals[6];
	JntVec jc;
	TRSF TC, TH;
	int j;

	/* Set values for the ST and JT transformations: */

	xyzToTrsf (&ST, ST_X, ST_Y, ST_Z);
	rpyToTrsf (&ST, ST_ROLL, ST_PITCH, ST_YAW);
	xyzToTrsf (&JT, JT_X, JT_Y, JT_Z);
	rpyToTrsf (&JT, JT_ROLL, JT_PITCH, JT_YAW);

	printf ("Allocating trajectory generator ...\n");

	if ((tg = tgenAlloc ("caesar")) == NULL)
	 { printErrors ();
	   exit (1);
	 }
	printf ("done\n");

#if 0
	tgenSetCartVel (tg, 200.0, -1);
	tgenSetCartAcc (tg, 1000.0, -1);
#endif

	printf ("Starting trajectory generator ...\n");
	if (tgenStart (tg, 0.01, LINK3OUT, LINK3IN, LINK2OUT, LINK2IN) < 0)
	 { printErrors();
	   exit (1);
	 }
	printf ("done\n");

	ctrl = (TspTgenCtrl*)tg->ctrl;

#if 0
	if (getRobotPosition (jc.v, "demohome", "caesar") < 0)
	 { tgenRelease (tg);
	   printErrors();
	   exit (1);
	 }
	printf ("moving to demohome (confirm):\n");
	while (getchar() != '\n')
	   ;
	tgenMoveJ (tg, &jc);

	do
	 { ProcWait (0.2*SEC_TO_TICKS);
	 }
	while (tgenIsBusy(tg));

	printf ("There.\n");
#endif

	while (1)
	 { char cmd[64];

	   printf ("FCAL> ");
	   fflush (stdout);
	   if (scanf ("%s", cmd) == EOF)
	    { tgenRelease (tg);
	      exit (0);
	    }
	   if (strcmp ("?", cmd) == 0)
	    { 
printf ("?:		help\n");
printf ("s:		show raw force values\n");
printf ("o <file>:	open file for calibration output\n");
printf ("c:		close calibration output file\n");
printf ("m <jnt> <ang>:	move jnt to a given angle\n");
printf ("t <x> <y> <z>:	translate in tool coordinates by x, y, and z\n");
printf ("w:		show robot position\n");
printf ("X <m1> <d1> <m2> <d2>: record data with weights on x axis\n");
printf ("Y <m1> <d1> <m2> <d2>: record data with weights on y axis\n");
printf ("r:		reset bias\n");
printf ("q:		quit\n");
	    }
	   else if (strcmp ("s", cmd) == 0)
	    { getRawData (rawVals);
	      subtractVi (unbiasedVals, rawVals, ctrl->rawForceBias, 6);
	      printVi ("Raw:      %4d\n", rawVals, 6);
	      printVi ("Unbiased: %4d\n", unbiasedVals, 6);
	      printVf ("FORCE:    %9.3f\n", (float*)&tg->force, 6);
	    }
	   else if (strcmp ("o", cmd) == 0)
	    { char fileName[128];
	      if (fp != NULL)
	       { fprintf (fp, "c = f / d\n");
		 fclose (fp);
	       }
	      if (getWord(fileName) == 0)
	       { fprintf (stderr, "'o' needs filename argument\n");
	       }
	      else
	       { if ((fp = fopen(fileName, "w")) == NULL)
		  { fprintf (stderr, "Can't open file\n");
		  }
	       }
	      dataCount = 0;
	    }
	   else if (strcmp ("c", cmd) == 0)
	    { if (fp != NULL)
	       { fprintf (fp, "c = (d'\f')'\n");
		 fclose(fp);
	       }
	      dataCount = 0;
	    }
	   else if (strcmp ("m", cmd) == 0)
	    { int jnt;
	      double angle;
	      JntVec jtemp;

	      if (getInt(&jnt) == 0 || getDouble(&angle) == 0)
	       { fprintf (stderr, "need to give joint number and angle\n");
	       }
	      else if (jnt <= 0 || jnt > 6)
	       { fprintf (stderr, "joint %d not valid\n", jnt);
	       }
	      else
	       { jtemp = tg->j6;

		 jtemp.v[jnt-1] = angle*DEGTORAD;
		 if (tgenMoveJ (tg, &jtemp) < 0)
		  { printErrors();
		    clearErrors();
		  }
		 else
		  { waitForMotionCompletion();
		  }
	       }
	    }
	   else if (strcmp ("t", cmd) == 0)
	    { int jnt;
	      double x, y, z;
	      TRSF TC;

	      if (getDouble(&x) == 0 || getDouble(&y) == 0 || 
		  getDouble(&z) == 0)
	       { fprintf (stderr, "need to give x,y,z translation values\n");
	       }
	      else
	       { TC = tg->T6;

		 multTrsfXyz (&TC, x, y, z);
		 if (tgenMoveC (tg, &TC, (TRSF*)0) < 0)
		  { printErrors();
		    clearErrors();
		  }
		 else
		  { waitForMotionCompletion();
		  }
	       }
	    }
	   else if (strcmp ("w", cmd) == 0)
	    { TRSF T6real;
	      JntVec j6real;

	      getRealT6 (&T6real);
	      getRealJ6 (&j6real);
	      printJf   ("Target: %8.3f\n", tg->j6.v, 6);
	      printJf   ("Angles: %8.3f\n\n", j6real.v, 6);
	      printTrsf ("T6:	  %m\n", &T6real);
	      printf    ("tgenCnt:%d\n", tg->timeCnt);
	    }
	   else if (strcmp ("X", cmd) == 0 || strcmp ("Y", cmd) == 0)
	    { double m1, d1, m2, d2;

	      if (getDouble (&m1) == 0 || getDouble (&d1) == 0 ||
		  getDouble (&m2) == 0 || getDouble (&d2) == 0)
	       { fprintf (stderr, "command needs four double values\n");
	       }
	      else
	       { TRSF TT, D1, D2, FT1, FT2;
		 FORCE g, f1, f2, ftotal;
		 double data[6], temp[6];
		 int i;
		
		 if (d1 != 0)
		  { d1 += SGN(d1)*PIN_DIAMETER;
		  }
		 if (d2 != 0)
		  { d2 += SGN(d2)*PIN_DIAMETER;
		  }

		 getRealT6 (&TT);
		 multTrsf (&TT, &TT, &ST);
		 multTrsf (&TT, &TT, &JT);
		 identTrsf (&D1);
		 identTrsf (&D2);
		 if (strcmp ("X", cmd) == 0)
		  { xyzToTrsf (&D1, d1, 0, 0);
		    xyzToTrsf (&D2, d2, 0, 0);
		  }
		 else
		  { xyzToTrsf (&D1, 0, d1, 0);
		    xyzToTrsf (&D2, 0, d2, 0);
		  }
		 invertTrsf (&FT1, &JT);
		 multLiTrsf (&FT1, &D1, &FT1);
		 invertTrsf (&FT2, &JT);
		 multLiTrsf (&FT2, &D2, &FT2);

		 zeroForce (&g);
		 g.f.z = m1 * 9.8;
		 rotateVect (&g.f, &g.f, &TT);
		 transForce (&f1, &g, &FT1);

		 zeroForce (&g);
		 g.f.z = m2 * 9.8;
		 rotateVect (&g.f, &g.f, &TT);
		 transForce (&f2, &g, &FT2);

		 addVf ((float*)&ftotal, (float*)&f1, (float*)&f2, 6);
		
		 zeroVd (data, 6);
		 for (i=0; i<NSAMP; i++)
		  { getUnbiasedData (temp);
		    addVd (data, data, temp, 6);
		    ProcWait (2*tg->sampInterval*SEC_TO_TICKS);
		  }
		 scaleVd (data, 1/(double)NSAMP, data, 6);

		 printVd ("Data:  %9.6f\n", data, 6);
		 printVf ("Force: %9.4f\n", (float*)&ftotal, 6);

		 if (fp)
		  { writeDataFile (fp, data, &ftotal);
		  }
	       }
	    }
	   else if (strcmp (cmd, "q") == 0)
	    { tgenRelease (tg);
	      exit (0);
	    }
	   else if (strcmp (cmd, "r") == 0)
	    { ctrl->resetBias = 1;
	      ProcWait (2*tg->sampInterval*SEC_TO_TICKS);	      
	      if (ctrl->resetBias)
	       { fprintf (stderr, "Warning: bias reset flag not cleared\n");
	       }
	    }
	   else
	    { printf ("Unrecognized command. Use '?' for help.\n");
	    }
	   while (getchar() != '\n')
	      ;
	 }
}

#endif
