#ifdef Tsp_TARGET

#include <conc.h>
#include "servo.h"
#undef WS_KINIF		/* Botch!! */
#undef WS_SERVOIF	/* */
#include "xkin.h"
#include "gripper.h"

int filteredForceData[6];
int filterWin = 1;
int timeDiff;

/* System topology:

	-------
	|     |
  servo | 14  | 1 ---------+      
	|     |            0
	-------		-------             -------
			|     |		    |     |
		   +- 3 |  X  | 2 ------- 3 |  19 | 2 --+
		   |    |     |		    |     |     |
	-------    |	-------		    -------     |
	|     |    +------ 1 ---------------------------+
    kin | 14  | 0 ---------+
       	|     |
	------

	X is the node this code runs on.
	Link 2 is used for host servo commands
	Link 3 is used for host kin commands
      	18 is the interface tram to the servo and kin transputers.
	19 is the interface tram to the host system.
*/

void servoComTask()
{
	/* Takes care of communication between the host and the servo */

	static unsigned char comBuf[256];

	Channel *toHost;
	Channel *fromHost;
	Channel *toServo;
	Channel *fromServo;

	int com;

	toServo = LINK0OUT;
	fromServo = LINK0IN;
	toHost = LINK2OUT;
	fromHost = LINK2IN;

	while (1)
	 { int len;

	   ChanIn (fromHost, comBuf, 2);
	   com = comBuf[0];
	   if ((com & 0xf0) == RCI_MSG_CODE)
	    { len = comBuf[1];
	    }
	   else
	    { len = MSG_LEN - 2;
	    }
	   if (len)
	    { ChanIn (fromHost, &comBuf[2], len);
	    }
	   ChanOut (toServo, comBuf, len+2);

	   switch (com)
	    { case INTERPOLATE:
	       { ChanIn (fromServo, comBuf, R_LEN_ONE);
		 ChanOut (toHost, comBuf, R_LEN_ONE);
		 break;
	       }
	      case LISTEN_MODE:
	       { ChanIn (fromServo, comBuf, R_LEN_LISTEN_MODE);
		 ChanOut (toHost, comBuf, R_LEN_LISTEN_MODE);
		 break;
	       }
	      case NORMAL_MODE:
	       { ChanIn (fromServo, comBuf, R_LEN_NORMAL_MODE);
		 ChanOut (toHost, comBuf, R_LEN_NORMAL_MODE);
		 break;
	       }
	      case RCI_GATHER:
	       { unsigned char tmpBuf[32];
		 int len;

		 ChanIn (fromServo, comBuf, 2);
		 len = comBuf[1];
		 if (len)
		  { ChanIn (fromServo, &comBuf[2], len);
		  }
#ifdef EXEC_TIME_HACK
		 ChanIn (fromServo, tmpBuf, 6);
#endif
		 ChanOut (toHost, &comBuf[1], len+1);
	       }
	      default:
	       { /* should do something smart here */
	       }
	    }
	 }
}

void forceComTask()
{
	/* Takes care of force/gripper request information from the host */

	static unsigned char comBuf[256];

	Channel *toHost;
	Channel *fromHost;

	toHost = LINK3OUT;
	fromHost = LINK3IN;

	while (1)
	 { int winSize;

	   winSize = ChanInChar (fromHost);
	   ProcToHigh();
	   filterWin = winSize;
	   bcopy (filteredForceData, comBuf, sizeof(int)*6);
	   comBuf[24] = timeDiff;
	   ProcToLow();
	   ChanOut (toHost, comBuf, sizeof(int)*6+1);
	 }
}

void gripperComTask ()
{
	static int gripperBuf[256/4];
	static double force[6];

	Channel *toKin;
	Channel *fromKin;

	char TaskIdReg;
	int l, status, win, i;

	int t0, t1;

	GRIP_REQ *req = (GRIP_REQ*)&gripperBuf[1];
	GRIP_REP *rep = (GRIP_REP*)&gripperBuf[2];

	toKin = LINK1OUT;
	fromKin = LINK1IN;
	
	bzero (force, 6*sizeof(double));

	win = 1;

	while (1)
	 { t0 = Time(0);

	   gripperBuf[0] = KIN_B430_FN;
	   req->type = GRIPPER_NOOP;
	   req->gripper.force = 0;

	   ChanOutChar(toKin, IF_TASK_START);
	   ChanOutChar(toKin, TaskIdReg);
	   ChanOutInt(toKin, B430_MSG_IN);
	   ChanOut(toKin, gripperBuf, B430_MSG_IN);
	   status = ChanInChar(fromKin);

	   if (status != 0)
	    { continue;
	    }
	   ChanOutChar(toKin, IF_TASK_OUTPUT);
	   ChanOutChar(toKin, TaskIdReg);
	   l = ChanInInt(fromKin);
	   ChanIn(fromKin, gripperBuf, B430_MSG_OUT);
	   status = ChanInChar(fromKin);

	   for (i=0; i<6; i++)
	    { force[i] = (force[i]*(win-1)+rep->force_feedback[i])/(double)win;
	    }
	   t1 = Time();
	   ProcToHigh();
	   filteredForceData[0] = force[0];
	   filteredForceData[1] = force[1];
	   filteredForceData[2] = force[2];
	   filteredForceData[3] = force[3];
	   filteredForceData[4] = force[4];
	   filteredForceData[5] = force[5];
	   timeDiff = t1-t0;
	   win = filterWin;
	   ProcToLow();
	 }
}

int main()
{
	Process *gripperProc;
	Process *servoProc;
	Process *forceProc;

	gripperProc = ProcAlloc (gripperComTask, 0x2000, /*num params=*/0);
	ProcRun (gripperProc);
	servoProc = ProcAlloc (servoComTask, 0x2000, /*num params=*/0);
	ProcRun (servoProc);
	forceProc = ProcAlloc (forceComTask, 0x2000, /*num params=*/0);
	ProcRun (forceProc);
}

#endif /* Tsp_TARGET */
