#include <math.h>
#include <vector.h>
#include <errorStack.h>
#include <varargInc.h>
#include "tgen.h"

#define MAX_ADD_REQ	2

void tgenLock (tg)
Tgen *tg;
{
	if (tg->lock)
	 { (*tg->lock)(tg);
	 }
}

void tgenUnlock (tg)
Tgen *tg;
{
	if (tg->unlock)
	 { (*tg->unlock)(tg);
	 }
}

void tgenAbort VA_ARGDEF((Tgen *tg, int code, char *fmt, ...))
{
	va_list ap;
#if !USE_STDARG
	Tgen *tg;
	int code;
	char *fmt;

	va_start(ap);
	tg = va_arg(ap, Tgen*);
	code = va_arg(ap, int);
	fnt = va_arg(ap, char*);
#else
	va_start (ap, fmt);
#endif
	if (tg->abort)
	 { (*tg->abort) (tg, code, fmt, ap);
	 }
	else
	 { extern ERROR_STACK_DESC *errorStack;
	   verrMsg (errorStack, code, fmt, ap);
	 }
}

MotionRequest *getNewMotionRequests (tg, limit)
Tgen *tg;
int limit;
{
	/* remove up to "limit" motion requests from the motion 
	   queue for processing */

	MotionRequest *list, *req;
	int k;

	if (tg->reqHead == NULL)
	 { return (NULL);
	 }
	tgenLock (tg);
	list = tg->reqHead;
	for (k=1, req=list; k<limit && req->next; k++)
	 {  req = req->next;
	 }
	tg->reqHead = req->next;
	req->next = NULL;
	if (req == tg->reqTail)
	 {  tg->reqTail = NULL;
	 }
	tgenUnlock (tg);
	return (list);
}

appendIdleRequest (tg)
Tgen *tg;
{
	MotionRequest *req;

	req = (MotionRequest*)malloc(sizeof(MotionRequest));
	*req = *tg->queueTail;
	req->segTime = 0.0;
	req->transTime = -1;
	req->flags = TGEN_STOP | TGEN_IDLE;
	req->id = 0;
	if (req->pathType == TGEN_CART_PATH)
	 { req->TB = req->TC;
	 }

	appendMotionRequest (tg, req);	
}

deleteIdleRequest (tg)
Tgen *tg;
{
	MotionRequest *req;

	req = tg->queueTail;
	tg->queueTail = req->prev;
	tg->queueTail->next = NULL;
	
	if (tg->nextnextMot->req == req)
	 { tg->nextnextMot->req = NULL;
	 }
	free (req);
}

appendMotionRequest (tg, req)
Tgen *tg;
MotionRequest *req;
{
	req->prev = tg->queueTail;
	req->next = NULL;
	tg->queueTail->next = req;
	tg->queueTail = req;

/* printf ("#Appending motion request\n"); */
	
	if (tg->nextMot->req == NULL)
	 { tg->nextMot->req = req;
	   motionControlInit (tg->nextMot, (MotionControl*)0, tg);

	   /* currMot *must* be an idle motion, so set segCnt accordingly */

	   tg->currMot->segCnt = tg->currMot->transCnt + 
				 tg->nextMot->transCnt;
	 }
	else if (tg->nextnextMot->req == NULL)
	 { tg->nextnextMot->req = req;
	   motionControlInit (tg->nextnextMot, tg->nextMot, tg);
	   maybeFixupSegCnt (tg);
	 }
}

double computeScale (ctrl)
MotionControl *ctrl;
{
	if (ctrl->req->flags & TGEN_IDLE)
	 { return (1.0);
	 }
	else
	 { return (ctrl->cnt/(double)ctrl->segCnt);
	 }
}

trajectoryCycle (tg)
Tgen *tg;
{
	MotionControl *curr;
	MotionControl *next;
	MotionRequest *newReqs;
	MotionRequest *req, *nextReq;
	char buf[32];

	tg->timeCnt++;

	if (newReqs = getNewMotionRequests(tg, MAX_ADD_REQ))
	 { if (tg->nextMot->req && tg->nextMot->req != tg->queueTail)
	    { deleteIdleRequest (tg);
	    }
	   for (req=newReqs; req; req=nextReq)
	    { nextReq = req->next;
	      appendMotionRequest (tg, req);
	    }
	   appendIdleRequest (tg);
	 }

	curr = tg->currMot;
	next = tg->nextMot;

	curr->cnt++;
	curr->scale = computeScale(curr);
	if (curr->cnt < curr->transCnt)
	 { tg->prevMot->cnt++;
	   tg->prevMot->scale = computeScale(tg->prevMot);
	   computeBlendedMotion (tg);
/* printf ("#%d Blend\n", curr->cnt); */
	 }
	else
	 { computeCruiseMotion (tg);
/* printf ("#%d Cruise\n", curr->cnt); */
	   if (tg->prevMot->status == TGEN_ACTIVE)
	    { tg->prevMot->status = TGEN_COMPLETED;
	    }
	 }
	if (tg->monitorFxn)
	 { (*tg->monitorFxn)(tg, tg->monitorArg, tg->monitorPtr);
	 }
	if (curr->req->pathType == TGEN_CART_PATH)
	 { if (tg->alterFxn)
	    { (*tg->alterFxn)(tg, tg->alterArg, tg->alterPtr);
	    }
	 }
	if (tg->flags & TGEN_ROBOT_ATTACHED)
	 { JntVec reqVel;
	   int velLim = 0;
	   int posLim = 0;
	   int j;

	   if (curr->req->pathType == TGEN_CART_PATH)
	    { TRSF T6final;
	      int status;

	      multTrsf (&T6final, &tg->T6, &tg->Alter);
	      status = invKinematics (tg->j6.v, &T6final, tg->conf, tg->j6.v,
						tg->kyn, tg->jls, (void*)0);
	      if (status & KYN_JOINT_LIMIT)
	       { tgenAbort (tg, TGJntLimit,
"trajectoryCycle(): joint limit exceeded for %s\n", 
			  jointDescStr(buf, status & KYN_JOINT_LIMIT));
		 return (-1);
	       }
	      else if (status & KYN_CANT_REACH)
	       { tgenAbort (tg, TGWorkspaceLimit, 
"trajectoryCycle(): position out of reach\n");
		 return (-1);
	       }
	    }
	   else /* TGEN_JNT_PATH */
	    { if (tg->flags & TGEN_ROBOT_ATTACHED)
	       { fwdKinematics (&tg->T6,&tg->conf,tg->j6.v,tg->kyn,(void*)0);
	       }
	    }
	   /* do limit checks */
	   subtractVf (reqVel.v, tg->j6.v, tg->oldJ6.v, NJNTS);
	   scaleVf (reqVel.v, 1/tg->sampInterval, reqVel.v, NJNTS);
	   for (j=0; j<NJNTS; j++)
	    { if (ABS(reqVel.v[j]) > tg->jls->maxVel[j])
	       { velLim |= (1<<j);
printf ("%g\n", ABS(reqVel.v[j]));
	       }
	      if (tg->j6.v[j] < tg->jls->jmin[j] || 
		  tg->j6.v[j] > tg->jls->jmax[j])
	       { posLim |= (1<<j);
	       }
	    }
	   if (posLim)
	    { tgenAbort (tg, TGJntLimit,
"trajectoryCycle(): joint limit exceeded for %s\n", 
			 jointDescStr(buf, posLim));
	      return (-1);
	    }
	   if (velLim)
	    { tgenAbort (tg, TGVelLimit,
"trajectoryCycle(): velocity limit exceeded for %s\n", 
			 jointDescStr(buf, velLim));
	      return (-1);
	    }
	 }
	tg->oldJ6 = tg->j6;
	if (next->req && curr->cnt >= curr->segCnt - next->transCnt)
	 { /* initiate next motion */
	   curr->segCnt = curr->cnt + next->transCnt;
	   advanceToNextMotion (tg);
	   currentMotionInit (tg);
	 }
	if (tg->flags & TGEN_DIAG_OUTPUT)
	 { appendDiagOutput (tg);
	 }
	return (1);
}

appendDiagOutput (tg)
Tgen *tg;
{
	TgenDiag *tgd, *diag;

	tgd = ALLOC(TgenDiag);
	tgd->next = NULL;
	if (tg->diagTail == NULL)
	 { tg->diagHead = tgd;
	   tgd->prev = NULL;
	 }
	else
	 { tg->diagTail->next = tgd;
	   tgd->prev = tg->diagTail;
	 }
	tg->diagTail = tgd;

	tgd->T6 = tg->T6;
	tgd->j6 = tg->j6;
	tgd->conf = tg->conf;
	if (tg->currMot->cnt == -tg->currMot->transCnt)	/* new motion */
	 { tgd->mctl = ALLOC(MotionControl);
	   *tgd->mctl = *tg->currMot;
	   /* XXX amend previous motion segCnt XXX */
	   for (diag=tgd->prev; diag && diag->mctl==NULL; diag=diag->prev)
	      ;
	   if (diag)
	    { diag->mctl->segCnt = tg->prevMot->segCnt;
	    }	   
	 }
	else
	 { tgd->mctl = NULL;
	 }
}

advanceToNextMotion (tg)
Tgen *tg;
{
	MotionControl *old;
	MotionControl *next;

	old = tg->prevMot;
	tg->prevMot = tg->currMot;
	tg->currMot = tg->nextMot;
	tg->nextMot = tg->nextnextMot;
	tg->nextnextMot = old;

	free (old->req);
	tg->prevMot->req->prev = NULL;

	next = tg->nextMot;
	if (next->req)
	 { tg->nextnextMot->req = next->req->next;
	   if (tg->nextnextMot->req)
	    { motionControlInit (tg->nextnextMot, next, tg);
	      maybeFixupSegCnt (tg);
	    }
	 }
	else
	 { tg->nextnextMot->req = NULL;
	 }
}

maybeFixupSegCnt(tg)
Tgen *tg;
{
	MotionControl *next;
	int minCnt;

	next = tg->nextMot;
	minCnt = next->transCnt + tg->nextnextMot->transCnt;
	next->segCnt = MAX(minCnt, next->segCnt);
}
