#include "../h/signalFxns.h"

SIGHAN_RETURN (*signalCatch (sig, fxn))()
int sig;
SIGHAN_RETURN (*fxn)();
{
#ifdef HAS_BSD_SIGNALS
	struct sigvec vec, ovec;

	vec.sv_flags = 0;
	vec.sv_mask = 0;
	vec.sv_handler = fxn;
	if (sigvec (sig, &vec, &ovec) < 0)
	 { return (SIGNAL_ERR);
	 }
	else
	 { return (ovec.sv_handler);
	 }
#else
#ifdef HAS_POSIX_SIGNALS
	struct sigaction act, oact;

	act.sa_flags = 0;
	sigemptyset (&act.sa_mask);
	act.sa_handler = fxn;
	if (sigaction (sig, &act, &oact) < 0)
	 { return (SIGNAL_ERR);
	 }
	else
	 { return (oact.sa_handler);
	 }
#else
	return (SIGNAL_ERR);
#endif
#endif
}

SigSet signalMaskSet (set)
SigSet set;
{
	SigSet oset;
#ifdef HAS_BSD_SIGNALS
	oset = sigsetmask (set);
#else
#ifdef HAS_POSIX_SIGNALS
	sigprocmask (SIG_SETMASK, &set, &oset);
#else
	oset = sigsetFill();
#endif
#endif
	return (oset);
}

SigSet signalMaskGet ()
{
	SigSet oset;
#ifdef HAS_BSD_SIGNALS
	oset = sigblock (0);
#else
#ifdef HAS_POSIX_SIGNALS
	sigprocmask (0, (SigSet*)0, &oset);
#else
	oset = sigsetFill();
#endif
#endif
	return (oset);
}

SigSet signalMaskAdd (set)
SigSet set;
{
	SigSet oset;
#ifdef HAS_BSD_SIGNALS
	oset = sigblock (set);
#else
#ifdef HAS_POSIX_SIGNALS
	sigprocmask (SIG_BLOCK, &set, &oset);
#else
	oset = sigsetFill();
#endif
#endif
	return (oset);
}

SigSet signalMaskDel (set)
SigSet set;
{
	SigSet oset;
#ifdef HAS_BSD_SIGNALS
	oset = sigblock (0);
	sigsetmask (oset & ~set);
#else
#ifdef HAS_POSIX_SIGNALS
	sigprocmask (SIG_UNBLOCK, &set, &oset);
#else
	oset = sigsetFill();
#endif
#endif
	return (oset);
}

SigSet signalMaskAll ()
{
	SigSet oset;
#ifdef HAS_BSD_SIGNALS
	oset = sigblock (~0);
#else
#ifdef HAS_POSIX_SIGNALS
	SigSet allsigs;
	allsigs = sigsetFill();
	sigprocmask (SIG_BLOCK, &allsigs, &oset);
#else
	oset = sigsetFill();
#endif
#endif
	return (oset);
}

SigSet sigsetClear()
{
	SigSet set;
#ifdef HAS_BSD_SIGNALS
	set = 0;
#else
#ifdef HAS_POSIX_SIGNALS
	sigemptyset (&set);
#else
	set = 0;
#endif
#endif
	return (set);
}

SigSet sigsetFill()
{
	SigSet set;
#ifdef HAS_BSD_SIGNALS
	set = ~0;
#else
#ifdef HAS_POSIX_SIGNALS
	sigfillset (&set);
#else
	set = ~0;
#endif
#endif
	return (set);
}

SigSet sigsetAdd(set, sig)
SigSet set;
int sig;
{
#ifdef HAS_BSD_SIGNALS
	set |= (1<<(sig-1));
#else
#ifdef HAS_POSIX_SIGNALS
	sigaddset (&set, sig);
#else
	set |= (1<<(sig-1));
#endif
#endif
	return (set);
}

SigSet sigsetDel(set, sig)
SigSet set;
int sig;
{
#ifdef HAS_BSD_SIGNALS
	set &= ~(1<<(sig-1));
#else
#ifdef HAS_POSIX_SIGNALS
	sigdelset (&set, sig);
#else
	set &= ~(1<<(sig-1));
#endif
#endif
	return (set);
}

SigSet sigsetInit (sig)
int sig;
{
	SigSet set;
#ifdef HAS_BSD_SIGNALS
	set = (1<<(sig-1));
#else
#ifdef HAS_POSIX_SIGNALS
	sigemptyset (&set);
	sigaddset (&set, sig);
#else
	set = (1<<(sig-1));
#endif
#endif
	return (set);
}

int sigsetHasSig (set, sig)
SigSet set;
int sig;
{
#ifdef HAS_BSD_SIGNALS
	return (set & (1<<(sig-1)));
#else
#ifdef HAS_POSIX_SIGNALS
	return (sigismember (&set, sig));
#else
	return (0);
#endif
#endif
}
