#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/uio.h>
#include <sys/conf.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/modctl.h>
#include <sys/poll.h>
#include <sys/kmem.h>
#include <sys/cmn_err.h>
#include <sys/sysmacros.h>
#include <sys/stat.h>

#if 0
#include <sys/edt.h>
#include <sys/pio.h>
#include <sys/eisa.h>
#endif

#include "eppreg.h"
#include "rcisys.h"
#include "rciioctl.h"

#define STATIC
#define MAPPED_IO	0

int eppdevflag = 0;
void *eppi;

int getEFLAGS()
{
	asm ("pushf");
	asm ("popw %eax");
}

int setIOPL()
{
	asm ("pushf");
	asm ("orw $0x3000,(%esp)");
	asm ("popf");	
}

#if 0
#define IRQ_MASK	0x0010
#endif

#define EPP_INTR_SPEC	0	/* device interrupt spec. */
#define EPP_REG_SPEC	0	/* register set to be mapped */
#define EPP_NULL_HANDLER (u_int(*)(caddr_t))nulldev

/* maximum number of instances supported by this driver */
#define MAX_INSTANCES		4
/* number of bits in the minor node reserved for the instance number */
#define NUM_INSTANCE_BITS	4
/* returns the instance number from a dev_t structure */
#define GET_INSTANCE(d)		(getminor(d) & ((1<<NUM_INSTANCE_BITS)-1))
/* bit setting or-ed with the instance number to get an mmap device */
#define MMAP_MINOR		(1<<NUM_INSTANCE_BITS)
/* returns true if the minor number corresponds to an mmap device */
#define IS_MMAP_DEV(m)	(((m) & ~((1<<NUM_INSTANCE_BITS)-1)) == MMAP_MINOR)

/* Hardware-specific device routines */
static void eppHardwareReset (EppRegs*, EppSoftc*);
static void eppHardwareIntrSetup (EppRegs*, EppSoftc*);
static void eppHardwareOpen (EppRegs*, EppSoftc*);
static void eppHardwareClose (EppRegs*, EppSoftc* );
static void eppHardwareIntrEnable (EppRegs*, EppSoftc*, int);
static  int eppHardwareIntrHandle (EppRegs*, EppSoftc*, int*);
static u_short eppGetCmdReg (EppRegs*, EppSoftc*);

/* Stuff relating to dynamic loading and autoconfiguration. There
   is *a lot* of boiler plate associated with this. */

static int eppidentify();
static int eppattach();
static int eppdetach();
static int eppgetinfo();

static int eppopen(dev_t *devp, int flag, int otyp, cred_t *credp);
static int eppclose(dev_t dev, int flag, int otyp, cred_t *credp);
static int eppread(dev_t dev, struct uio *uiop, cred_t *credp);
static int eppwrite(dev_t dev, struct uio *uiop, cred_t *credp);
static int eppioctl(dev_t dev, int cmd, int arg, int mode, 
			cred_t *credp, int *rvalp);
static int eppmmap(dev_t dev, off_t off, int prot);
static int eppchpoll(dev_t dev, short events, int anyyet,
		    short *reventsp, struct pollhead **phpp);

static u_int eppintr (caddr_t);
static u_int eppsoftintr (caddr_t);

static struct cb_ops epp_cb_ops = {
	eppopen,
	eppclose,
	nodev,
	nodev,
	nodev,
	eppread,
	eppwrite,
	eppioctl,
	nodev,
	eppmmap,
	nodev,
	eppchpoll,
	nodev,
	NULL,
	D_NEW | D_MP
};

static struct dev_ops epp_ops = {
	DEVO_REV,
	0,
	eppgetinfo,
	eppidentify,
	nulldev,
	eppattach,
	eppdetach,
	nodev,
	&epp_cb_ops,
	(struct bus_ops *)NULL
};

static struct modldrv modldrv = {
	&mod_driverops,
	"epp driver",
	&epp_ops
};

static struct modlinkage modlinkage = {
	MODREV_1,
	&modldrv,
	NULL
};

int _init (void)
{
	int status;

	ddi_soft_state_init (&eppi, sizeof(EppSoftc), MAX_INSTANCES);
	status = mod_install (&modlinkage);
	if (status != 0)
	 { ddi_soft_state_fini (&eppi);
	 }
	return (status);
}

int _fini (void)
{
	int status;

	status = mod_remove (&modlinkage);
	if (status == 0)
	 { ddi_soft_state_fini (&eppi);
	 }
	return (status);
}

int _info (struct modinfo *modinfop)
{
	return (mod_info (&modlinkage, modinfop));
}

static int eppidentify (dip)
dev_info_t *dip;
{
	/* For SBus devices, the name checked by this routine must match
	   the name stored on the device itself. */

	if (strcmp (ddi_get_name(dip), "epp") == 0)
	 { return (DDI_IDENTIFIED);
	 }
	else
	 { return (DDI_NOT_IDENTIFIED);
	 }
}

static int eppgetinfo (dev_info_t *dip, ddi_info_cmd_t infocmd, 
			  void *arg, void **result)
{
	EppSoftc *info;
	int status;

	switch (infocmd)
	 { case DDI_INFO_DEVT2INSTANCE:
	    { *result = (void*)GET_INSTANCE((dev_t)arg);
	      status = DDI_SUCCESS;
	      break;
	    }
	   case DDI_INFO_DEVT2DEVINFO:
	    { info = ddi_get_soft_state(eppi, GET_INSTANCE((dev_t)arg));
	      if (info == NULL)
	       { return (DDI_FAILURE);
	       }
	      *result = (void*)info->dip;
	      status = DDI_SUCCESS;
	      break;
	    }
	   default:
	    { status = DDI_FAILURE;
	      break;
	    }
	 }
	return (status);
}

static int eppfreeResources (dev_info_t *dip, EppSoftc *info)
{
	if ((info->resources & EPP_MMAP_MINOR_CREATED) ||
	    (info->resources & EPP_MAIN_MINOR_CREATED))
	 { ddi_remove_minor_node (dip, NULL);
	 }
	if (info->resources & EPP_IHANDLER_SET)
	 { ddi_remove_intr (dip, EPP_INTR_SPEC, &info->iblkCookie);
	 }
	if (info->resources & EPP_SOFT_IHANDLER_SET)
	 { ddi_remove_softintr (info->softIntrId);
	 }
	if (info->resources & EPP_LOW_MUTEX_SET)
	 { mutex_destroy (&info->lowMux);
	 }
	if (info->resources & EPP_INTR_MUTEX_SET)
	 { mutex_destroy (&info->intrMux);
	 }
	if (info->resources & EPP_REGS_MAPPED)
	 { ddi_regs_map_free (&info->regsHandle);
	 }
	ddi_soft_state_free (eppi, info->instance);
}

static int eppattach (dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	EppSoftc *info;
	int instance;
	ddi_device_acc_attr_t dev_attr;

	static char *mainNames[] = { "0", "1", "2", "3" };
	static char *mmapNames[] = { "mem0", "mem1", "mem2", "mem3" };

	switch (cmd)
	 { case DDI_ATTACH:
	    { instance = ddi_get_instance(dip);
	      if (instance >= MAX_INSTANCES)
	       { return (DDI_FAILURE);
	       }
	      if (ddi_soft_state_zalloc(eppi, instance) != 0)
	       { return (DDI_FAILURE);
	       }
	      info = ddi_get_soft_state(eppi, instance);

	      info->dip = dip;
	      info->instance = instance;
	
#if MAPPED_IO
	      /* map device registers */
	      if (ddi_map_regs (dip, EPP_REG_SPEC, (caddr_t*)&info->regs,
				/*offset=*/0, /*len=*/0) != DDI_SUCCESS)
	       { eppfreeResources (dip, info);
		 cmn_err (CE_CONT, "epp: couldn't map regs\n");
		 return (DDI_FAILURE);
	       }
	      info->resources |= EPP_REGS_MAPPED;
#else
#if 0
	      /* Couldn't get this to work ...  */
	       { int buf[10];
		 int len;

		 if (ddi_prop_op (instance, dip, 
			PROP_LEN_AND_VAL_BUF,
			DDI_PROP_DONTPASS,
			"reg", (caddr_t)buf, &len) != DDI_PROP_SUCCESS)
		  { cmn_err (CE_CONT, "epp: couldn't find reg prop\n");
		  }
		 else
		  { cmn_err (CE_CONT, "epp: reg=%x %x %x len=%d\n", 
				buf[0], buf[1], buf[2], len);
		  }
	       }
#endif
	      dev_attr.devacc_attr_version = DDI_DEVICE_ATTR_V0;
	      dev_attr.devacc_attr_endian_flags = DDI_NEVERSWAP_ACC;
	      dev_attr.devacc_attr_dataorder = DDI_STRICTORDER_ACC;
	      if (ddi_regs_map_setup (dip, EPP_REG_SPEC, 
		  (caddr_t*)&info->regs, /*offset=*/0, /*len=*/0,
		   &dev_attr, &info->regsHandle) == DDI_SUCCESS)
	       { cmn_err (CE_CONT, "epp: regs_map_setup worked: 0x%x\n", 
			  info->regs);
		 info->resources |= EPP_REGS_MAPPED;
	       }
	      else
	       { cmn_err (CE_CONT, "epp: regs_map_setup failed\n");
		 info->regs = 0;
	       }
	      info->iobase = 0x240;	/* XXX XXX XXX */
	      printf ("eflags=0x%x\n", getEFLAGS());
	      setIOPL();
	      printf ("eflags=0x%x\n", getEFLAGS());
#endif
	      /* make device quiescent */
	      eppHardwareReset (info->regs, info);

	      /* set up interrupt-level mutex for the device. Start
		 by assigning a dummy interrupt handler so the
		 mutex is associated with the proper interrupt context,
		 without risk of an a handler getting called. */

	      if (ddi_add_intr (dip, EPP_INTR_SPEC,&info->iblkCookie,
		     NULL, EPP_NULL_HANDLER, NULL) != DDI_SUCCESS)
	       { eppfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }

	      mutex_init (&info->intrMux, "epp intr mutex", MUTEX_DRIVER,
			  (void*)info->iblkCookie);
	      ddi_remove_intr (dip, EPP_INTR_SPEC, info->iblkCookie);
	      info->resources |= EPP_INTR_MUTEX_SET;

	      /* see if the device interrupt is delivered in a high-level
		 interrupt context. If it is, we need to set up a 
		 low-level interrupt handler and mutex as well,
		 because we can't call cmn_err() or pollwakeup()
		 from a high-level interrupt context. */

	      if (ddi_intr_hilevel (dip, EPP_INTR_SPEC))
	       { info->hasHighLevelIntr = 1;
		 
		 /* set up low-level mutex and handler */

		 if (ddi_add_softintr (dip, DDI_SOFTINT_HIGH, 
			&info->softIntrId, &info->softIblkCookie, NULL, 
			EPP_NULL_HANDLER, NULL) != DDI_SUCCESS)
		  { eppfreeResources (dip, info);
		    return (DDI_FAILURE);
		  }
		 mutex_init (&info->lowMux, "epp low mutex", MUTEX_DRIVER,
			     (void*)info->softIblkCookie);
		 ddi_remove_softintr (info->softIntrId);
		 info->resources |= EPP_LOW_MUTEX_SET;

		 if (ddi_add_softintr (dip, DDI_SOFTINT_HIGH, 
			&info->softIntrId, &info->softIblkCookie, NULL, 
			eppsoftintr, (caddr_t)info) != DDI_SUCCESS)
		  { eppfreeResources (dip, info);
		    return (DDI_FAILURE);
		  }
		 info->resources |= EPP_SOFT_IHANDLER_SET;

		 info->muxp = &info->lowMux;
		 cmn_err (CE_CONT, "epp: interrupts are high level\n");
	       }
	      else
	       { info->hasHighLevelIntr = 0;
		 info->muxp = &info->intrMux;
		 cmn_err (CE_CONT, "epp: interrupts are low level\n");
	       }

	      if (ddi_add_intr (dip, EPP_INTR_SPEC, &info->iblkCookie,
		     &info->idevCookie, eppintr,(caddr_t)info) != DDI_SUCCESS)
	       { eppfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= EPP_IHANDLER_SET;

	      /* program device interrupt vector */
	      eppHardwareIntrSetup (info->regs, info);

	      if (ddi_create_minor_node (dip, mainNames[instance], S_IFCHR, 
		     instance, DDI_PSEUDO, 0) != DDI_SUCCESS)
	       { eppfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= EPP_MAIN_MINOR_CREATED;

	      if (ddi_create_minor_node (dip, mmapNames[instance], S_IFCHR, 
		     instance|MMAP_MINOR, DDI_PSEUDO, 0) != DDI_SUCCESS)
	       { eppfreeResources (dip, info);
		 return (DDI_FAILURE);
	       }
	      info->resources |= EPP_MMAP_MINOR_CREATED;

	      ddi_report_dev(dip);
cmn_err (CE_CONT, "succeeded\n");
	      return (DDI_SUCCESS);
	    }
	   default:
	    { return (DDI_FAILURE);
	    }
	 }
}

static int eppdetach (dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	EppSoftc *info;
	
	switch (cmd)
	 { case DDI_DETACH:
	    { info = ddi_get_soft_state (eppi, ddi_get_instance(dip));

	      /* make device quiescent */
	      eppHardwareReset (info->regs, info);
	      eppfreeResources (dip, info);
	      return (DDI_SUCCESS);
	    }
	   default:
	    { return (DDI_FAILURE);
	    }
	 }	
}

static u_int eppintr (caddr_t arg)
{
	/* Called in response to an interrupt from the EPP */

	EppSoftc *info = (EppSoftc*)arg;
	int needSoftIntr = 0;
	int pending;
	int stray;

	mutex_enter (&info->intrMux);
	pending = eppHardwareIntrHandle (info->regs, info, &stray);
	if (pending == 0 && stray == 0)
	 { mutex_exit (&info->intrMux);
	   return (DDI_INTR_UNCLAIMED);
	 }

	if (pending)
	 { if (info->hasHighLevelIntr)
	    { if (info->highIntrCnt < info->qsize)
	       { info->highIntrCnt++;
		 if (!info->softIntrRunning)
		  { needSoftIntr = 1;
		  }
	       }
	    }
	   else
	    { if (info->intrCnt < info->qsize)
	       { info->intrCnt++;		/* bump the counter */
		 mutex_exit (&info->intrMux);
		 pollwakeup (&info->pollhead, POLLRDNORM);
		 mutex_enter (&info->intrMux);
	       }
	    }
	 }
	if (stray)
	 { /* declare a stray interrupt */
	   if (info->hasHighLevelIntr)
	    { info->stray = stray;
	      if (!info->softIntrRunning)
	       { needSoftIntr = 1;
	       }
	    }
	   else
	    { cmn_err (CE_CONT, "epp%d: stray interrupt(s); mask=0x%x\n",
		   		 info->instance, stray);
	    }
	 }
	mutex_exit (&info->intrMux);
	if (needSoftIntr)
	 { ddi_trigger_softintr (info->softIntrId);
	 }
	return (DDI_INTR_CLAIMED);
}

static u_int eppsoftintr (caddr_t arg)
{
	/* Used only when the device interrupts occur at high interrupt
	   priority. Calls pollwakeup and cmn-err because the high-level
	   interrupt handler can't. */

	EppSoftc *info = (EppSoftc*)arg;
	int stray;
	int cnt;

	mutex_enter (&info->lowMux);
	mutex_enter (&info->intrMux);

	/* don't know why info->softIntrRunning needs to be checked
	   in the next phrase, but it was in the manual example */
	if ((info->stray == 0 && info->highIntrCnt <= 0) ||
	    info->softIntrRunning)
	 { mutex_exit (&info->intrMux);
	   mutex_exit (&info->lowMux);
	   return (DDI_INTR_UNCLAIMED);	
	 }
	info->softIntrRunning = 1;
	while (info->stray || info->highIntrCnt > 0)
	 { /* ASSERT (mutex_owned (&info->intrMux)); */

	   if ((stray = info->stray) != 0)
	    { info->stray = 0;
	    }
	   if ((cnt = info->highIntrCnt) > 0)
	    { info->highIntrCnt--;
	    }
	   mutex_exit (&info->intrMux);

	   if (stray)
	    { cmn_err (CE_CONT, "epp%d: stray interrupt(s); mask=0x%x\n",
				 info->instance, stray);
	    }
	   if (cnt)
	    { if (info->intrCnt < info->qsize)
	       { info->intrCnt++;
		 mutex_exit (&info->lowMux);
		 pollwakeup (&info->pollhead, POLLRDNORM);
		 mutex_enter (&info->lowMux);
	       }
	    }

	   mutex_enter (&info->intrMux);
	 }
	info->softIntrRunning = 0;
	mutex_exit (&info->intrMux);
	mutex_exit (&info->lowMux);
	return (DDI_INTR_CLAIMED);	
}

STATIC int eppopen (dev_t *dev, int flag, int otyp, cred_t *crp)
{
	/* This device is an exclusive open device: only one process
	   may open it at a time. This prevents contention problems
	   and makes the driver software easier to implement. */

	EppSoftc *info;

	if ((info = ddi_get_soft_state(eppi,GET_INSTANCE(*dev))) == NULL)
	 { return (ENXIO);
	 }
	if (otyp != OTYP_CHR)
	 { return (EINVAL);
	 }
#if MAPPED_IO
	if (!info->regs)	/* paranoid; this check shouldn't be needed */
	 { return (ENXIO);
	 }
#endif

	/* now determine whether this is a main device or an mmap device.
	   Enforce exclusive open for main devices */
	mutex_enter (info->muxp);
	if (IS_MMAP_DEV(getminor(*dev)))
	 { /* OK, just return */
	 }
	else
	 { if (info->flags & EPP_BUSY)
	    { mutex_exit (info->muxp);
	      return (EBUSY);
	    }
	   else
	    { info->flags |= EPP_BUSY;
	    }
	   info->qsize = TQ_DEFAULT_SIZE;
	   info->intrCnt = 0;
	   bzero ((char*)&info->pollhead, sizeof(struct pollhead));

	   eppHardwareOpen (info->regs, info);
	 }
	mutex_exit (info->muxp);
	return (0);
}

static void doclose (EppSoftc *info)
{
	eppHardwareClose (info->regs, info);
	info->flags = 0;
}

STATIC int eppclose (dev_t dev, int flag, int otyp, cred_t *crp)
{
	EppSoftc *info;

	if ((info = ddi_get_soft_state(eppi,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	if (IS_MMAP_DEV(getminor(dev)))
	 { return (0);
	 }
	/* check for soft-close */
	mutex_enter (info->muxp);
	if (! (info->flags & EPP_BUSY) )
	 { mutex_exit (info->muxp);
	   return (0);
	 }
	mutex_exit (info->muxp);
	doclose (info);
	return (0);
}

STATIC int eppread (dev_t dev, uio_t *uiop, cred_t *crp)
{
	/* The read routine returns a single byte (set to 1) if an interrupt
	   arrived. */

	EppSoftc *info;
	int cnt;
	int ack = 1;

	if ((info = ddi_get_soft_state(eppi,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	if (IS_MMAP_DEV(getminor(dev)))
	 { return (ENXIO);
	 }
	/* check for soft-close */
	mutex_enter (info->muxp);
	if (! (info->flags & EPP_BUSY) )
	 { mutex_exit (info->muxp);
	   return (EBADF);
	 }
	cnt = info->intrCnt;
	if (cnt)
	 { info->intrCnt--;
	 }
	mutex_exit (info->muxp);

	if (cnt)
	 { if (ureadc (ack, uiop) != 0)
	    { return (EFAULT);
	    }
	 }
	return (0);
}

STATIC int eppwrite (dev_t dev, uio_t *uiop, cred_t *crp)
{
	return (ENXIO);
}

STATIC int eppmmap (dev_t dev, off_t off, int prot)
{
	EppSoftc *info;

	if ((info = ddi_get_soft_state(eppi,GET_INSTANCE(dev))) == NULL)
	 { return (-1);
	 }
	if (off < 0 || off >= sizeof(EppRegs))
	 { return (-1);
	 }
#if MAPPED_IO
	return (hat_getkpfnum((char*)info->regs + off));
#else
	return (-1);
#endif
}

STATIC int eppchpoll (dev_t dev, short events, int anyyet,
		      short *reventsp, struct pollhead **phpp)
{
	/* Check to see if any interrupt events have arrived. If not,
	   then queue the process for a possible select wakeup. */

	EppSoftc *info;
	int cnt;

	if ((info = ddi_get_soft_state(eppi,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	if (IS_MMAP_DEV(getminor(dev)))
	 { return (ENXIO);
	 }
	/* check for soft-close */
	mutex_enter (info->muxp);
	if (! (info->flags & EPP_BUSY) )
	 { mutex_exit (info->muxp);
	   return (EBADF);
	 }
	cnt = info->intrCnt;
	mutex_exit (info->muxp);

	/* not really sure how system prevents races with the interrupt
	   handler here, but this is what the example says to do ... */
	if ((events & (POLLIN|POLLRDNORM)) && cnt > 0)
	 { *reventsp = events & (POLLIN|POLLRDNORM);
	 }
	else
	 { *reventsp = 0;
	   if (!anyyet)
	    { *phpp = &info->pollhead;
	    }
	 }
	return (0);
}

STATIC int eppioctl (dev_t dev, int cmd, int arg, int mode, 
		     cred_t *crp, int *rvalp)
{
	EppSoftc *info;
	int isMmapDev;

	if ((info = ddi_get_soft_state(eppi,GET_INSTANCE(dev))) == NULL)
	 { return (ENXIO);
	 }
	if ((isMmapDev = IS_MMAP_DEV(getminor(dev))) == 0)
	 { /* check for soft-close */
	   mutex_enter (info->muxp);
	   if (! (info->flags & EPP_BUSY) )
	    { mutex_exit (info->muxp);
	      return (EBADF);
	    }
	   mutex_exit (info->muxp);
	 }

	switch (cmd)
	 { default:
	    { return (ENOTTY);
	      break;
	    }
	   case RCIIOC_CONFIRM:
	    { break;
	    }
	   case RCIIOC_INTR_ENABLE:
	    { mutex_enter (&info->intrMux);
	      eppHardwareIntrEnable (info->regs, info, (int)arg);
	      mutex_exit (&info->intrMux);
	      break;
	    }
	   case RCIIOC_SET_INTRQ:
	    { int size;

	      if (isMmapDev)
	       { return (ENOTTY);
	       }
	      size = (int)arg;
	      if (size <= 0)
	       { return (EINVAL);
	       }
	      else
	       { mutex_enter (info->muxp);
		 info->qsize = size;
		 info->intrCnt = 0;
		 mutex_exit (info->muxp);
	       }
	      break;
	    }
	   case RCIIOC_GET_INTRQ:
	    { int size;

	      if (isMmapDev)
	       { return (ENOTTY);
	       }
	      size = info->qsize;
	      if (copyout ((caddr_t)&size, (caddr_t)arg, sizeof(int)))
	       { return (EFAULT);
		 break;
	       }
	      break;
	    }
/* RCIIOC_SOFT_CLOSE is a special hack. On some machines, calling
   close(2) does not always result in the driver close routine being 
   called right away. So we fake it with an ioctl, just to be sure.
   It is the responsiblity of the application to call close(2) directly
   after the SOFT_CLOSE. Once SOFT_CLOSE has been called, the driver 
   will allow another exclusive open on the main device, and so concurrent
   accesses by the original descriptor are possible (but probably not
   desirable). */   

	   case RCIIOC_SOFT_CLOSE:
	    { if (isMmapDev)
	       { return (ENOTTY);
	       }
	      doclose (info);      
	      break;
	    }
	 }
	return (0);
}

/*------------------------ */

#if MAPPED_IO
#define OUTB(p,v)	regs->p = (v)
#define OUTW(p,v)	regs->p = (v)
#define INB(p)		regs->p
#define INW(p)		regs->p
#else
#define OUTB(p,v)	outb(info->iobase+EPP_##p,v)
#define OUTW(p,v)	outw(info->iobase+EPP_##p,v)
#define INB(p)		inb(info->iobase+EPP_##p)
#define INW(p)		inw(info->iobase+EPP_##p)
#endif

static void eppHardwareReset (EppRegs *regs, EppSoftc *info)
{
	/* make the device quiescent. Called at attach and detach time */

	/* This is the device initialization sequence described in the
	   EPP manual, page 4-37. The counter initializtion is ignored */

	OUTW(cfg1stat,EPP_LRESET1);
	OUTW(cfg2,EPP_LRESET2);
	OUTW(cfg3,0);
	OUTW(cfg1stat,0);
	OUTW(cfg2,0);
	OUTW(dmaclr1,0);
	OUTW(dmaclr2,0);
	OUTW(cntintclr,0);

	/* Now set the directions */

	OUTW(cfg3,(EPP_WRITEC|EPP_WRITED));
	OUTW(cfg1stat,(EPP_DIOAEN|EPP_DIOBEN|EPP_PULSE1));
}

static void eppHardwareIntrSetup (EppRegs* regs, EppSoftc* info)
{
	/* if the device interrupt vector or IPL is programmable,
	   do this here based on info provided in info->idevCookie. */

	/* Nothing to do for EPP */
}

static void eppHardwareClose(EppRegs *regs, EppSoftc *info)
{
	/* make the device quiet when it is closed */

	/* disable interrupts */
	OUTW(cfg1stat,(EPP_DIOAEN|EPP_DIOBEN|EPP_PULSE1)); 
}

static void eppHardwareOpen(EppRegs *regs, EppSoftc *info)
{
	/* initialize the device when it is openned */	

	OUTW(cfg3,(EPP_WRITEC|EPP_WRITED));
	OUTW(cfg1stat,(EPP_DIOAEN|EPP_DIOBEN|EPP_PULSE1));
}

static void eppHardwareIntrEnable(EppRegs *regs, EppSoftc *info, int enable)
{
	/* enable/disable device interrupts */

	int temp;

	if (enable)
	 { OUTW(cfg1stat,(EPP_DIOAEN|EPP_DIOBEN|EPP_PULSE1|EPP_INTEN1));
	   temp = INW(portCD); /* just paranoid; clear interrupts */
	 }
	else
	 { OUTW(cfg1stat,(EPP_DIOAEN|EPP_DIOBEN|EPP_PULSE1));
	 }
}

static int eppHardwareIntrHandle(EppRegs *regs, EppSoftc *info, int *stray)
{
	/* clear interrupts, return true if a valid interrupt is pending,
	   and return a mask of any stray interrupts in *stray. */
	int pending;
	int temp;

	pending = (INW(cfg1stat) & 
		   (EPP_DRDY1|EPP_DRDY2|EPP_DMATC1|EPP_DMATC2|EPP_CNTINT));

	if (pending & EPP_DRDY1)
	 { temp = INW(portAB);
	 }
	if (pending & (EPP_DRDY2|EPP_DMATC1|EPP_DMATC2|EPP_CNTINT))
	 { /* These shouldn't happen, but we check for them anyway, and
	      try to clear the interrupt in each case. */
	   if (pending & EPP_DRDY2)
	    { temp = INW(portCD);
	    }
	   if (pending & EPP_DMATC1)
	    { OUTW(dmaclr1,0);
	    }
	   if (pending & EPP_DMATC2)
	    { OUTW(dmaclr2,0);
	    }
	   if (pending & EPP_CNTINT)
	    { OUTW(cntintclr,0);
	    }
	 }
	if (info->flags & EPP_BUSY)
	 { *stray = (pending & ~EPP_DRDY1);
	   return (pending & EPP_DRDY1);
	 }
	else
	 { /* device isn't open; therefore, any interrupt is a stray */
	   *stray = pending;
	   return (0);
	 }
}
